
#include "BasicTypes.h"
#include "tcl_vpt.h"
#include <stdlib.h>

Tcl_Interp *TclInterp = NULL;

int Basictypes_Init(Tcl_Interp *interp)
{
  TclInterp = interp;
  return TCL_OK;
}

/*
#define assert(ex) {if (!(ex)){ \
  (void)fprintf(stderr, "Assertion failed: file %s, line %d\n", \
        __FILE__, __LINE__); exit(2);}}
*/

VLObject::~VLObject() {}

void VLValue::setCharPtr(VLConstCharPtr val) 
{
    unshare();
    if (m_val)
      Tcl_SetStringObj(m_val, (char *)val, -1);
    else {      
      m_val = Tcl_NewStringObj((char *)val, -1);
      Tcl_IncrRefCount(m_val);
    }
}


Tcl_Obj *VLValue::eval(void)
{
        int ret = Tcl_GlobalEvalObj(TclInterp, m_val);
    if (ret != TCL_OK) 
    {
        throw VLString("Error while executing command: ", VLString( Tcl_GetObjResult(TclInterp)));
    }
    return Tcl_GetObjResult(TclInterp);
}
        


VLValue::VLValue():  m_val(0)
{
}

VLValue::VLValue(const VLValue &v):  m_val(v.m_val)
{
    if (m_val)
    Tcl_IncrRefCount(m_val);
}

VLValue::VLValue(Tcl_Obj *val): m_val(val)
{
    if (m_val)
      Tcl_IncrRefCount(m_val);
}

void VLValue::setResult(void)
{
  Tcl_SetObjResult(TclInterp, m_val);    
}



const VLValue &VLValue::operator=(const VLValue &v)
{
    if (&v != this) {
    if (v.m_val != m_val) {
        if (m_val) {

          Tcl_DecrRefCount(m_val);
        }
        m_val = v.m_val;
        if (m_val) {
          Tcl_IncrRefCount(m_val);
        }
    }
    }
    
    return *this;
}

VLValue::~VLValue()
{
    // bug workaround: when global strings are deallocated, there is a STATUS_ACCESS_VIOLATION
    if (PROG_IS_EXITING) return; 

    if (m_val) {
      Tcl_DecrRefCount(m_val);
      m_val = NULL;
    }
}
    
Tcl_Obj *VLValue::getObj() const
{
    if (m_val)
    return m_val;
    else
    return Tcl_NewObj();
}

void VLValue::unshare()
{
    if (m_val && Tcl_IsShared(m_val)){
      Tcl_Obj *ob = Tcl_DuplicateObj(m_val);
      Tcl_DecrRefCount(m_val);
      m_val = ob;
      Tcl_IncrRefCount(m_val);
    } 
}

void VLValue::setObj(Tcl_Obj *val)
{
    if (val != m_val) {
    if (m_val)
        Tcl_DecrRefCount(m_val);
    m_val = val;
    if (m_val) {
        Tcl_IncrRefCount(m_val);
    }
    }
}


void VLValue::prepare_modify() 
{
    if (!m_val)
       setObj(Tcl_NewObj());
    else
       unshare();
}

VLCharPtr VLValue::getStringRep() const 
{
    if (m_val)
        return Tcl_GetStringFromObj(m_val, (int *)0); 
    else
    return "";    
}
VLLong::VLLong(long val): VLValue()
{
    setLong(val);
}

VLLong::operator long() const
{
    return getLong();
}
    
long VLLong::getLong() const 
{
    long l = 0;
    if (m_val)
    Tcl_GetLongFromObj((Tcl_Interp *)0, m_val, &l);
    return l;
}
    
void VLLong::setLong(long val) 
{
    unshare();
    if (m_val)
    Tcl_SetLongObj(m_val, val);
    else {
    m_val = Tcl_NewLongObj(val);
    Tcl_IncrRefCount(m_val);
    }
    
}
VLInt::VLInt(int val)
{
    setInt(val);
}

VLInt::operator int() const
{
    return getInt();
}
    
int VLInt::getInt() const 
{
    int val = 0;
    if (m_val)
    Tcl_GetIntFromObj((Tcl_Interp *)0, m_val, &val);
    return val;
}
    
void VLInt::setInt(int val) 
{
    unshare();
    if (m_val)
    Tcl_SetIntObj(m_val, val);
    else {
    m_val = Tcl_NewIntObj(val);
    Tcl_IncrRefCount(m_val);
    }
    
}
/*
char *VLInt::getNewArgs(char *buffer)
{
   //sprintf (buffer, "ptr%x.i", (unsigned int)this);
   sprintf (buffer, "ptr%p.i", this);
   return buffer;
}
*/
char *VLInt::getValArgs(char *buffer)
{
   sprintf (buffer, "%d", getInt());
   return buffer;
}

VLBoolean::VLBoolean()
{
    setBoolean(0);
}

VLBoolean::VLBoolean(int val): VLValue()
{
    setBoolean(val);
}

VLBoolean::operator int() const
{
    return getBoolean();
}
    
int VLBoolean::getBoolean() const 
{
    int val = 0;
    if (m_val)
    Tcl_GetBooleanFromObj((Tcl_Interp *)0, m_val, &val);
    return val;
}
    
void VLBoolean::setBoolean(int val) 
{
    unshare();
    if (m_val)
    Tcl_SetBooleanObj(m_val, val);
    else {
    m_val = Tcl_NewBooleanObj(val);
    Tcl_IncrRefCount(m_val);
    }
    
}
VLString::VLString(VLConstCharPtr val): VLValue()
{
    setCharPtr(val);
}


VLString::VLString(int val): VLValue()
{
    m_val = Tcl_NewIntObj(val);
}


VLString::operator VLCharPtr() const
{
    return getCharPtr();
}


VLCharPtr VLString::getTempCharPtr() const
{
    if (m_val)
        return Tcl_GetString(m_val);
    else
        return "";
}


VLCharPtr VLString::getCharPtr() const 
{
    // FIXME: Check for memory leaks
    return strdup(getTempCharPtr());
}


int VLString::equals(const VLString &x) {
    return strcmp(getTempCharPtr(), x.getTempCharPtr())==0;
}


void VLString::setCharPtr(VLConstCharPtr val) 
{
    unshare();
    if (m_val)
      Tcl_SetStringObj(m_val, (char *)val, -1);
    else {      
      m_val = Tcl_NewStringObj((char *)val, -1);
      Tcl_IncrRefCount(m_val);
    }
}


void VLString::append(VLCharPtr val)
{
    prepare_modify();
    Tcl_AppendStringsToObj(m_val,val,NULL);
}


int VLString::getLength()
{
    if (m_val)
        return Tcl_GetCharLength(m_val);
    else
    return 0;
}

VLString VLString::getSubstring(int from, int count)
{
    if (m_val)
    {
        VLString result = VLString(Tcl_GetRange(m_val,from,from - 1 + count));
        return result;
    }
    else
    return 0;
    
    // char *s = getCharPtr();
    // assert(from+count <= getLength());
    // s[from+count] = 0;
    // return VLString(s+from);
   
}

VLDouble::VLDouble(double val): VLValue()
{
    setDouble(val);
}

VLDouble::operator double() const
{
    return getDouble();
}

double VLDouble::getDouble() const 
{
    double val = 0;
    if (m_val)
    Tcl_GetDoubleFromObj((Tcl_Interp *)0, m_val, &val);
    return val;
}

void VLDouble::setDouble(double val) 
{
    unshare();
    if (m_val)
    Tcl_SetDoubleObj(m_val, val);
    else {
    m_val = Tcl_NewDoubleObj(val);
    Tcl_IncrRefCount(m_val);
    }

}
VLList::VLList(): VLValue()
{
}


VLList::VLList(VLConstCharPtr s): VLValue(Tcl_NewStringObj(VLCharPtr(s), -1))
{
}



VLList *VLList::append(Tcl_Obj *obj)
{
    prepare_modify();
    Tcl_ListObjAppendElement((Tcl_Interp *)0, m_val, obj);
    return this;
}

VLList *VLList::append(int i)
{
    prepare_modify();

    Tcl_ListObjAppendElement((Tcl_Interp *)0, m_val, Tcl_NewIntObj(i));
    return this;
}

VLList *VLList::append(VLCharPtr s)
{
    prepare_modify();

    Tcl_ListObjAppendElement((Tcl_Interp *)0, m_val, Tcl_NewStringObj(s, -1));
    return this;
}

VLList *VLList::append(double d)
{
    prepare_modify();

    Tcl_ListObjAppendElement((Tcl_Interp *)0, m_val, Tcl_NewDoubleObj(d));
    return this;
}

VLList *VLList::append(const VLValue &v)
{
    return append(v.getObj());
}

VLList *VLList::appendList(const VLList &v)
{
    prepare_modify();
    Tcl_ListObjAppendList((Tcl_Interp *)0, m_val, v.getObj());
    return this;
}


/* by cs: Dies verursacht schwer zu findende Fehler: 
Tcl_ListObjAppendElement((Tcl_Interp *)0, m_val, TclVP_NewVoidPointerObj(ptr));
*/

VLList *VLList::append(VLCharPtr name, int i)
{
    prepare_modify();

    Tcl_ListObjAppendElement((Tcl_Interp *)0, m_val, Tcl_NewStringObj(name, -1));
    Tcl_ListObjAppendElement((Tcl_Interp *)0, m_val, Tcl_NewIntObj(i));
    return this;
}

VLList *VLList::append(VLCharPtr name, VLCharPtr s)
{
    prepare_modify();

    Tcl_ListObjAppendElement((Tcl_Interp *)0, m_val, Tcl_NewStringObj(name, -1));
    Tcl_ListObjAppendElement((Tcl_Interp *)0, m_val, Tcl_NewStringObj(s, -1));
    return this;
}

VLList *VLList::append(VLCharPtr name, double d)
{
    prepare_modify();

    Tcl_ListObjAppendElement((Tcl_Interp *)0, m_val, Tcl_NewStringObj(name, -1));
    Tcl_ListObjAppendElement((Tcl_Interp *)0, m_val, Tcl_NewDoubleObj(d));
    return this;
}

VLList *VLList::append(VLCharPtr name, void *ptr)
{
    prepare_modify();

    Tcl_ListObjAppendElement((Tcl_Interp *)0, m_val, Tcl_NewStringObj(name, -1));
    Tcl_ListObjAppendElement((Tcl_Interp *)0, m_val, TclVP_NewVoidPointerObj(ptr));
    return this;
}

int VLList::getLength() 
{
    int len = 0;

    if (m_val)
    Tcl_ListObjLength((Tcl_Interp *)0, m_val, &len);
    return len;
}

VLValue VLList::getIndex(int i)
{
   Tcl_Obj *objPtr;
   if(Tcl_ListObjIndex((Tcl_Interp *)0, m_val, i, &objPtr) == TCL_OK) return VLValue(objPtr);
   assert(0);
   return VLValue();
}

VLPointArray::VLPointArray()
{
}


VLPointArray::VLPointArray(int x, int y)
{
    setXY(0, x, y);
}

VLPointArray::VLPointArray(int x1, int y1, int x2, int y2)
{
    setXY(0, x1, y1);
    setXY(1, x2, y2);
}

VLPointArray::VLPointArray(VLPoint a, VLPoint b)
{
    setXY(0, a);
    setXY(1, b);
}

int VLPointArray::getX(int index, int defaultval) const
{
    double x = defaultval;
    Tcl_Obj *X = (Tcl_Obj *)0;
    
    if (m_val) {
    if (Tcl_ListObjIndex((Tcl_Interp *)0, m_val, index * 2, &X) == TCL_OK) {
        if (X != (Tcl_Obj *)0) {
        Tcl_GetDoubleFromObj((Tcl_Interp *)0, X, &x);
        }
    }
    }
    
    return int(x);
}

int VLPointArray::getY(int index, int defaultval) const
{
    double y = defaultval;
    Tcl_Obj *Y = (Tcl_Obj *)0;
    
    if (m_val) {
    if (Tcl_ListObjIndex((Tcl_Interp *)0, m_val, index * 2 + 1, &Y) == TCL_OK) {
        if (Y != (Tcl_Obj *)0) {
        Tcl_GetDoubleFromObj((Tcl_Interp *)0, Y, &y);
        }
    }
    }
    
    return int(y);
}
 
int VLPointArray::getNumbPoints() const
{
    int len = 0;
    
    if (m_val)
    Tcl_ListObjLength((Tcl_Interp *)0, m_val, &len);
    return len / 2;
}

void VLPointArray::removePoints(int index, int count)
{
    // Tcl_ListObjReplace strzt ab, wenn index + count > len && count <= len gilt.

    int len;
    if (!m_val) return;
    len = Tcl_ListObjLength((Tcl_Interp *)0, m_val, &len);

    if (2 * index + 2 * count > len) 
    count = len / 2 - index;

    Tcl_ListObjReplace((Tcl_Interp *)0, m_val, 2 * index, 2 * count, 0, (Tcl_Obj **)0);
}

void VLPointArray::setXY(int index, VLPoint p)
{
    setXY(index,p.getX(), p.getY());
}

void VLPointArray::setXY(int index, int x, int y)
{
    Tcl_Obj *coords[2];
    int len;
    
    unshare();
    len = getNumbPoints();

    if (index > len) {
    coords[1] = Tcl_NewIntObj(0);
        coords[0] = Tcl_NewIntObj(0);
    while (index > len) {
        if (m_val) 
        Tcl_ListObjReplace((Tcl_Interp *)0, m_val, 2*len, 0, 2, coords);
        else {
        m_val = Tcl_NewListObj(2, coords);
        Tcl_IncrRefCount(m_val);
        }
        
        len++;
    }
    }
    
    coords[0] = Tcl_NewIntObj(x);
    coords[1] = Tcl_NewIntObj(y);


    if (m_val) 
    Tcl_ListObjReplace((Tcl_Interp *)0, m_val, index * 2, (index < len) ? 2 : 0, 2, coords);
    else {
    m_val = Tcl_NewListObj(2, coords);
    Tcl_IncrRefCount(m_val);
    }

}


void VLPointArray::print(char *s="")
  {
  printf("%s\n",s);
  for(int i=0; i<getNumbPoints(); i++)
    {
    printf("(%d,%d)\n",getX(i),getY(i)); 
    }
  }

/*
char *VLPointArray::getNewArgs(char *buffer)
{
   //sprintf(buffer, "ptr%x.x ptr%x.y", (unsigned int)this, (unsigned int)this);
   sprintf(buffer, "ptr%p.x ptr%p.y", this, this);
   return buffer;
}
*/
char *VLPointArray::getValArgs(char *buffer)
{
   sprintf(buffer, "%d %d", getX(), getY());
   return buffer;
}


void VLPointArray::append(VLPoint p)
{
  append(p.getX(), p.getY());
}

VLPointArray VLPointArray::addToAll(VLPoint p)
{
  VLPointArray arr;
  for(int i=0;i< getNumbPoints();i++)
  {
      arr.append(getPoint(i).add(p));
  }
  return arr;
}

VLPoint VLPointArray::getPoint(int index)
{
  return VLPoint(getX(index), getY(index));
}


VLPoint VLPointArray::getFirst()
{
  return getPoint(0);
}

VLPoint VLPointArray::getLast()
{
  return getPoint(getNumbPoints()-1);
}

VLRegion VLPoint::getSquare(int size) {
  return VLRegion(VLPoint(getX()-size, getY()-size), VLPoint(2*size, 2*size));
}

    
void *VLVoidPointer::getPtr() const 
{
    void *l = 0;
    if (m_val) {
      int status = TclVP_GetVoidPointerFromObj((Tcl_Interp *)0, m_val, &l);
      if(status != TCL_OK) {
       throw VLString("Invalid pointer value: ",  VLString(getObj()));
      }
    }
    return l;
}
    
void VLVoidPointer::setPtr(void *val) 
{
    unshare();
    if (m_val)
    TclVP_SetVoidPointerObj(m_val, val);
    else {
    m_val = TclVP_NewVoidPointerObj(val);
    Tcl_IncrRefCount(m_val);
    }
    
}



