
#ifndef BASICTYPES_H
#define BASICTYPES_H

#include <stddef.h>
#include "tcl.h"
#include "string.h"
#include "tcl_vpt.h"
#include "stdlib.h"
#include <assert.h>
#include <math.h>
#include "Tclinter_Object.h"


extern Tcl_Interp *TclInterp;

int Basictypes_Init(Tcl_Interp *interp);


inline int vlMax(int a, int b)    { return (a>b ? a : b); }
inline int vlAbs(int a)           { return (a>0 ? a : -a); }
inline int vlSign(int s, int val) { return (s>=0 ? val : -val); }
inline int vlSign(int s)          { return (s==0 ? 0 : (s>0?1:-1) ); }
typedef char *VLCharPtr;
typedef char const *VLConstCharPtr;

extern int PROG_IS_EXITING; // defined in main.c (tclinterGen)

class VLObject
{
public:
    virtual ~VLObject();
};

typedef VLObject *VLObjectPtr;

class VLValue: public VLObject
{
protected:
    Tcl_Obj *m_val;

    void unshare();
    void prepare_modify();

public:
    Tcl_Obj *eval(void); /* by cs */
    VLValue();
    VLValue(const VLValue &v);
    VLValue(Tcl_Obj *val);
    VLValue(VLConstCharPtr val): m_val(NULL) { setCharPtr(val); }
    void setCharPtr(VLConstCharPtr val);
    const VLValue &operator=(const VLValue &v);
    void setResult(void);

    int isNull(void) { 
      char *s = getStringRep();
      return s[0] == 0;
    }

    virtual ~VLValue();

    virtual Tcl_Obj *getObj() const;
    virtual void setObj(Tcl_Obj *val);

    VLCharPtr getStringRep() const;
};

typedef VLValue *VLValuePtr;
        

class VLLong: public VLValue
{
public:
    VLLong(long val = 0);
    VLLong(const VLValue &v) : VLValue(v) {};

    operator long() const;
    
    long getLong() const;
    void setLong(long val);
};

typedef VLLong *VLLongPtr;
#define VLLONG(v) (new VLLong(v))


class VLInt: public VLValue
{
public:
    VLInt(Tcl_Obj * obj) : VLValue(obj) {};
    VLInt(const VLValue &v) : VLValue(v) {};
    VLInt(int val);
    VLInt() {};

    operator int() const;
    
    int getInt() const;
    void setInt(int val);

    //virtual char *getNewArgs(char *);
    virtual char *getValArgs(char *);
};

typedef VLInt *VLIntPtr;
        
#define VLINT(v) (new VLInt(v))
#define VLINTVAL(v) ((v)->getInt())

class VLBoolean: public VLValue
{
public:
    VLBoolean(const VLValue &v) : VLValue(v) {};
    VLBoolean(int val);
    VLBoolean();

    operator int() const;
    
    int getBoolean() const;
    void setBoolean(int val);
    
};

typedef VLBoolean *VLBooleanPtr;


class VLString: public VLValue
{
public:
    VLString(Tcl_Obj * obj) : VLValue(obj) {};
    VLString(const VLValue &v) : VLValue(v) {};
    VLString(VLString a, VLString b) { append(a); append(b); }
    VLString() {};
    VLString(VLConstCharPtr val);
    VLString(int val);

    const bool operator==(const VLString &other) { return equals(other); }
    operator VLCharPtr() const;
    
    VLCharPtr getCharPtr() const;
    VLCharPtr getTempCharPtr() const;
    void setCharPtr(VLConstCharPtr val);
    void append(VLCharPtr val);
    int getLength();
    VLString getSubstring(int from, int count);

    int equals(const VLString &x);
};

typedef VLString *VLStringPtr;


inline int isEmptyString(VLCharPtr s) { return s[0]==0; }
        
inline VLCharPtr GetVLCharPtr(VLString *s) { return s->getCharPtr(); }


class VLDouble: public VLValue
{
public:   
    VLDouble(double val = 0.0);
    VLDouble(const VLValue &v) : VLValue(v) {};

    operator double() const;
    
    double getDouble() const;
    void setDouble(double val);
};

typedef VLDouble *VLDoublePtr;
        
#define VLDOUBLE(v) (new VLDouble(v))

class VLList: public VLValue
{
public:
    VLList();
    VLList(Tcl_Obj * obj) : VLValue(obj) {}
    VLList(const VLValue &v) : VLValue(v) {}
    VLList(VLConstCharPtr s);
    VLList(VLCharPtr s1, const VLList &v) { append(s1); appendList(v); }
    VLList(VLCharPtr s1, VLCharPtr s2, const VLList &v) { append(s1); append(s2); appendList(v); }
    VLList(VLCharPtr s1, VLValue s2, const VLList &v) { append(s1); append(s2); appendList(v); }
    VLList(VLValue x, const VLList &v) { append(x); appendList(v); }
    VLList *append(Tcl_Obj *obj);
    VLList *append(int i);
    VLList *append(VLCharPtr s);
    VLList *append(double d);
    VLList *append(const VLValue &v);
    VLList *append(VLCharPtr name, int i);
    VLList *append(VLCharPtr name, VLCharPtr s);
    VLList *append(VLCharPtr name, double d);
    VLList *append(VLCharPtr name, void *ptr);
    VLList *appendList(const VLList &v);

    int getLength();
    VLValue getIndex(int i);
    VLList getAppend(const VLList &v) { VLList l = VLList(); l.appendList(*this); l.appendList(v); return l; }
};

typedef VLList *VLListPtr;


inline VLList vlList() { return VLList(); }
inline VLList vlList(VLValue a) { return VLList(a, vlList()); }
inline VLList vlList(VLValue a, VLValue b) { return VLList(a, vlList(b)); }
inline VLList vlList(VLValue a, VLValue b, VLValue c) { return VLList(a, vlList(b,c)); }
inline VLList vlList(VLValue a, VLValue b, VLValue c, VLValue d) { return VLList(a, vlList(b,c,d)); }
inline VLList vlList(VLValue a, VLValue b, VLValue c, VLValue d, VLValue e) { return VLList(a, vlList(b,c,d,e)); }
inline VLList vlList(VLValue a, VLValue b, VLValue c, VLValue d, VLValue e, VLValue f) { return VLList(a, vlList(b,c,d,e,f)); }
inline VLList vlList(VLValue a, VLValue b, VLValue c, VLValue d, VLValue e, VLValue f, VLValue g) { return VLList(a, vlList(b,c,d,e,f,g)); }
inline VLList vlList(VLValue a, VLValue b, VLValue c, VLValue d, VLValue e, VLValue f, VLValue g, VLValue h) { return VLList(a, vlList(b,c,d,e,f,g,h)); }
inline VLList vlList(VLValue a, VLValue b, VLValue c, VLValue d, VLValue e, VLValue f, VLValue g, VLValue h, VLValue i) { return VLList(a, vlList(b,c,d,e,f,g,h,i)); }
inline VLList vlList(VLValue a, VLValue b, VLValue c, VLValue d, VLValue e, VLValue f, VLValue g, VLValue h, VLValue i, VLValue j) { return VLList(a, vlList(b,c,d,e,f,g,h,i,j)); }
inline VLList vlList(VLValue a, VLValue b, VLValue c, VLValue d, VLValue e, VLValue f, VLValue g, VLValue h, VLValue i, VLValue j, VLValue k) { return VLList(a, vlList(b,c,d,e,f,g,h,i,j, k)); }
inline VLList vlList(VLValue a, VLValue b, VLValue c, VLValue d, VLValue e, VLValue f, VLValue g, VLValue h, VLValue i, VLValue j, VLValue k, VLValue l) { return VLList(a, vlList(b,c,d,e,f,g,h,i,j, k, l)); }


#define VLEVAL(l) ((l)->eval())
#define VLLIST() (new VLList())
#define VLLIST1(x) ((new VLList())->append(x))
#define VLLIST2(x,y) (VLLIST1(x)->append(y))
#define VLLIST3(x,y,z) (VLLIST2(x,y)->append(z))
#define VLAPPEND(l,x) ((l)->append(x))
#define VLAPPEND2(l,name,x) ((l)->append(name,x))
#define NOVLLIST (VLListPtr(0))


class VLPoint;

class VLPointArray: public VLValue
{
public:
    VLPointArray();
    VLPointArray(const VLValue &v) : VLValue(v) {};
    VLPointArray(int x, int y);       // Makes a Point Array consisting of one Point
    VLPointArray(int x1, int y1, int x2, int y2);
    VLPointArray(VLPoint a, VLPoint b);
    VLPointArray(Tcl_Obj * obj): VLValue(obj) {};

    
    int getX(int index = 0, int defaultval = 0) const;    // Returns X of Point number index
    int getY(int index = 0, int defaultval = 0) const;    // Returns Y of Point number index.
    int getNumbPoints() const;        // Returns Number of Points.
    
    void removePoints(int index, int count = 1);
                                      // Removes the specified points.
    
    void setXY(int index, VLPoint p);

    void setXY(int index, int x, int y);
                                      // Changes the Point coordinates or
                                      // inserts at the end. Also inserts null
                                      // elements if required.

    void append(int x, int y) { setXY(getNumbPoints(), x, y); }
    void append(VLPoint p);
    VLPointArray addToAll(VLPoint p);

    VLPoint getPoint(int index);
    VLPoint getFirst();
    VLPoint getLast();

    void print(char *);

    //virtual char *getNewArgs(char *buffer);
    virtual char *getValArgs(char *buffer);
};

typedef VLPointArray *VLPointArrayPtr;

#define NOVLPOINT ((VLPointArrayPtr)0)
#define VLPOINT(x,y) (new VLPointArray(x,y))
#define VLREGION(x1,y1,x2,y2) (new VLPointArray(x1,y1,x2,y2))


class VLRegion;

class VLPoint: public VLPointArray
{
public:    
    VLPoint(): VLPointArray() {}
    VLPoint(const VLValue &v) : VLPointArray(v) {}
    VLPoint(Tcl_Obj * obj): VLPointArray(obj) {};
    VLPoint(int aX, int aY): VLPointArray(aX, aY) {}    
    VLPoint(int flip, int aX, int aY) { if(flip) { setXY(0,aY,aX); } else { setXY(0,aX,aY); } }
    int getX() const { return VLPointArray::getX(0); }
    int getY() const { return VLPointArray::getY(0); }
    int getX(int flip) const { return (flip?getY():getX()); }
    int getY(int flip) const { return (flip?getX():getY()); }
    VLPoint add(int dx, int dy) { return VLPoint(getX()+dx, getY()+dy); }
    VLPoint add(int flip, int dx, int dy) { if (flip) return VLPoint(getX()+dy, getY()+dx); else return VLPoint(getX()+dx, getY()+dy); }
    VLPoint sub(int dx, int dy) { return VLPoint(getX()-dx, getY()-dy); }
    VLPoint add(VLPoint p) { return VLPoint(getX()+p.getX(), getY()+p.getY()); }
    VLPoint sub(VLPoint p) { return VLPoint(getX()-p.getX(), getY()-p.getY()); }
    VLPoint mul(int f) { return VLPoint(f*getX(), f*getY()); }
    VLPoint div(int f) { return VLPoint(getX()/f, getY()/f); }
    VLPoint max(VLPoint p) { return VLPoint(vlMax(getX(),p.getX()), vlMax(getY(),p.getY())); }
    VLPoint scale(int length) { double f = length/sqrt(getX()*getX()+getY()*getY()); return VLPoint( (int)(f*getX()+0.5), (int)(f*getY()+0.5) ); }
    VLPoint neg() { return VLPoint(-getX(), -getY()); }
    int isZero() { return getX()==0 && getY()==0; }
    VLRegion getSquare(int size);
    int equals(const VLPoint &x) { return getX()==x.getX() && getY()==x.getY(); }
};

typedef VLPoint *VLPointPtr;


inline VLPoint vlMax(VLPoint a, VLPoint b) { return a.max(b); }


const int ALIGN_CENTER = 0;
const int ALIGN_LEFT   = 1;
const int ALIGN_RIGHT  = 2;
const int ALIGN_SCALE  = 3;

const int ALIGN_TOP    = 1;
const int ALIGN_BOTTOM = 2;


class VLRegion: public VLPointArray
{
public:
    VLRegion() {}
    VLRegion(VLPoint aPos, VLPoint aSize) { setPos(aPos); setSize(aSize); }
    VLRegion(const VLValue &v) : VLPointArray(v) {}

    void setPos(VLPoint aPos) { setXY(0, aPos.getX(), aPos.getY()); }
    void setPos2(VLPoint pos2) { setXY(1, pos2.getX(), pos2.getY()); }
    void setSize(VLPoint aSize) { setPos2( getPos().add(aSize) ); }
    void setPS(VLPoint pos, VLPoint size) { setPos(pos); setSize(size); }

    VLPoint getPos() const { return VLPoint(getX(), getY()); }
    VLPoint getPos2() const { return VLPoint(getX(1), getY(1)); }
    VLPoint getSize() const { return getPos2().sub(getPos()); }
    VLPoint getCenter() const { return getPos().add(getSize().div(2)); }

    VLPoint getPosNW() { return getPos(); }
    VLPoint getPosSW() { return VLPoint(getPos().getX(), getPos2().getY()); }
    VLPoint getPosNE() { return VLPoint(getPos2().getX(), getPos().getY()); }
    VLPoint getPosSE() { return getPos2(); }

    VLPoint getPosN() { return VLPoint(getCenter().getX(), getPos().getY()); }
    VLPoint getPosS() { return VLPoint(getCenter().getX(), getPos2().getY()); }
    VLPoint getPosW() { return VLPoint(getPos().getX(), getCenter().getY()); }
    VLPoint getPosE() { return VLPoint(getPos2().getX(), getCenter().getY()); }

    VLRegion enlarge(VLPoint p) { return VLRegion(getPos().add(p.div(-2)), getSize().add(p)); }
    VLRegion shrink(VLPoint p) { return enlarge(p.neg()); }

    void doEnlargeR(VLPoint p) { setSize(getSize().add(p)); }
    void doShrinkC(VLPoint p) { setPS(getPos().add(p.div(2)), getSize().add(p.neg())); }

    void doMove(VLPoint direction) { VLPoint size=getSize(); setPos(getPos().add(direction)); setSize(size); }

    int equals(const VLRegion &x) { return getPos().equals(x.getPos()) && getPos2().equals(x.getPos2()); }

    bool intersects(int a1, int a2, int b1, int b2) { return a1<b1 && a2>b1 || a1>=b1 && a1<b2; }

    bool intersects(VLRegion other) { 
      return intersects( getPos().getX(), getPos2().getX(), other.getPos().getX(), other.getPos2().getX() )
          && intersects( getPos().getY(), getPos2().getY(), other.getPos().getY(), other.getPos2().getY() );
    }

    void doAlign(int fullPos, int fullSize, int reqSize, int alignment, int *resPos, int *resSize) {
        switch(alignment) {
           case ALIGN_CENTER: 
             *resPos = fullPos + (fullSize-reqSize)/2;
             *resSize = reqSize;
             break;
           case ALIGN_LEFT:
             *resPos = fullPos;
             *resSize = reqSize;
             break;
           case ALIGN_RIGHT:
             *resPos = fullPos + (fullSize-reqSize);
             *resSize = reqSize;
             break;
           case ALIGN_SCALE:
             *resPos = fullPos;
             *resSize = fullSize;
             break;
           default:
             assert(0);
        }             
    }

    VLRegion align(VLPoint reqSize, int alignX, int alignY) {
        int x, y, w, h;
        doAlign(getPos().getX(), getSize().getX(), reqSize.getX(), alignX, &x, &w);
        doAlign(getPos().getY(), getSize().getY(), reqSize.getY(), alignY, &y, &h);
        return VLRegion(VLPoint(x,y), VLPoint(w,h));
    }

    VLPoint getLineIntersectionPoint(VLPoint direction) {
        // berechnet den Schnittpunkt des Rechtecks mit der Halbgeraden ausgehend
        // vom Mittelpunkt des Rechtecks in Richtung "direction".

        if (direction.isZero()) {
          direction = VLPoint(10,0);
        }
        VLPoint delta;
        int absDirX = vlAbs(direction.getX());
        int absDirY = vlAbs(direction.getY());
        int sizeX = getSize().getX();
        int sizeY = getSize().getY();

        if ( vlAbs(direction.getY())*getSize().getX() > vlAbs(direction.getX())*getSize().getY() ) {
           // Schnitt oben oder unten
           delta = VLPoint( direction.getX() * getSize().getY() / (2 * vlAbs(direction.getY())),
                          vlSign(direction.getY(), getSize().getY()/2) );
                          
        }
        else {
           // Schnitt rechts oder links
           //printf("BasicTypes XXX: %i %i\n", direction.getX(), 2 * vlAbs(direction.getX()));
           int sizeX = getSize().getX();
           delta = VLPoint( vlSign(direction.getX(), getSize().getX()/2),
                          direction.getY() * getSize().getX() / (2 * vlAbs(direction.getX())) );
        }
        return getCenter().add(delta);
    }
};

typedef VLRegion *VLRegionPtr;


inline VLRegion VLRegionFromPoints(VLPoint a, VLPoint b) { VLRegion r; r.setPos(a); r.setPos2(b); return r; }
// XXX: VLRegionFromPoints -> vlRegionFromPoints

typedef void *VoidPtr;
typedef void *VoidPointer;

class VLVoidPointer: public VLValue
{
public:
    VLVoidPointer(Tcl_Obj * obj) : VLValue(obj) {};
    VLVoidPointer(const VLValue &v) : VLValue(v) {};
    VLVoidPointer(void *val = 0) { setPtr(val); }

    operator VoidPointer() const { return getPtr(); }

    int isNull() { return getPtr()==NULL; }
    void *getPtr() const;
    void setPtr(void *val);
};

typedef VLVoidPointer *VLVoidPointerPtr;
        
#define VLVOIDPOINTER(v) (new VLVoidPointer(v))


class VLObjectPointer: public VLVoidPointer {
 public:
  VLObjectPointer(Tclinter_Object *obj=0): VLVoidPointer(obj) {}

  VLObjectPointer(Tcl_Obj *obj): VLVoidPointer(obj) { check(); }
  VLObjectPointer(const VLValue &v): VLVoidPointer(v) { check(); }

  const bool operator==(const VLObjectPointer &other) { return getValue()==other.getValue(); }

  void check() {
    if (!isValid()) throw VLString("Not a valid object pointer: ",VLString(getObj()));
  }

  bool isValid() {
    return isNull() || Tclinter_Object::exists((Tclinter_Object *)getPtr());
  }

  operator TclinterObjectPtr() const { return getValue(); }


  TclinterObjectPtr getValue() const { 
    return (TclinterObjectPtr)getPtr();
  }
};

typedef VLObjectPointer TreeContext;
typedef VLObjectPointer VLObjectPointerPtr2;
// BOGUS: die "2" ist nur provisorisch wegen PaderWAVE Spezifikation



#include "VLHashtable.h"

#endif
