
// DynamicalArray
//
// Version 1.3
//
// Zielsetzung: Mglichst komfortable Anwendung. Aber Achtung: Elemente werden
// evtl. kopiert (problematisch, wenn sie nicht primitiv sind). DynamicalArray
// ist hauptschlich fr Objektzeiger gedacht. 
//
// Template, damit kein Typecast bentigt wird.
// Effizienz: Array-Zugriff: O(1), addElement: durchschnittlich O(1)
// Speicherverschwendung: weniger als bei linearer Liste, wenn
// Zeiger verwaltet werden


#ifndef DYNAMICAL_ARRAY_H
#define DYNAMICAL_ARRAY_H

#include <stdlib.h>
#include <assert.h>

typedef int (*qsort_compare)(const void *, const void *);
// void qsort( void *base, size_t num, size_t width, qsort_compare comp );


template <class T>
class DynamicalArray
{
public:
	DynamicalArray(void) { init(0,1); }
        DynamicalArray(const DynamicalArray &x);
        DynamicalArray(int s) { init(s,s); }
        DynamicalArray(int s, T val) { init(s,s); for(int i=0; i<s; i++) value(i)=val; }
	~DynamicalArray() { delete [] buffer; }
        void operator=(const DynamicalArray &x);
	typedef int (*compareFkt)(T*,T*);
        void setSize(int s);
        void setSize(int s, T val) { setSize(s); for(int i=0; i<s; i++) value(i)=val; }
	int getSize(void) { return validSize; }
	void sort(compareFkt comp) { qsort(buffer,validSize,sizeof(T),(qsort_compare)comp); }
	T &value(int i) { assert(0<=i && i<validSize); return buffer[i]; }
	const T &value(int i) const { assert(0<=i && i<validSize); return buffer[i]; }
	void removeLast() { assert(validSize>0); validSize--; }
	void removeAll() { validSize=0; }
	void add(T);

        // Andere Notationen:
	T &operator [](int i) { return value(i); }
	const T &operator [](int i) const { return value(i); }
	T &last(void) { return value(validSize-1); }
	void assign(int i,T val) { value(i)=val; }

private:
        void init(int aValidSize, int aBufferSize)
        { validSize=aValidSize; bufferSize=aBufferSize; buffer=new T[bufferSize]; }
	int validSize, bufferSize;
	T *buffer;
};



template <class T> 
inline DynamicalArray<T>::DynamicalArray(const DynamicalArray &x)
{
  validSize=x.validSize;
  bufferSize=x.bufferSize;
  buffer=new T[bufferSize];
  for(int i=0; i<bufferSize; i++) buffer[i]=x.buffer[i];
}

template <class T> 
inline void DynamicalArray<T>::operator=(const DynamicalArray &x)
{
  delete [] buffer;
  validSize=x.validSize;
  bufferSize=x.bufferSize;
  buffer=new T[bufferSize];
  for(int i=0; i<bufferSize; i++) buffer[i]=x.buffer[i];
}  

template <class T> 
inline void DynamicalArray<T>::setSize(int s)
{
  delete [] buffer;
  validSize=s;
  bufferSize=s;
  buffer=new T[bufferSize];
}  


template <class T> 
inline void DynamicalArray<T>::add(T elem)
{
	if(validSize==bufferSize)
	{
		T *newBuffer=new T[2*bufferSize];
		for(int i=0; i<bufferSize; i++) newBuffer[i]=buffer[i];
		delete [] buffer;
		buffer=newBuffer;
		bufferSize*=2;
	}
	buffer[validSize++]=elem;
}


#endif



