
#include "VisAutoPatternCode.h"


int VisTileDrawing::expandX(float fLeft,float fRight,int minSize)
{
  int left=int(floor(fLeft));
  int right=int(ceil(fRight));
  //Suche erste und letzte dragLine
  int i,first,end, currSize, staticSize, dynamicalSize;
  first=0;
  while(first<dragLineX.getSize() && dragLineX[first].pos<left) first++;
  end=first;
  while(end<dragLineX.getSize() && dragLineX[end].pos<right) end++;
  if(minSize!=0) assert(first<end);
  //berechne aktuelle Groesse
  staticSize=sumX[right]-sumX[left];
  dynamicalSize=0;
  for(i=first; i<end; i++) dynamicalSize+=dragLineX[i].getRepSize();
  currSize=staticSize+dynamicalSize;
  int delta=std::max(0,minSize-currSize);

  //Suche drag-Lines mit hoechster Prioritaet
  int maxpri=0;
  for(i=first; i<end; i++) maxpri=std::max(maxpri,dragLineX[i].priority);
  int countmaxpri=0;
  for(i=first; i<end; i++) if(dragLineX[i].priority==maxpri) countmaxpri++;

  //vergroessere moeglichst gleichmaessig
  i=first;
  while(delta>0)
  {
    assert(i<end);
    if(dragLineX[i].priority==maxpri)
    {
      int delta2=delta/(countmaxpri--);
      int deltaRep=(delta2+dragLineX[i].size-1)/dragLineX[i].size;
      dragLineX[i].repetitions+=deltaRep;
      delta-=deltaRep*dragLineX[i].size;
    }
    i++;
  }
  return currSize-delta;
}


int VisTileDrawing::expandY(float fLeft,float fRight,int minSize)
{
  int left=int(floor(fLeft));
  int right=int(ceil(fRight));
  //Suche erste und letzte dragLine
  int i,first,end, currSize, staticSize, dynamicalSize;
  first=0;
  while(first<dragLineY.getSize() && dragLineY[first].pos<left) first++;
  end=first;
  while(end<dragLineY.getSize() && dragLineY[end].pos<right) end++;
  if(minSize!=0) assert(first<end);
  //berechne aktuelle Groesse
  staticSize=sumY[right]-sumY[left];
  dynamicalSize=0;
  for(i=first; i<end; i++) dynamicalSize+=dragLineY[i].getRepSize();
  currSize=staticSize+dynamicalSize;
  int delta=std::max(0,minSize-currSize);

  //Suche drag-Lines mit hoechster Prioritaet
  int maxpri=0;
  for(i=first; i<end; i++) maxpri=std::max(maxpri,dragLineY[i].priority);
  int countmaxpri=0;
  for(i=first; i<end; i++) if(dragLineY[i].priority==maxpri) countmaxpri++;

  //vergroessere moeglichst gleichmaessig
  i=first;
  while(delta>0)
  {
    assert(i<end);
    if(dragLineY[i].priority==maxpri)
    {
      int delta2=delta/(countmaxpri--);
      int deltaRep=(delta2+dragLineY[i].size-1)/dragLineY[i].size;
      dragLineY[i].repetitions+=deltaRep;
      delta-=deltaRep*dragLineY[i].size;
    }
    i++;
  }
  return currSize-delta;
}


VisDrawingTrafo *VisTileDrawing::getDrawingTrafo(void)
{
  if(sizeX.getSize()==0)
  {
    initDrawing();
    int size, i, j;
    VLConstCharPtr name;
    // Bestimme die groessen der einzelnen Kacheln in x-Richtung
    for(i=0; i<tileCountX; i++)
    {
      size=1;
      for(j=0; j<tileCountY; j++)
      {
        name=tileName[i+tileCountX*j];
        if(name) { size=vlMeasureImage(name).getX(); break; }
      }
      sizeX.add(size);
    }
    // das gleiche in y-Richtung
    for(j=0; j<tileCountY; j++)
    {
      size=1;
      for(i=0; i<tileCountX; i++)
      {
        name=tileName[i+tileCountX*j];
        if(name) { size=vlMeasureImage(name).getY(); break; }
      }
      sizeY.add(size);
    }
    // Nun Summen bilden und Groessen der dragLines berechnen
    for(i=0; i<dragLineX.getSize(); i++) 
      { dragLineX[i].size=sizeX[dragLineX[i].pos]; sizeX[dragLineX[i].pos]=0; }
    for(i=0; i<dragLineY.getSize(); i++) 
      { dragLineY[i].size=sizeY[dragLineY[i].pos]; sizeY[dragLineY[i].pos]=0; }
    sumX.add(0); for(i=0; i<tileCountX; i++) sumX.add(sumX[i]+sizeX[i]);
    sumY.add(0); for(i=0; i<tileCountY; i++) sumY.add(sumY[i]+sizeY[i]);
    for(i=0; i<dragLineX.getSize(); i++) sizeX[dragLineX[i].pos]=dragLineX[i].size; 
    for(i=0; i<dragLineY.getSize(); i++) sizeY[dragLineY[i].pos]=dragLineY[i].size; 
  }
  return new VisTileDrawing(*this);
}

    
VisContainerInterface2 *VisTileDrawing::expand(void)
{
  //printf("expanding TileDrawing...\n");
  int i;
  for(i=0; i<container.getSize(); i++)
    if(container[i].interface)
    {
      if(container[i].align.x==VisAlignBuoy)
      {
        expandX(container[i].p1.x, container[i].buoy.x, 
          container[i].interface->alignmentPoint.x);
        expandX(container[i].buoy.x, container[i].p2.x, 
          container[i].interface->preferedSize.x-container[i].interface->alignmentPoint.x);
      }
      else expandX(container[i].p1.x, container[i].p2.x, container[i].interface->preferedSize.x);

      if(container[i].align.y==VisAlignBuoy)
      {
        expandY(container[i].p1.y, container[i].buoy.y, 
          container[i].interface->alignmentPoint.y);
        expandY(container[i].buoy.y, container[i].p2.y, 
          container[i].interface->preferedSize.y-container[i].interface->alignmentPoint.y);
      }
      else expandY(container[i].p1.y, container[i].p2.y, container[i].interface->preferedSize.y);
    }
  containerInterface.preferedSize=
   VisIntVector(expandX(0,tileCountX,0),expandY(0,tileCountY,0));
  containerInterface.alignmentPoint=
   VisIntVector(expandX(0,alignmentPosition.x,0),expandY(0,alignmentPosition.y,0));
  containerInterface.nodePtr=nodePtr;
  return &containerInterface;
}


void VisTileDrawing::draw(int derivpos, TreeContext primIdNodePtr)
{
  //printf("drawing TileDrawing...\n");
  //An Containergroesse anpassen
  if(containerInterface.align.x==VisAlignScale) expandX(0,tileCountX,containerInterface.size.x);
  else containerInterface.xFit();
  if(containerInterface.align.y==VisAlignScale) expandY(0,tileCountY,containerInterface.size.y);
  else containerInterface.yFit();

  //kacheln zeichnen
  VisIntVector tile,rep,nextDragLine,coords;
  coords.x=containerInterface.position.x;
  nextDragLine.x=0;
  for(tile.x=0; tile.x<tileCountX; tile.x++)
  {
    if(nextDragLine.x < dragLineX.getSize() && dragLineX[nextDragLine.x].pos==tile.x)
    { rep.x=dragLineX[nextDragLine.x].repetitions; nextDragLine.x++; }
    else rep.x=1;
    while(rep.x-->0)
    {
      //das ganze in y-Richtung:
      coords.y=containerInterface.position.y;
      nextDragLine.y=0;
      for(tile.y=0; tile.y<tileCountY; tile.y++)
      {
        if(nextDragLine.y < dragLineY.getSize() && dragLineY[nextDragLine.y].pos==tile.y)
        { rep.y=dragLineY[nextDragLine.y].repetitions; nextDragLine.y++; }
        else rep.y=1;
        while(rep.y-->0)
        {
          VLConstCharPtr name=tileName[tile.x+tileCountX*tile.y];
          if(name)
          {
            char tclObjParam[100];
            sprintf(tclObjParam,
                    "-anchor nw -image %s",(VLConstCharPtr )name);

            vlCreateObject(nodePtr, "image", *VLPointArrayPtr(coords), tclObjParam);
          }
          coords.y+=sizeY[tile.y];
        }
      }
      coords.x+=sizeX[tile.x];
    }
  }

  // Einfuegestellen
  for(int i=0; i<insertionLine.getSize(); i++)
  {
    VisTileDrawingInsertionLine *l=&insertionLine[i];
    TreeContext afterNodePtr = listRootNodePtr;
    if(l->after)
    {
      int id=getContainerId(l->after);
      assert(id>=0);
      // assert(container[id].interface); *** muss wieder eingebaut werden
      if(container[id].interface)
      afterNodePtr=container[id].interface->nodePtr;
    } 

    vlCreateInsertionLine(
      l->after ? vlAfter(afterNodePtr) : vlSon(afterNodePtr),
      VLPoint(getPosX(l->p1.x),getPosY(l->p1.y)), 
      VLPoint(getPosX(l->p2.x),getPosY(l->p2.y))
    );
  }

  //Positionen der Unterobjekte berechnen
  for(int i=0; i<container.getSize(); i++)
    if(container[i].interface)
      {
        container[i].interface->position=
          containerInterface.position
          +VisIntVector(expandX(0,container[i].p1.x,0),expandY(0,container[i].p1.y,0));
        container[i].interface->size=
          containerInterface.position
          +VisIntVector(expandX(0,container[i].p2.x,0),expandY(0,container[i].p2.y,0))
          -container[i].interface->position;
        container[i].interface->align=container[i].align;
        container[i].interface->buoy=
          containerInterface.position
          +VisIntVector(expandX(0,container[i].buoy.x,0),expandY(0,container[i].buoy.y,0))
          -container[i].interface->position;
      }
}

void VisTileDrawing::setSubElement(VisContainerInterface2 *elem,VLConstCharPtr name)
{
  int posId=getContainerId(name);
  if(posId<0)
  {
    if (!hasDummyContainer(name)) {
      vlError("VisTileDrawing::setSubElement: Container nicht gefunden " __FILE__);
      printf("Container \"%s\" nicht gefunden! Vorhanden sind:\n",name);
      for(int i=0; i<containerNames.getSize(); i++) 
        printf("%s\n",containerNames[i]);
      //assert(0);
    }
    elem -> position = VisIntVector(-20000,0);
    return;
  }
  assert(posId < container.getSize());
  container[posId].interface=elem;
}



void VisVectorDrawingCoord::mult(int factor)
{
  for(int i=0; i<getSize(); i++) {
    value(i) = factor*value(i);
  }
}


int VisVectorDrawingCoord::min(void)
{
  int m=INT_MAX;
  for(int i=0; i<getSize(); i++) if(value(i)<m) m=value(i);
  return m;
}

int VisVectorDrawingCoord::max(void)
{
  int m=INT_MIN;
  for(int i=0; i<getSize(); i++) if(value(i)>m) m=value(i);
  return m;
}

void VisVectorDrawingCoord::stretch(int left, int right, int delta, int quantum)
{ 
  delta=quantum*((delta+quantum-1)/quantum);
  double factor=1+delta/double(right-left);
  for(int i=0; i<getSize(); i++)
    if(value(i)>left)
      if(value(i)<right) value(i) = left + roundToInt(factor*(value(i)-left));
      else value(i) += delta;
}


void VisVectorDrawingCoord::scale(int left, int right, double factor,int quantum)
{
  int delta=roundToInt((factor-1)*(right-left));
  stretch(left,right,delta,quantum);
}



class VisVectorDrawingTrafo; // forward-Deklaration


int rotRealization[] = {0, 3, 6, 5};

VisVectorDrawing* VisVectorDrawing::rotate(bool flip, int rot) {
// 1. flip: Spiegelung an der y-Achse
// 2. rot: Rotieren im Uhrzeigersinn um 90*rot Grad. Gueltige Werte: 0, 1, 2, 3.


  if(!inited)
  {
    draw();
    if ( (rotRealization[rot]>>2)&1 ^ flip )
      flipX();
    if ( (rotRealization[rot]>>1)&1 ) 
      flipY();
    if ( (rotRealization[rot]>>0)&1 ) 
      flipXY();

    // berechne fuer jedes Container-Intervall das erste 
    // Stretch-Intervall, das dessen Groesse beeinflusst
    int contId2;
    startintId2X.setSize(contPosX.getSize()/2);
    endintId2X.setSize(contPosX.getSize()/2);
    for(contId2=0; contId2<contPosX.getSize(); contId2+=2)
    {
      int contLeft=contPosX[contId2];
      int contRight=contPosX[contId2+1];
      int startintId2=0;
      int endintId2=0;
      while(startintId2<intPosX.getSize() && intPosX[startintId2+1]<=contLeft)
        startintId2+=2;
      startintId2X[contId2/2]=startintId2;
      endintId2=startintId2;
      while(endintId2<intPosX.getSize() && intPosX[endintId2]<contRight)
        endintId2+=2;
      assert(startintId2<endintId2);
      endintId2X[contId2/2]=endintId2-2;
    }
    // das gleiche fuer Y:
    startintId2Y.setSize(contPosY.getSize()/2);
    endintId2Y.setSize(contPosY.getSize()/2);
    for(contId2=0; contId2<contPosY.getSize(); contId2+=2)
    {
      int contLeft=contPosY[contId2];
      int contRight=contPosY[contId2+1];
      int startintId2=0;
      int endintId2=0;
      while(startintId2<intPosY.getSize() && intPosY[startintId2+1]<=contLeft)
        startintId2+=2;
      startintId2Y[contId2/2]=startintId2;
      endintId2=startintId2;
      while(endintId2<intPosY.getSize() && intPosY[endintId2]<contRight)
        endintId2+=2;
      assert(startintId2<endintId2);
      endintId2Y[contId2/2]=endintId2-2;
    }
    if(!boundingBox)
    {
      boundPosX.add(std::min(std::min(contPosX.min(),intPosX.min()),figurePosX.min()));
      boundPosX.add(std::max(std::max(contPosX.max(),intPosX.max()),figurePosX.max()));
      boundPosY.add(std::min(std::min(contPosY.min(),intPosY.min()),figurePosY.min()));
      boundPosY.add(std::max(std::max(contPosY.max(),intPosY.max()),figurePosY.max()));
      boundPosX.add(0);
      boundPosY.add(0);
    }
    inited=true; 
  }
  return this;
}


VisDrawingTrafo *VisVectorDrawing::getDrawingTrafo(void)
{ 
  rotate(0,0);
  VisVectorDrawingTrafo *trafo=new VisVectorDrawingTrafo();
  trafo->init(*this);
  return trafo;
}





void VisVectorDrawingTrafo::init(VisVectorDrawing &d)
{
  drawing=&d; 
  contPosX=d.contPosX; contPosY=d.contPosY; 
  intPosX=d.intPosX; intPosY=d.intPosY;
  figurePosX=d.figurePosX; figurePosY=d.figurePosY;
  boundPosX=d.boundPosX; boundPosY=d.boundPosY;
  insertionLineX=d.insertionLineX; insertionLineY=d.insertionLineY;
  contXPreferedSize.setSize(contPosX.getSize()/2);
  contYPreferedSize.setSize(contPosY.getSize()/2);
  interface.setSize(drawing->contProp.getSize());
  int i;
  for(i=0; i<contXPreferedSize.getSize(); i++) contXPreferedSize[i]=0;
  for(i=0; i<contYPreferedSize.getSize(); i++) contYPreferedSize[i]=0;
  for(i=0; i<interface.getSize(); i++) interface[i]=NULL;
}   


void VisVectorDrawingTrafo::scaleX(int intId2, double factor)
{
  int left=intPosX[intId2];
  int right=intPosX[intId2+1];
  int quantum=drawing->intQuantumX[intId2/2];
  contPosX.scale(left,right,factor,quantum);
  intPosX.scale(left,right,factor,quantum);
  figurePosX.scale(left,right,factor,quantum);
  boundPosX.scale(left,right,factor,quantum);
  insertionLineX.scale(left,right,factor,quantum);
}

void VisVectorDrawingTrafo::scaleY(int intId2, double factor)
{
  int left=intPosY[intId2];
  int right=intPosY[intId2+1];
  int quantum=drawing->intQuantumY[intId2/2];
  contPosY.scale(left,right,factor,quantum);
  intPosY.scale(left,right,factor,quantum);
  figurePosY.scale(left,right,factor,quantum);
  boundPosY.scale(left,right,factor,quantum);
  insertionLineY.scale(left,right,factor,quantum);
}


void VisVectorDrawingTrafo::stretchX(int intId2, int delta)
{
  int left=intPosX[intId2];
  int right=intPosX[intId2+1];
  int quantum=drawing->intQuantumX[intId2/2];
  contPosX.stretch(left,right,delta,quantum);
  intPosX.stretch(left,right,delta,quantum);
  figurePosX.stretch(left,right,delta,quantum);
  boundPosX.stretch(left,right,delta,quantum);
  insertionLineX.stretch(left,right,delta,quantum);
}


void VisVectorDrawingTrafo::stretchY(int intId2, int delta)
{
  int left=intPosY[intId2];
  int right=intPosY[intId2+1];
  int quantum=drawing->intQuantumY[intId2/2];
  contPosY.stretch(left,right,delta,quantum);
  intPosY.stretch(left,right,delta,quantum);
  figurePosY.stretch(left,right,delta,quantum);
  boundPosY.stretch(left,right,delta,quantum);
  insertionLineY.stretch(left,right,delta,quantum);
}

void VisVectorDrawingTrafo::evaluateInterfaces(void)
{
  for(int i=0; i<interface.getSize(); i++)
  {
    VisContainerInterface2 *p=interface[i];    
    if(!p) continue;
    p->align.x=drawing->contProp[i].alignX;
    p->align.y=drawing->contProp[i].alignY;
    if(drawing->contProp[i].alignX==VisAlignBuoy)
    {
      contXPreferedSize[drawing->contProp[i].idX]=p->alignmentPoint.x;
      contXPreferedSize[drawing->contProp[i].idX+1]=p->preferedSize.x-p->alignmentPoint.x;
    }
    else contXPreferedSize[drawing->contProp[i].idX]=p->preferedSize.x;

    if(drawing->contProp[i].alignY==VisAlignBuoy)
    {
      contYPreferedSize[drawing->contProp[i].idY]=p->alignmentPoint.y;
      contYPreferedSize[drawing->contProp[i].idY+1]=p->preferedSize.y-p->alignmentPoint.y;
    }
    else contYPreferedSize[drawing->contProp[i].idY]=p->preferedSize.y;
  }
}


void VisVectorDrawingTrafo::fit(void)
{
  //printf("expanding VectorDrawing...\n");
  int contId;
  // gehe alle Container durch
  // faktoren: Wie viel Prozent ueberdeckt der Container das Intervall?
  for(contId=0; contId<contXPreferedSize.getSize(); contId++)
  {
    int contLeft=contPosX[2*contId];
    int contRight=contPosX[2*contId+1];
    int contSize=contRight-contLeft;
    if(contXPreferedSize[contId]<=contSize) continue;
    int firstintId2=drawing->startintId2X[contId];
    int lastintId2=drawing->endintId2X[contId];

    // nachtraeglich eingefuegte Intervall-Prioritaeten:
    int i2,maxpri=0;
    for(i2=firstintId2; i2<=lastintId2; i2+=2)
      maxpri=std::max(maxpri,drawing->intPriorityX[i2/2]);
    while(drawing->intPriorityX[firstintId2/2]!=maxpri) firstintId2+=2;
    while(drawing->intPriorityX[lastintId2/2]!=maxpri) lastintId2-=2;
    int nonmaxpricount=0;
    for(i2=firstintId2; i2<=lastintId2; i2+=2) 
      if(drawing->intPriorityX[i2/2]!=maxpri) nonmaxpricount++;

    int stickOutLeft=std::max(0, contLeft-intPosX[firstintId2]);
    int stickOutRight=std::max(0, intPosX[lastintId2+1]-contRight);
    if(firstintId2 == lastintId2)
    { // Container schneidet nur ein Intervall
      double factor = 1 - (stickOutLeft+stickOutRight) 
                        / double(intPosX[firstintId2+1]-intPosX[firstintId2]);
      isometricalStretchX(firstintId2,roundToInt((contXPreferedSize[contId]-contSize)/factor));
    }
    else
    { // beruecksichtige alle geschnittenen Intervalle
      double factor1  = 1 - stickOutLeft
                          / double(intPosX[firstintId2+1]-intPosX[firstintId2]);
      double factorN  = 1 - stickOutRight
                          / double(intPosX[lastintId2+1]-intPosX[lastintId2]);
      double stretch  = (contXPreferedSize[contId]-contSize)
                  / (factor1*factor1+factorN*factorN+(lastintId2-firstintId2-2)/2-nonmaxpricount);
      isometricalStretchX(firstintId2,roundToInt(factor1*stretch));
      for(int intId2=firstintId2+2; intId2<lastintId2; intId2+=2)
        if(drawing->intPriorityX[intId2/2]==maxpri)
          isometricalStretchX(intId2,roundToInt(stretch));
      isometricalStretchX(lastintId2,roundToInt(factorN*stretch));
    }
  }
  // fuer Y analog
  for(contId=0; contId<contYPreferedSize.getSize(); contId++)
  {
    int contLeft=contPosY[2*contId];
    int contRight=contPosY[2*contId+1];
    int contSize=contRight-contLeft;
    if(contYPreferedSize[contId]<=contSize) continue;
    int firstintId2=drawing->startintId2Y[contId];
    int lastintId2=drawing->endintId2Y[contId];

    // nachtraeglich eingefuegte Intervall-Prioritaeten:
    int i2,maxpri=0;
    for(i2=firstintId2; i2<=lastintId2; i2+=2)
      maxpri=std::max(maxpri,drawing->intPriorityY[i2/2]);
    while(drawing->intPriorityY[firstintId2/2]!=maxpri) firstintId2+=2;
    while(drawing->intPriorityY[lastintId2/2]!=maxpri) lastintId2-=2;
    int nonmaxpricount=0;
    for(i2=firstintId2; i2<=lastintId2; i2+=2) 
      if(drawing->intPriorityY[i2/2]!=maxpri) nonmaxpricount++;

    int stickOutLeft=std::max(0, contLeft-intPosY[firstintId2]);
    int stickOutRight=std::max(0, intPosY[lastintId2+1]-contRight);
    if(firstintId2 == lastintId2)
    { // Container schneidet nur ein Intervall
      double factor = 1 - (stickOutLeft+stickOutRight) 
                        / double(intPosY[firstintId2+1]-intPosY[firstintId2]);
      isometricalStretchY(firstintId2,roundToInt((contYPreferedSize[contId]-contSize)/factor));
    }
    else
    { // beruecksichtige alle geschnittenen Intervalle
      double factor1  = 1 - stickOutLeft
                          / double(intPosY[firstintId2+1]-intPosY[firstintId2]);
      double factorN  = 1 - stickOutRight
                          / double(intPosY[lastintId2+1]-intPosY[lastintId2]);
      double stretch  = (contYPreferedSize[contId]-contSize)
                  / (factor1*factor1+factorN*factorN+(lastintId2-firstintId2-2)/2-nonmaxpricount);
      isometricalStretchY(firstintId2,roundToInt(factor1*stretch));
      for(int intId2=firstintId2+2; intId2<lastintId2; intId2+=2)
        if(drawing->intPriorityY[intId2/2]==maxpri)
          isometricalStretchY(intId2,roundToInt(stretch));
      isometricalStretchY(lastintId2,roundToInt(factorN*stretch));
    }
  }
}


void VisVectorDrawingTrafo::draw(int derivpos, TreeContext primIdNodePtr)
{
#ifdef DEBUG_COORD
  printf("drawing VectorDrawing...\n");
  printf("Interf: x:%i y:%i\n",containerInterface.position.x,containerInterface.position.y);
#endif

  // Groesse an den Container anpassen (evtl. expandieren):
  if(containerInterface.align.x != VisAlignScale) containerInterface.xFit();
  else
  {
    int maxpri=0; 
    for(int i=0; i<drawing->intPriorityX.getSize(); i++) 
      maxpri=std::max(maxpri,drawing->intPriorityX[i]);
    int currSize=0;
    int intId2;
    // currSize: wie gross ist die Summe alle Stretch-Intervalle (mit max. prioritaet) ?
    for(intId2=0; intId2<intPosX.getSize(); intId2+=2)
      if(drawing->intPriorityX[intId2/2]==maxpri) 
        currSize+=intPosX[intId2+1]-intPosX[intId2];
    // wantedSize: wie gross soll die Summe aller Stretch-Intervalle werden?
    int wantedSize=containerInterface.size.x-((boundPosX[1]-boundPosX[0])-currSize);
    for(intId2=0; intId2<intPosX.getSize(); intId2+=2)
      if(drawing->intPriorityX[intId2/2]==maxpri)
      {
        double factor=wantedSize/double(currSize);
        currSize-=intPosX[intId2+1]-intPosX[intId2];
        scaleX(intId2,factor);
        wantedSize-=intPosX[intId2+1]-intPosX[intId2];
      }
  }
  // fuer y analog:
  if(containerInterface.align.y != VisAlignScale) containerInterface.yFit();
  else
  {
    int maxpri=0; 
    for(int i=0; i<drawing->intPriorityY.getSize(); i++) 
      maxpri=std::max(maxpri,drawing->intPriorityY[i]);
    int currSize=0;
    int intId2;
    // currSize: wie gross ist die Summe alle Stretch-Intervalle (mit max. prioritaet) ?
    for(intId2=0; intId2<intPosY.getSize(); intId2+=2)
      if(drawing->intPriorityY[intId2/2]==maxpri) 
        currSize+=intPosY[intId2+1]-intPosY[intId2];
    // wantedSize: wie gross soll die Summe aller Stretch-Intervalle werden?
    int wantedSize=containerInterface.size.y-((boundPosY[1]-boundPosY[0])-currSize);
    for(intId2=0; intId2<intPosY.getSize(); intId2+=2)
      if(drawing->intPriorityY[intId2/2]==maxpri)
      {
        double factor=wantedSize/double(currSize);
        currSize-=intPosY[intId2+1]-intPosY[intId2];
        scaleY(intId2,factor);
        wantedSize-=intPosY[intId2+1]-intPosY[intId2];
      }
  }

  //Grafische Objekte zeichnen
  int dx=containerInterface.position.x-boundPosX[0];
  int dy=containerInterface.position.y-boundPosY[0];
  int i;
  for(i=0; i<drawing->figureProp.getSize(); i++)
  {
    VisVectorDrawingFigureProp *p=&(drawing->figureProp[i]);
    VLPointArrayPtr coords=new VLPointArray;
    for(int j=0; j<p->points; j++)
      coords->setXY(j,figurePosX[j+p->posId]+dx, figurePosY[j+p->posId]+dy);

    vlCreateObject(nodePtr, p->figure, *coords, p->figureStyle);
  }
  // InsertionLines einfuegen

  for(i=0; i<drawing->insertionLineAfter.getSize(); i++)
  {
    TreeContext afterNodePtr = listRootNodePtr;
    if(drawing->insertionLineAfter[i])
    {
      int id=drawing->getContainerId(drawing->insertionLineAfter[i]);
      assert(id>=0);
      // assert(interface[id]); *** muss wieder eingebaut werden
      if(interface[id])
      afterNodePtr=interface[id]->nodePtr;
    } 
    vlCreateInsertionLine(
      drawing->insertionLineAfter[i] ? vlAfter(afterNodePtr) : vlSon(afterNodePtr),
      VLPoint(insertionLineX[2*i+0]+dx,insertionLineY[2*i+0]+dy),
      VLPoint(insertionLineX[2*i+1]+dx,insertionLineY[2*i+1]+dy)
    );
  }

  // Unterpositionen festlegen
  for(i=0; i<interface.getSize(); i++)
  { 
    VisContainerInterface2 *p=interface[i];
    if(!p) continue;
    VisVectorDrawingContainerProp *c=&(drawing->contProp[i]);
    p->position=VisIntVector(contPosX[c->idX*2]+dx, contPosY[c->idY*2]+dy);
    p->size=VisIntVector(contPosX[c->idX*2+ ((c->alignX==VisAlignBuoy)?3:1)]-contPosX[c->idX*2], 
                         contPosY[c->idY*2+ ((c->alignY==VisAlignBuoy)?3:1)]-contPosY[c->idY*2]);
    p->buoy=VisIntVector(contPosX[c->idX*2+1]-contPosX[c->idX*2], 
                         contPosY[c->idY*2+1]-contPosY[c->idY*2]);
    p->align.x=c->alignX;
    p->align.y=c->alignY;
  }
}


void VisVectorDrawingTrafo::setSubElement (VisContainerInterface2 *elem,VLConstCharPtr name)
{
  int posId=drawing->getContainerId(name);
  if(posId<0)
  {
    if (!drawing->hasDummyContainer(name)) {
      vlError("VisVectorDrawingTrafo::setSubElement: Container nicht gefunden " __FILE__);
      printf("Container \"%s\" nicht gefunden! Vorhanden sind:\n",name);
      for(int i=0; i<drawing->containerNames.getSize(); i++) 
        printf("%s\n",drawing->containerNames[i]);
      //assert(0);
    }
    elem -> position = VisIntVector(-20000,0);
    return;
  }
  assert(posId < interface.getSize());
  interface[posId]=elem;
}


VisContainerInterface2 * VisVectorDrawingTrafo::expand(void)
{
  evaluateInterfaces();
  fit();
  containerInterface.preferedSize=preferedSize();
  containerInterface.alignmentPoint=
    VisIntVector(boundPosX[2]-boundPosX[0],boundPosY[2]-boundPosY[0]);
  containerInterface.nodePtr=nodePtr;
  return &containerInterface;
}



