


#include "labelPositioning.h"
#include "VisBasicCode.h"
// wird fuer VisFloatVector benoetigt


// INTERNAL 

// abscoord <-> linecoord 
// Erste Version: Positionierung eines Punktes

VisFloatVector abscoord2linecoord(VisFloatVector linestart, VisFloatVector lineend, VisFloatVector abspos)
{
   VisFloatVector direction=lineend-linestart;
   VisFloatVector normal=direction.rightnormal();
   VisFloatVector relpos1=abspos-linestart;
   return VisFloatVector( scalprod(relpos1,normal), 
                          scalprod(relpos1,direction)/scalprod(direction,direction));
}


VisFloatVector linecoord2abscoord(VisFloatVector linestart, VisFloatVector lineend, VisFloatVector relpos)
{
   VisFloatVector direction=lineend-linestart;
   VisFloatVector normal=direction.rightnormal();
   return linestart + relpos.x*normal + relpos.y*direction;
}


// abscoord <-> linecoord 
// Zweite Version: Positionierung eines Rechtecks


VisFloatVector getBorderPoint(VisFloatVector size, VisFloatVector direction)
// Liefere Koordinaten eines Randpunktes eines Rechtecks(!) relativ zu dessen Mittelpunkt,
// so dass die "Tangente" an diesem Punkt die Richtung "direction" hat.
// Das besondere: Diese Funktion ist stetig!
// getBorderpoint(direction,...) * direction.rightnormal >= 0 !!!
{
   //Berechne Punkt auf der Ellipse zu Box, an der die Tangente die
   //Richtung direction hat:
   VisFloatVector point;
   double alpha=atan2(-direction.x*size.y, direction.y*size.x);
   point.setXY(cos(alpha)*size.x/2, sin(alpha)*size.y/2);

   // Verschiebe diesen Punkt, so dass die Box (rechteckig!) komplett
   // auf einer Seite liegt:
   VisFloatVector normal=direction.rightnormal();
   double dist=(fabs(normal.x*size.x)+fabs(normal.y*size.y))/2;
   point.move((dist-scalprod(point,normal))*normal);
   return point;
}


VisFloatVector abscoord2linecoord(VisFloatVector linestart, VisFloatVector lineend, 
   VisFloatVector boxsize, VisFloatVector lefttoppos)
{
   //Aus linestart, lineend und boxsize abgeleitette Werte:
   VisFloatVector direction=lineend-linestart;
   VisFloatVector normal=direction.rightnormal();
   VisFloatVector borderpoint=getBorderPoint(boxsize,direction);

   VisFloatVector midpos = lefttoppos+boxsize/2;
   double relpos = scalprod(midpos-linestart,normal) / scalprod(borderpoint,normal);

   VisFloatVector anchor,res;
   if(relpos>1)
   {
     anchor=midpos-borderpoint;
     res=abscoord2linecoord(linestart,lineend,anchor);
     res.x+=1; 
     assert(res.x>1);
   }
   else if(relpos<-1)
   {
     anchor=midpos+borderpoint;
     res=abscoord2linecoord(linestart,lineend,anchor);
     res.x-=1; 
     assert(res.x<-1);
   }
   else
   {
     anchor=midpos-relpos*borderpoint;
     res=abscoord2linecoord(linestart,lineend,anchor);
     res.x=relpos; //res.x war vorher 0!
     assert(res.x<=1 && res.x>=-1);
   }
   return 1000*res;
}


VisFloatVector linecoord2abscoord(VisFloatVector linestart, VisFloatVector lineend, 
   VisFloatVector boxsize, VisFloatVector relpos)
// Ergebnis: Linke obere Ecke der Box
{
   //Aus linestart, lineend und boxsize abgeleitette Werte:
   VisFloatVector direction=lineend-linestart;
   VisFloatVector normal=direction.rightnormal();
   VisFloatVector borderpoint=getBorderPoint(boxsize,direction);

   relpos=relpos/1000;
   VisFloatVector midpos,anchor;
   if(relpos.x>1)
   {
     relpos.x-=1;
     anchor=linecoord2abscoord(linestart,lineend,relpos);
     midpos=anchor+borderpoint; 
   }
   else if(relpos.x<-1)
   {
     relpos.x+=1;
     anchor=linecoord2abscoord(linestart,lineend,relpos);
     midpos=anchor-borderpoint; 
   }
   else
   { 
     double alpha=relpos.x;
     relpos.x=0;
     anchor=linecoord2abscoord(linestart,lineend,relpos);
     midpos=anchor+alpha*borderpoint;
   }
   return midpos-boxsize/2;
}


VLPoint pointFromVector(VisFloatVector v)
{
  return VLPoint((int)v.x, (int)v.y);
}

VisFloatVector vectorFromPoint(VLPoint p)
{
  return VisFloatVector(p.getX(), p.getY());
}


// INTERFACE 



VLPoint abscoord2linecoord(VLPoint linestart, VLPoint lineend, VLPoint boxsize, VLPoint boxpos)
{
  return pointFromVector(abscoord2linecoord(
                           vectorFromPoint(linestart),
                           vectorFromPoint(lineend),
                           vectorFromPoint(boxsize),
                           vectorFromPoint(boxpos)
                           ));
}

VLRegion linecoord2abscoord(VLPoint linestart, VLPoint lineend, VLPoint boxsize, VLPoint relpos)
{
  VLPoint absPos = pointFromVector(linecoord2abscoord(
                           vectorFromPoint(linestart),
                           vectorFromPoint(lineend),
                           vectorFromPoint(boxsize),
                           vectorFromPoint(relpos)
                           ));
  return VLRegion(absPos, boxsize);
}

