
#include "nonOverlappingSolver.h"


NonOverlappingSolver::NonOverlappingSolver(TreeContext aSetSizeAttr, int aMinDistance) {
  setSizeAttr = aSetSizeAttr;
  minDistance = aMinDistance;
}


VLRegionPtr NonOverlappingSolver::addRegion(TreeContext aPosAttr, VLPoint size) {
  NonOverlappingRegion *region = new NonOverlappingRegion();
  region->region = new VLRegion( VLPoint(dynamic_cast<ObservedType*>(aPosAttr.getValue())->getValue()), size );
  region->region->doEnlargeR(VLPoint(minDistance,minDistance));
  region->posAttr = aPosAttr;
  regions.push_back(region);
  return region->region;
}


NonOverlappingRegion *NonOverlappingSolver::findLeftmostOverlapping() {
  NonOverlappingRegion *current = NULL;
  for (NonOverlappingRegionList::iterator p = regions.begin(); p != regions.end(); p++) {
    for (NonOverlappingRegionList::iterator q = regions.begin(); q != regions.end(); q++) {
      if ( (*p) != (*q) && (*p)->region->intersects( *((*q)->region) )) {
         if ( !current 
           || (*p)->region->getPos().getX() < current->region->getPos().getX() 
           || ( (*p)->region->getPos().getX() == current->region->getPos().getX() && (*p)->region->getPos().getY() < current->region->getPos().getY() ) ) {
             current = (*p);
         }
      }
    }
  }
  return current;
}


bool NonOverlappingSolver::findOverlapping(NonOverlappingRegion *&o1, NonOverlappingRegion *&o2) {
  NonOverlappingRegion *first = findLeftmostOverlapping();
  if (!first) return 0;

  for (NonOverlappingRegionList::iterator p = regions.begin(); p != regions.end(); p++) {
    if ( (*p) != first && (*p)->region->intersects(*(first->region)) ) {
        o1 = first;
        o2 = (*p);
        return 1;
    }
  }
  assert(0);
  return 0;
}


void NonOverlappingSolver::updatePersistentValues() {
  for (NonOverlappingRegionList::iterator p = regions.begin(); p != regions.end(); p++) {
    VLList cmd;
    cmd.append("edit::setValue");
    cmd.append((*p)->posAttr);
    cmd.append((*p)->region->getPos());
    cmd.eval();
  }
}


void NonOverlappingSolver::solve() {
  NonOverlappingRegion *reg1 = NULL;
  NonOverlappingRegion *reg2 = NULL;
  NonOverlappingRegion *reg1x = NULL;
  NonOverlappingRegion *reg2x = NULL;
  NonOverlappingRegion *reg1y = NULL;
  NonOverlappingRegion *reg2y = NULL;

  while(findOverlapping(reg1, reg2)) {

    if(reg1->region->getPos().getX() <= reg2->region->getPos().getX()) {
      reg1x = reg1;
      reg2x = reg2;
    } else {
      reg1x = reg2;
      reg2x = reg1;
    }   
    int deltaX = reg1x->region->getPos2().getX() - reg2x->region->getPos().getX();

    if(reg1->region->getPos().getY() <= reg2->region->getPos().getY()) {
      reg1y = reg1;
      reg2y = reg2;
    } else {
      reg1y = reg2;
      reg2y = reg1;
    }   
    int deltaY = reg1y->region->getPos2().getY() - reg2y->region->getPos().getY();

    if (deltaX<deltaY) {
      reg2x->region->doMove(VLPoint(deltaX, 0));
      regions.remove(reg2x);
      regions.push_front(reg2x);
    } else {
      reg2y->region->doMove(VLPoint(0, deltaY));
      regions.remove(reg2y);
      regions.push_front(reg2y);
    }
  }
  updatePersistentValues();
}
