

#include "tcl.h"
#include <string.h>

int setVoidPointerFromString(Tcl_Interp *interp, Tcl_Obj *objPtr);
void updateStringFromVoidPointer(Tcl_Obj *objPtr);
void dupVoidPointer(Tcl_Obj *srcPtr,Tcl_Obj *dupPtr);

Tcl_ObjType voidPointerType = {
  "voidpointer",
  NULL,
  dupVoidPointer,
  updateStringFromVoidPointer,
  setVoidPointerFromString
};

int setVoidPointerFromString(Tcl_Interp *interp, Tcl_Obj *objPtr) {
  void *pointer;
  int length;
  char *str;
  Tcl_FreeInternalRepProc *freeProc;

  /* get string rep - we use the string only to the null char */
  str = Tcl_GetStringFromObj(objPtr,NULL);
  /* test for conversion error */
  if(sscanf(str,"%p",&pointer)!=1) {
    /* return TCL_ERROR after set error result if interp is given */
      if(interp != NULL) {
      char buffer[256];
      sprintf (buffer, "conversion to voidpointer type failed for '%s'", str);
      
      Tcl_SetResult(interp,buffer,TCL_VOLATILE);
      }
      
    return TCL_ERROR;
  }

  /* If old type has a free proc - call it */
  /* typePtr can be NULL if internal rep is invalid */
  if(objPtr->typePtr) {
    freeProc = objPtr->typePtr->freeIntRepProc;
    if(freeProc != NULL) {
      (*freeProc)(objPtr);
    }
  }

  /* set new internal rep and new type */
  objPtr -> internalRep.otherValuePtr = pointer;
  objPtr->typePtr = &voidPointerType;

  return TCL_OK;
}

/* STARTLEN ^= 16 (64bit pointer = 8 byte =  16 hex chars) + 2 (header) + 1 (null char) < 20*/
#define STARTLEN 20

void updateStringFromVoidPointer(Tcl_Obj *objPtr) {
  int nchars;
  int size;
  char *str;

  size = STARTLEN;
  str = Tcl_Alloc(size);

  if (!objPtr->internalRep.otherValuePtr) {
     strcpy(str, "0x0");
     nchars = 3;
  } else {
    while(1) {
      nchars = sprintf(str,"%p",objPtr->internalRep.otherValuePtr);
      if(nchars>-1)
        break;
      size += STARTLEN;
      str = Tcl_Realloc(str,size);
    }
  }

  str = Tcl_Realloc(str,nchars+1);

  objPtr->bytes = str;
  objPtr->length = nchars;
}

void dupVoidPointer(Tcl_Obj *srcPtr,Tcl_Obj *dupPtr) {
  dupPtr->internalRep.otherValuePtr = srcPtr->internalRep.otherValuePtr;
  dupPtr->typePtr = &voidPointerType;
}

Tcl_Obj* TclVP_NewVoidPointerObj(void *pointer) {
  Tcl_Obj *objPtr;

  objPtr = Tcl_NewObj();
  objPtr->typePtr = &voidPointerType;
  objPtr->internalRep.otherValuePtr = pointer;
  Tcl_InvalidateStringRep(objPtr);
  
  return objPtr;
}

void TclVP_SetVoidPointerObj(Tcl_Obj *objPtr, void *pointer) {
  Tcl_FreeInternalRepProc *freeProc;

  if(objPtr->typePtr) {
     /* If old type has a free proc - call it */
    freeProc = objPtr->typePtr->freeIntRepProc;
    if(freeProc != NULL) {
      (*freeProc)(objPtr);
    }
  }
  objPtr->internalRep.otherValuePtr = pointer;
  Tcl_InvalidateStringRep(objPtr);
}

int TclVP_GetVoidPointerFromObj(Tcl_Interp *interp,Tcl_Obj *objPtr,void **pointerPtr) {
  int status;

  if(objPtr->typePtr != &voidPointerType) {
    /*
    fprintf(stderr,"convert type\n");
    */
    status = Tcl_ConvertToType(interp,objPtr,&voidPointerType);
    if(status != TCL_OK)
      return status;
  }
  *pointerPtr = objPtr->internalRep.otherValuePtr;
  return TCL_OK;
}


void TclVP_registerVoidPointerType(void) {
  // Tcl_RegisterObjType(&voidPointerType);
}

