
# ---- namensraeume --------------------------------------------------------------------------------------

set names::isNamespace(Library) 1
set names::isNamespace(WebNode) 1
set names::isNamespace(DataSourceTable) 1
set names::isNamespace(Formular) 1
set names::prefix(StyleSheetBlock) SSB
set names::prefix(GlobalFrame) GF
set names::isNamespace(DataSourceInsert) 1
set names::isNamespace(DataSourceProjection) 1
set names::isNamespace(DataSourceDelete) 1


#main::addInitProc <procName>
# -------------------------------------------------------------------------------------------------------


checkutil::addCheck WebPage {
  set names [c::getList $obj.formalParameter.CHILDREN.name.VALUE]
  set components [c::getList "$obj.frame.CHILDREN.components.CHILDREN"]
  
  foreach {comp} $components {
    if ![equal [c::getClassName $comp] "Formular"] {
      continue
    }
    set formelements [c::getList $comp.elements.CHILDREN]
    foreach {formelem} $formelements {
      lappend names [c::get $formelem.name.VALUE]
    }
  }
  
  foreach el $names { 
    if {![info exists arr($el)]} {
      set arr($el) 1
      continue
    }
    
    # we have found a double entry
    return "$el is not unique within the parameter and form element names!"
  }
  
  return ""
}

checkutil::addCheck WebNodeActualParam.formalParam {
  if [c::tryGet "$obj.INCLUDING Formular" form] {
    set action [c::get $form.action.VALUE]
    return [_assertEqualContexts $action "WebAction" $value "must be of the assoc. Action!"]
  }
  return ""
}

checkutil::addCheck WebNodeActualParam {
  return [_assertActualParamsConsistency $obj]
}

checkutil::addCheck DataSourceActualParam {
  return [_assertActualParamsConsistency $obj]
}

proc _assertActualParamsConsistency {obj} {

  set formalParam [c::get $obj.formalParam.VALUE]
  if [nullptr $formalParam] { return "" }
  set formalParamTypeClassName [common::getTypeClassName $formalParam]
  
  set expression [c::getList $obj.expressions.CHILDREN]
  if [equal [llength $expression] 0] { return "" }
  
  set exprClassName [c::getClassName $expression]
  if [equal $exprClassName "WebNodeVariableExpression"] {
    set content [c::get $expression.content.VALUE]
    if [nullptr $content] { return "" }
    
    set contentClassName [c::getClassName $content]
    set actualParamTypeClassName ""
    if [c::nodeIsSubtypeOf $content WebNodeFormalParam] {
      set actualParamTypeClassName [common::getTypeClassName $content]
    } elseif [c::nodeIsSubtypeOf $content FormularElement] {
      if [equal [c::getClassName $obj] "DataSourceActualParam"] {
        return "FormularElement is not allowed here!"
      }
      set actualParamTypeClassName [common::getTypeClassNamesByFormElement $content]
    }
    
    foreach {entry} $actualParamTypeClassName {
      if [equal $formalParamTypeClassName $entry] {
        return ""
      }
      if {[equal $formalParamTypeClassName "DBLong"] && [equal $entry "DBInteger"]} {
        # integer kann problemlos einer Variable vom Typ long zugewiesen werden!
        return ""
      }
    }
    return "types of '[c::get $formalParam.name.VALUE]'$formalParamTypeClassName and '[c::get $content.name.VALUE]'$actualParamTypeClassName are inconsistent!"
    
  } elseif [equal $exprClassName "WebNodeSessionExpression"] {
    set dbcolumn [c::get $expression.content.VALUE]
    if [nullptr $dbcolumn] { return "" }
    
    set actualParamTypeClassName [common::getTypeClassName $dbcolumn]
    if [equal $formalParamTypeClassName $actualParamTypeClassName] {
      return ""
    }
    if {[equal $formalParamTypeClassName "DBLong"] && [equal $actualParamTypeClassName "DBInteger"]} {
      # integer kann problemlos einer Variable vom Typ long zugewiesen werden!
      return ""
    }
    return "types of '[c::get $formalParam.name.VALUE]'$formalParamTypeClassName and '[c::get $dbcolumn.name.VALUE]'$actualParamTypeClassName are inconsistent!"
    
  } elseif [equal $exprClassName "WebNodeConstantExpression"] {
    set actualParamTypeClassName [common::getPossibleTypeClassNamesOfConstantExpression $expression]
    foreach {entry} $actualParamTypeClassName {
      if [equal $formalParamTypeClassName $entry] {
        return ""
      }
      if {[equal $formalParamTypeClassName "DBLong"] && [equal $entry "DBInteger"]} {
        # ein integer kann problemlos einer Variable vom Typ long zugewiesen werden!
        return ""
      }
      if {[equal $formalParamTypeClassName "DBDouble"] && [equal $entry "DBFloat"]} {
        # ein float kann problemlos einer Variable vom Typ double zugewiesen werden!
        return ""
      }
    }
    return "types of '[c::get $formalParam.name.VALUE]'$formalParamTypeClassName and '[c::get $expression.value.VALUE]'$actualParamTypeClassName are inconsistent!"
  } 
  
  return ""
}

checkutil::addCheck WebNodeVariableExpression.content {
    set error "must reference a field of the assoc. form or a formal parameter of the current web page!"

    if [c::nodeIsSubtypeOf $value FormularElement] {
        if ![c::tryGet "$obj.INCLUDING Formular" result] {
         return $error
        }
         set currentForm [c::get "$obj.INCLUDING Formular"]

        set referedForm [c::get "$value.INCLUDING Formular"]
        if ![equal $currentForm $referedForm] {
            return $error
        }
    }
    if [c::nodeIsSubtypeOf $value WebNodeFormalParam] {
       return [_assertFormalParameterIsFromCurrentWebNode $obj $value]
    }
    return ""
}

proc _assertEqualContexts {context contextName value errmsg} {
  if ![c::tryGet "$value.INCLUDING $contextName" refContext] {
    return $errmsg
  }
  if ![equal $context $refContext] {
    return $errmsg
  }
  return ""
}
 
proc _assertFormalParameterIsFromCurrentWebNode {obj value} {
    set error "must reference a formal parameter of the current webpage!"

    if ![c::tryGet "$obj.INCLUDING WebNode" currentWebnode] {
      return $error
    }
    
    return [_assertEqualContexts $currentWebnode "WebNode" $value $error]
}

checkutil::addCheck Formular.dataSourceId {
  return [_assertFormalParameterIsFromCurrentWebNode $obj $value]
}

checkutil::addCheck OutputFragment.content {
  return [_assertFormalParameterIsFromCurrentWebNode $obj $value]
}

checkutil::addCheck Formular.dataSource {
  set className [c::getClassName $value]
    
  if {![equal $className "DataSourceTable"]} {
    return "must reference a database table!"
  }
  
  return ""
}

checkutil::addCheck Formular {
  
  set dataSource [c::get $obj.dataSource.VALUE]
  set dataSourceId [c::get $obj.dataSourceId.VALUE]
  
  if {![nullptr $dataSource] && [nullptr $dataSourceId]} {
    return "if 'dataSource' is set the 'dataSourceId' must reference a formal parameter!"
  }
  
  if {[nullptr $dataSource] && ![nullptr $dataSourceId]} {
    return "if 'dataSourceId' is set the 'dataSource' must reference a datasource!"
  }
  
  return ""
}

checkutil::addCheck FormularDataSourceSelectOption {
  set dataSource [c::get $obj.dataSource.VALUE]
  if [nullptr $dataSource] { return "dataSource must not be null!" }
  
  set value [c::get $obj.value.VALUE]
  if [nullptr $value] { return "value must not be null!" }
  
  set label [c::get $obj.label.VALUE]
  if [nullptr $label] { return "label must not be null!" }
  
  set valueDataSource [c::get "$value.INCLUDING DataSourceTable"]
  set labelDataSource [c::get "$label.INCLUDING DataSourceTable"]
  
  if ![equal $dataSource $valueDataSource] {
    return "value must be of the selected datasource!"
  }
  if ![equal $dataSource $labelDataSource] {
    return "label must be of the selected datasource!"
  }
  
  return ""
}

checkutil::addCheck FormularSelectElement {
  set selectElemTypes [common::getTypeClassNamesByFormElement $obj]
  set options [c::getList $obj.options.CHILDREN]
  if [equal [llength $options] 0] { return "" }
  
  set optionsClassName [c::getClassName $options]
  if [equal $optionsClassName "FormularConstantSelectOption"] {
    # TODO
    return ""
  } elseif [equal $optionsClassName "FormularDataSourceSelectOption"] {
    set value [c::get $options.value.VALUE]
    if [nullptr $value] { return "value must not be null!" }
    set valueTypeClass [common::getTypeClassName $value]
    
    foreach {elemType} $selectElemTypes {
      if [equal $elemType $valueTypeClass] { return "" }
    }
    return "types of its converter and the value of FormularDataSourceSelectOption are inconsistent!"
  }
  return ""
}


checkutil::addCheck FormularElement.pvalue {
  if [nullptr $value] {
    return ""
  }
  set error "must reference a column of the assoc. database table!"
  set formular [c::get "$obj.INCLUDING Formular"]
  set dataSource [c::get $formular.dataSource.VALUE]
  if ![c::tryGet "$value.INCLUDING DataSourceTable" referedDataSource] {
    return $error
  }
  if ![equal $dataSource $referedDataSource] {
    return $error
  }
  return ""  
}


checkutil::addCheck Fragment.link {
  
  if [nullptr $value] { return "" }
  if [equal [c::getClassName $value] "SubWebZone"] {
    return "webzones are not allowed to be linked!"
  }
  
  return ""
}

checkutil::addCheck DataSourceProjection.sourceDataSource {
   set className [c::getClassName $value]

   if {[equal $className "DataSourceJoin"] || [equal $className "DataSourceTable"]} {
	 return ""
   }
   return "The source of a projection must be a database table or a join datasource!"
}

checkutil::addCheck OrderedDataSource.sourceDataSource {
   set className [c::getClassName "$value"]
   if {[equal $className "DataSourceProjection"] || [equal $className "DataSourceTable"]} {
	 return ""
   }
   return "The source of an ordered datasource must be a database table or a projection filter!"
}

checkutil::addCheck LimitDataSource.sourceDataSource {
  set className [c::getClassName "$value"]
   if {[equal $className "OrderedDataSource"] || [equal $className "DataSourceProjection"] || [equal $className "DataSourceTable"]} {
	 return ""
   }
   return "The source of a limit datasource must be a database table, a projection filter or at least an ordered datasource!"  
}


checkutil::addCheck DataSourceModifyOperation.targetDataSource {
  set className [c::getClassName "$value"]
  if ![equal $className "DataSourceTable"] {
    return "modify operations must point to a database table!"
  }
  return ""
}

checkutil::addCheck DataSourceModifyOperation.name {
  return [checkName $value]
}

checkutil::addCheck DataSource.name {
  return [checkName $value]
}

checkutil::addCheck WebRole.name {
  return [checkName $value]
}

checkutil::addCheck DBColumn.name {
  return [checkName $value]
}

checkutil::addCheck Projection.name {
  return [checkName $value]
}

checkutil::addCheck WebApplication.name {
  if [equal [string match "* *" $value] 1] {
    return "'$value' must not contain any spaces!"
  }
  return ""
}

checkutil::addCheck DataSourceFormalParam.name {
  return [checkName $value]
}

checkutil::addCheck FormularElement.name {
  set chkResult [checkName $value]
  if ![equal $chkResult ""] {
    return $chkResult
  }
  
  if ![string is lower $value] {
    return "'$value' must be in lower case (a-z)!"
  }
  
  return ""
}

checkutil::addCheck WebNode.name {
  set chkResult [checkName $value]
  if ![equal $chkResult ""] {
    return $chkResult
  }
  
  if ![equal [string match "\[A-Z\]*" $value] 1] {
    return "'$value' must start with an upper case letter (A-Z)!"
  }
  
  return ""
}

checkutil::addCheck WebNodeFormalParam.name {
  set chkResult [checkName $value]
  if ![equal $chkResult ""] {
    return $chkResult
  }
  
  if ![equal [string match "\[a-z\]*" $value] 1] {
    return "'$value' must start with a lower case letter (a-z)!"
  }
  
  return ""
}

proc checkName {name} {
  
  if [equal [string match "*-*" $name] 1] {
    return "'$name' must not contain any hyphen!"
  }
  
  if [equal [string match "* *" $name] 1] {
    return "'$name' must not contain any spaces!"
  }
  
  if [common::isJavaKeyword $name] {
    return "'$name' is a java keyword. It is not allowed here! Please choose another identifier!"
  }
  
  return ""
}


checkutil::addCheck ListOutputElement.dataSource {
  set className [c::getClassName $value]
    
  if {[equal $className "DataSourceJoin"]} {
    return "must reference a datasource different from 'DataSourceJoin'!"
  }
  
  return ""
}

checkutil::addCheck ListOutputRefContent.column {
  set errmsg "must reference a column available in the associated datasource!"
  if ![c::tryGet "$obj.INCLUDING ListOutputElement" listoutputElement] {
    return $errmsg
  }
  if [nullptr $value] { return $errmsg }
  # WORKAROUND!
  if ![equal [c::getClassName $value] "DBColumn"] { return "" }
  
  set dataSource [c::get $listoutputElement.dataSource.VALUE]  
  set sourceColumns [getColumns $dataSource true]
  if [equal [lsearch -exact $sourceColumns $value] -1] {
    return $errmsg
  }
  return ""
}

checkutil::addCheck ListOutputProjectionContent.column {
  set errmsg "must reference a projection column available in the associated datasource!"
  if ![c::tryGet "$obj.INCLUDING ListOutputElement" listoutputElement] {
    return $errmsg
  }
  if [nullptr $value] { return $errmsg } 
  set dataSource [c::get $listoutputElement.dataSource.VALUE]
  set sourceColumns [getProjectionColumns $dataSource]
  if [equal [lsearch -exact $sourceColumns $value] -1] {
    return $errmsg
  }
  return ""
}

checkutil::addCheck ListOutputRefContent {
  if ![c::tryGet "$obj.INCLUDING ListOutputElement" listoutputElement] {
    return "must be in subtree of ListOutputElement!"
  }
  
  set dataSource [c::get $listoutputElement.dataSource.VALUE]
  if [isProjectionDataSource $dataSource] {
    return "not allowed on ProjectionFilter or sub datasources! Please use 'Projection Column Content' instead."
  }
  
  return ""
}

checkutil::addCheck ListOutputProjectionContent {
  if ![c::tryGet "$obj.INCLUDING ListOutputElement" listoutputElement] {
    return "must be in subtree of ListOutputElement!"
  }
  
  set dataSource [c::get $listoutputElement.dataSource.VALUE]
  if ![isProjectionDataSource $dataSource] {
    return "only allowed on ProjectionFilter or sub datasources! Please use 'DB Column Content' instead."
  }
  
  return ""
}

checkutil::addCheck ListOutputActualParam {
  set expression [c::getList $obj.expression.CHILDREN]
  if [equal [llength $expression] 0] {
    return ""
  }
  
  set exprClassName [c::getClassName $expression]
  if {[equal $exprClassName "ColumnValueExpression"] || [equal $exprClassName "ProjectionColumnValueExpression"]} {
    return ""
  }
  
  return "only ColumnValueExpression or ProjectionColumnValueExpression are allowed here!"
}

checkutil::addCheck ListOutputActualParam.formalParam {
  set errmsg "must reference a formal parameter of the corresponding webpage or action!"
  if ![c::tryGet "$obj.INCLUDING ActionConnection" actionConnection] {
    return $errmsg
  }
  set actionWebNode [c::get $actionConnection.action.VALUE]
  if ![c::tryGet "$value.INCLUDING WebNode" referedWebNode] {
    return $errmsg
  }
  if ![equal $actionWebNode $referedWebNode] {
    return $errmsg
  }
  return ""
}

# returns DataSourceTable or DataSourceProjection
proc getDataSourceSource {dataSource} {
  set className [c::getClassName $dataSource]
    
  if {[equal $className "OrderedDataSource"] || [equal $className "LimitDataSource"]} {
    return [getDataSourceSource [c::get $dataSource.sourceDataSource.VALUE]]
  }
  
  return $dataSource
}

proc getDataSourceSourceClassName {dataSource} {
  return [c::getClassName [getDataSourceSource $dataSource]]
}

proc checkJoinDataSource {dataSource} {
  if [equal [c::getClassName $dataSource] "DataSourceTable"] {
    return ""
  }
 
  return "A the moment a join may only consist of two database tables! Please fix your specification!"
}

proc checkJoinDataSourceColumn {col dataSource} {
  if [nullptr $col] {
    return ""
  }
  
  set columnName [c::get "$col.name.VALUE"]
  set columns [c::getList "$dataSource.columns.CHILDREN.name.VALUE"]
  set index [lsearch -exact $columns $columnName]
  if [equal $index -1] {
    return "The columns for joining two tables must reference the corresponding database tables!"
  }
  
  return ""
}
checkutil::addCheck DataSourceJoin.dataSource1 {
  return [checkJoinDataSource $value]
}

checkutil::addCheck DataSourceJoin.dataSource2 {
  return [checkJoinDataSource $value]
}

checkutil::addCheck DataSourceTable {
  set columnObjs [c::getList $obj.columns.CHILDREN]
  
  set primKeyCount 0
  foreach {columnObj} $columnObjs {
    if [c::get $columnObj.primaryKey.VALUE] {
      set primKeyCount [expr $primKeyCount +1]
    }
  }
  
  if [equal $primKeyCount 0] {
    return "one column of db table '[c::get $obj.name.VALUE]' must be marked as primary key!"
  }
  
  if {$primKeyCount > 1} {
    return "only one column of db table '[c::get $obj.name.VALUE]' may be marked as primary key!"
  }
  
  return ""
}

checkutil::addCheck DataSourceJoin {
  
  for {set i 1} {$i<3} {incr i} {
    set check [checkJoinDataSourceColumn [c::get "$obj.columnRef$i.VALUE"] [c::get "$obj.dataSource$i.VALUE"]]
    if ![equal $check ""] {
      return $check
    }
  }
  return ""
}


checkutil::addCheck DataSourceProjection {

  set sourceColumns [getColumns $obj]  
  set condition [c::getList "$obj.condition.CHILDREN"]
  
  if {[nullptr $condition] || [equal $condition ""]} {
    return ""
  }
 
  set conditionColumns [getColumnsFromCondition $condition]
  foreach {listEntry} $conditionColumns {
   set index [lsearch -exact $sourceColumns $listEntry]
   if [equal $index -1] {
     return "referenced columns in the condition expression must be available in the corresponding datasource!"
    }
  }
  return ""
}

checkutil::addCheck OrderColumn.orderColumn {
  if [nullptr $value] { return "must not be null!" }
    
  set errmsg "The columns to order by must be of the corresponding database table(s)!"
  if ![c::tryGet "$obj.INCLUDING OrderedDataSource" orderedDS] {
    return $errmsg
  }

  # WORKAROUND! Muss ein Bug in Devil sein.
  if ![equal [c::getClassName $value] "DBColumn"] {
    return ""
  }
  
  set sourceColumns [getColumns $orderedDS]
  if [equal [lsearch -exact $sourceColumns $value] -1] {
    return $errmsg
  }

  return ""
}

checkutil::addCheck Projection.column {
  if [nullptr $value] { return "must not be null!" }
  
  set errmsg "The projection columns must be of the corresponding database table(s)!"
  if ![c::tryGet "$obj.INCLUDING DataSourceProjection" proDS] {
    return $errmsg
  }

  # WORKAROUND! Muss ein Bug in Devil sein.
  if ![equal [c::getClassName $value] "DBColumn"] {
    return ""
  }
  
  set sourceColumns [getColumns $proDS]
  if [equal [lsearch -exact $sourceColumns $value] -1] {
    return $errmsg
  }
  
  return "" 
}

checkutil::addCheck ParameterValueExpression.parameter {
    
  if [c::tryGet "$obj.INCLUDING DataSourceModifyOperation" context] {
    set errmsg "must reference a parameter of the corresponding ModifyOperation!"
    return [_assertEqualContexts $context "DataSourceModifyOperation" $value $errmsg]
    
  } elseif [c::tryGet "$obj.INCLUDING DataSourceProjection" context] {
    set errmsg "must reference a parameter of the corresponding ProjectionFilter!"
    return [_assertEqualContexts $context "DataSourceProjection" $value $errmsg]
  } elseif [c::tryGet "$obj.INCLUDING LimitDataSource" context] {
    set errmsg "must reference a parameter of this limitdatasource or the superior projection datasource!"
    set params [_getFormalParametersOfDataSource $context]
    if [equal [lsearch -exact $params $value] -1] {
      return $errmsg
    }
  }
  
  return ""
}

proc _getFormalParametersOfDataSource {dataSource} {
  set className [c::getClassName $dataSource]
  set parameterList ""
  
  if {[equal $className "LimitDataSource"] || [equal $className "DataSourceProjection"]} {
    set parameterList [concat $parameterList [c::getList "$dataSource.formalParameters.CHILDREN"]]
    
    set parentParams [_getFormalParametersOfDataSource [c::get "$dataSource.sourceDataSource.VALUE"]]
    if ![equal $parentParams ""] {
      set parameterList [concat $parameterList $parentParams]
    }
  }
  
  if [equal $className "OrderedDataSource"] {
    return [_getFormalParametersOfDataSource [c::get "$dataSource.sourceDataSource.VALUE"]]
  }
  return $parameterList
}

checkutil::addCheck ColumnValueExpression.column {
  if [nullptr $value] { return "must not be null!" }
  
  if [c::tryGet "$obj.INCLUDING DataSourceModifyOperation" context] {
    set errmsg "must reference a column of the target database table!"
    set targetDataSource [c::get $context.targetDataSource.VALUE]
    if [nullptr $targetDataSource] { return "" }
    return [_assertEqualContexts $targetDataSource "DataSourceTable" $value $errmsg]
    
  } elseif [c::tryGet "$obj.INCLUDING ListOutputElement" context] {
    set dataSource [c::get $context.dataSource.VALUE]
    if [nullptr $dataSource] {
      return ""
    }
    set columns [getColumns $dataSource]
    if [equal [lsearch -exact $columns $value] -1] {
      return "must reference a column of the corresponding datasource!"
    }
  } elseif [c::tryGet "$obj.INCLUDING DataSourceProjection" context] {
    set columns [getColumns $context]
    if [equal [lsearch -exact $columns $value] -1] {
      return "must reference a column of this datasource!"
    }
  }
  
  return ""
}
  
checkutil::addCheck ProjectionColumnValueExpression.column {
  if [nullptr $value] { return "must not be null!" }
  
  if [c::tryGet "$obj.INCLUDING ListOutputElement" context] {
    set dataSource [c::get $context.dataSource.VALUE]
    if [nullptr $dataSource] {
      return ""
    }
    set columns [getProjectionColumns $dataSource]
    if [equal [lsearch -exact $columns $value] -1] {
      return "must reference a projection column of the corresponding datasource!"
    }
  }
  
  return ""
}

checkutil::addCheck ColumnValueExpression {
  if [c::tryGet "$obj.INCLUDING ListOutputElement" context] {
    set dataSource [c::get $context.dataSource.VALUE]
    if [nullptr $dataSource] {
      return ""
    }    
    if [isProjectionDataSource $dataSource] {
      return "not allowed on ProjectionFilter or sub datasources! Please use 'Projection Column Expression' instead."
    }
  }
  
  return "" 
}

checkutil::addCheck ProjectionColumnValueExpression {
  if [c::tryGet "$obj.INCLUDING ListOutputElement" context] {
    set dataSource [c::get $context.dataSource.VALUE]
    if [nullptr $dataSource] {
      return ""
    }    
    if ![isProjectionDataSource $dataSource] {
      return "only allowed on ProjectionFilter or sub datasources! Please use 'Column Expression' instead."
    }
  }
  
  return "" 
}
checkutil::addCheck DataSourceValue {
  set expression [c::getList $obj.expression.CHILDREN]
  if [equal [llength $expression] 0] { return "required sub-element" }
  
  set className [c::getClassName $expression]
  if {![equal $className "ParameterValueExpression"] && ![equal $className "ConstantExpression"] && ![equal $className "PredefinedExpression"]} {
    return "may only contain a ParameterValueExpression, ConstantExpression or PredefinedExpression!"
  }
  
  #Typchecks
  set column [c::get $obj.column.VALUE]
  if [nullptr $column] { return "must not be null" }
  set formalParamTypeClassName [common::getTypeClassName $column]
  set actualParamTypeClassName ""
  if [equal $className "ParameterValueExpression"] {
    set formalParam [c::get $expression.parameter.VALUE]
    if [nullptr $formalParam] { return "must not be null" }
    set actualParamTypeClassName [common::getTypeClassName $formalParam]
    
  } elseif [equal $className "ConstantExpression"] {
    set actualParamTypeClassName [common::getPossibleTypeClassNamesOfConstantExpression $expression]
  } elseif [equal $className "PredefinedExpression"] {
    set value [c::get $expression.value.VALUE]
    if [string equal -length 12 $value "current_date"] {
      set actualParamTypeClassName "DBDate"
    } elseif [string equal -length 17 $value "current_timestamp"] {
      set actualParamTypeClassName "DBTimeStamp"
    } {
      set actualParamTypeClassName [common::getAllTypes]
    }
  }
  
  foreach {entry} $actualParamTypeClassName {
    if [equal $formalParamTypeClassName $entry] {
      return ""
    }
    if {[equal $formalParamTypeClassName "DBLong"] && [equal $entry "DBInteger"]} {
      return ""
    }
    if {[equal $formalParamTypeClassName "DBDouble"] && [equal $entry "DBFloat"]} {
      # ein float kann problemlos einer Variable vom Typ double zugewiesen werden!
      return ""
    }
  }
  return "types are inconsistent! $formalParamTypeClassName and $actualParamTypeClassName"
  
}

checkutil::addCheck DataSourceValue.column {
  if [nullptr $value] {
    return "reference of the datasource value column must not be empty!"
  }
 
  set targetDataSource [c::get "$obj.INCLUDING DataSourceModifyOperation.targetDataSource.VALUE"]
  
  if ![nullptr $targetDataSource] {
   set columns [c::getList "$targetDataSource.columns.CHILDREN"]
   set index [lsearch -exact $columns $value]
   if [equal $index -1] {
    return "column must reference the corresponding one in '[c::get $targetDataSource.name.VALUE]'"
   }
  }
  return ""
}

proc getColumnsFromCondition {objId} {
  set className [c::getClassName "$objId"]
  if [equal $className "BinaryExpression"] {

    set leftOperand [c::getList "$objId.leftOperand.CHILDREN"]
    set rightOperand [c::getList "$objId.rightOperand.CHILDREN"]
    if ![nullptr leftOperand] {
      if ![nullptr rightOperand] {
       return [concat [getColumnsFromCondition [lindex $leftOperand 0]] [getColumnsFromCondition [lindex $rightOperand 0]]]
     }
    }
  }
  if [equal $className "ColumnValueExpression"] {
   return [c::get "$objId.column.VALUE"]
  }
  return ""
}

checkutil::addCheck LimitDataSource {
 
 set formalParams [getFormalParams $obj]
 
 set firstRow [c::getList "$obj.firstRow.CHILDREN"]
 set maxRows [c::getList "$obj.maxRows.CHILDREN"]


 if ![nullptr $firstRow] {
   if ![nullptr $maxRows] {
    set first [lindex $firstRow 0]
    set max [lindex $maxRows 0]
    set params [concat [getParameterNamesFromRow $first ] " " [getParameterNamesFromRow $max]]
  
    foreach {listEntry} $params {
      set index [lsearch -exact $formalParams $listEntry]
      if [equal $index -1] {
        return "Expressions for first row/max rows may only reference its own formal parameters!"
      }
    }
   }
 }
 return ""
}

proc getFormalParams {dataSource} {
  
  if [nullptr $dataSource] {
   return ""
  }
  
  set className [c::getClassName $dataSource]
  
  if {[equal $className "OrderedDataSource"]} {
    return [getFormalParams [c::get "$dataSource.sourceDataSource.VALUE"]]
  }
  
  if {[equal $className "DataSourceProjection"] || [equal $className "LimitDataSource"]} {
    return [c::getList "$dataSource.formalParameters.CHILDREN.name.VALUE"]
  }
  
  return ""
}

proc getParameterNamesFromRow {objId} {
  set className [c::getClassName "$objId"]

  if [equal $className "BinaryExpression"] {

    set leftOperand [c::getList "$objId.leftOperand.CHILDREN"]
    set rightOperand [c::getList "$objId.rightOperand.CHILDREN"]

    if ![nullptr $leftOperand] {
      if ![nullptr $rightOperand] {
        return [concat [getParameterNamesFromRow [lindex $leftOperand 0]] " " [getParameterNamesFromRow [lindex $rightOperand 0]]]

     }
    }
  }
  if [equal $className "ParameterValueExpression"] {

    return [c::get "$objId.parameter.VALUE.name.VALUE"]

  }
  return ""
}


#deprecated
proc getColumnNames {dataSource {untilProjectionOnly false}} {
  set columns [getColumns $dataSource $untilProjectionOnly]
  set columnNames ""
  foreach {column} $columns {
    if [equal [c::getClassName $column] "DBColumn"] {
      set columnNames [concat $columnNames [c::get $column.name.VALUE]]
    } {
      set columnNames [concat $columnNames $column]
    }
  }
  return $columnNames
}

proc getColumns {objId {untilProjectionOnly false}} {
 
 if [nullptr $objId] {
   return ""
 }
 
 set className [c::getClassName $objId]
 set columns ""
 
 if [equal $className "DataSourceJoin"] {
   return [concat [getColumns [c::get "$objId.dataSource1.VALUE"]] [getColumns [c::get "$objId.dataSource2.VALUE"]]]
 }
     
 if {[equal $className "OrderedDataSource"] || [equal $className "LimitDataSource"]} {
   return [getColumns [c::get "$objId.sourceDataSource.VALUE"] $untilProjectionOnly]
 }
 
 if [equal $className "DataSourceProjection"] {
   set columnList [c::getList "$objId.projectionColumns.CHILDREN"]
   foreach {projColumn} $columnList {
     if [equal [c::getClassName $projColumn] "ComputationProjection"] {
       set columns [concat $columns [c::get $projColumn.name.VALUE]]
     } elseif {$untilProjectionOnly} {
       set columns [concat $columns [c::get $projColumn.column.VALUE]]
     }
   }
   
   if {$untilProjectionOnly} {
       return $columns
   }
   set columns [concat $columns [getColumns [c::get "$objId.sourceDataSource.VALUE"]]]
 }
 
 if [ equal $className "DataSourceTable"] {
  set columnList [c::getList "$objId.columns.CHILDREN"]
  foreach {listEntry} $columnList {
    set columns [concat $columns $listEntry]
  }
 }
  return $columns
}

proc getProjectionColumns {objId} {
 
 if [nullptr $objId] {
   return ""
 }
 
 set className [c::getClassName $objId]
 
 if {[equal $className "DataSourceTable"] || [equal $className "DataSourceJoin"]} {
   return ""
 }
     
 if {[equal $className "OrderedDataSource"] || [equal $className "LimitDataSource"]} {
   return [getProjectionColumns [c::get "$objId.sourceDataSource.VALUE"]]
 }
 
 if [equal $className "DataSourceProjection"] {
   return [c::getList "$objId.projectionColumns.CHILDREN"]
 }
 
 return ""
}

### neu
proc getFormalParamName {objId} {
 return [c::getList $objId.actualParams.CHILDREN.formalParam.VALUE.name.VALUE]
}


# --------------------------------------------------------------------------------------------


set body  	"font-family Arial \
             font-style italic \
             background-color #E0E0E0 \
             font-size 12 \
             color blue"


set h    	"font-family Arial \
             font-style italic \
             font-size 12 \
             color blue \
             background-color #E0E0E0"


set p       "font-family Arial \
             font-style italic \
             font-size 12 \
             color blue \
             background-color #E0E0E0"


set table  "font-family Arial \
		    font-style italic \
			font-size 12 \
			color blue \
			background-color #E0E0E0 \
			margin-top 1 \
			margin-bottom 1 \
			margin-left 1 \
			margin-right 1 \
	  		border-width 1 \
	   		border-style solid \
			border-spacing 5"


set td 		"vertical-align top \
  		 	 background-color #E0E0E0"


set tr 		"background-color #E0E0E0"


set div 	"margin-top 1 \
			 margin-bottom 1 \
			 margin-left 1 \
			 margin-right 1 \
             border-width 1 \
             border-style solid \
             border-color red \
      		 background-color #E0E0E0 "

 set li  "margin 20"
 
proc synchook::StyleSheetPair {obj} {
    set attribute [c::get "$obj.attribute.VALUE"]
    set category [c::get "$obj.category.VALUE"]
    set valueCategory [c::get "$obj.valueCategory.VALUE"]
    set valueObject [c::get "$obj.value"]
    set file [c::getRoot $obj]
    set valueObjectList [c::getList $obj.value.CHILDREN]

    if {$category == "Custom"} {
      set valueClass [getValueCategoryClass $valueCategory]
    } {
      set valueClass [getValueClass $attribute]
    }

    if { [empty $valueObjectList] } {
      edit::insert [c::icBelow $valueObject] [edit::createObject $valueClass $file]
    } elseif {[c::getClassName $valueObjectList] != "$valueClass"} {
    	edit::delete $valueObjectList
    	edit::insert [c::icBelow $valueObject] [edit::createObject $valueClass $file]
    } else {
    #"Do nothing"
    }
 }

## neu
proc synchook::DataBaseOperation {obj} {
 set filter [c::get $obj.filter.VALUE]
 if ![nullptr $filter] {
   set defList [c::get $filter.formalParams]
   set repList [c::get $obj.params]

   syncutil::createMissing $defList $repList formalParam createMissingDataSourceActualParam
   syncutil::deleteSuperfluous $defList $repList formalParam
   syncutil::order $defList $repList formalParam
 }
}

proc synchook::ActionConnection {obj} {
  set action [c::get $obj.action.VALUE]
  if [nullptr $action] {
    return
  }
  set defList [c::get $action.formalParameter]
  set repList [c::get $obj.actualParams]

  syncutil::createMissing $defList $repList formalParam createMissingListOutputActualParam
  syncutil::deleteSuperfluous $defList $repList formalParam
  syncutil::order $defList $repList formalParam
}

proc synchook::DataSourceInsert {obj} {
  set filter [c::get $obj.targetDataSource.VALUE]
  if [nullptr $filter] {
    return
  }
  
  if 1 return
  # TODO Parameter koppeln, aber so, dass Primrschlssel und Spalten 
  # vom Typ (DBHasOne, DBHasMany) exkludiert werden
  
  set defList [c::get $filter.columns]
  set repList [c::get $obj.values]
  
  syncutil::createMissing $defList $repList column createMissingDataSourceValue
  syncutil::deleteSuperfluous $defList $repList column
  syncutil::order $defList $repList column
}

proc synchook::ListOutputElement {obj} {
  set dataSource [c::get $obj.dataSource.VALUE]
  if [nullptr $dataSource] {
    return
  }
  
  set defList [getDataSourceDefList $dataSource]
  if [equal $defList ""] {
    return
  }
  
  set repList [c::get $obj.dataSourceActualParams]

  syncutil::createMissing0 $defList $repList formalParam createMissingDataSourceActualParam
  syncutil::deleteSuperfluous0 $defList $repList formalParam
  syncutil::order0 $defList $repList formalParam
}

proc getDataSourceDefList {dataSource} {
  set className [c::getClassName $dataSource]
  
  if [equal $className "LimitDataSource"] {
    return [concat [c::getList $dataSource.formalParameters.CHILDREN] [getDataSourceDefList [c::get $dataSource.sourceDataSource.VALUE]]]
  }
  
  if {[equal $className "OrderedDataSource"]} {
    return [getDataSourceDefList [c::get $dataSource.sourceDataSource.VALUE]]
  }
  
  if {[equal $className "DataSourceProjection"]} {
    return [c::getList $dataSource.formalParameters.CHILDREN]
  }
  
  return ""
}

#----------

proc synchook::Table {obj} {
    set rowList    [c::get $obj.rows]
    set columnList [c::get $obj.columns]
    set file [c::getRoot $obj]

    ### 1. There must be at least one row
    if { [empty [c::getList $obj.rows.CHILDREN]] } {
        edit::insert [c::icBelow $rowList] [edit::createObject TableRow $file]
    }

    ### 2. There must be at least one column
    if { [empty [c::getList $obj.columns.CHILDREN]] } {
        edit::insert [c::icBelow $columnList] [edit::createObject TableColumn $file]
    }
}


proc synchook::TableRow {obj} {
    ### Enforce correct structure of table

    set defList [c::get "$obj.INCLUDING Table.columns"]
    set repList [c::get "$obj.cells"]

    syncutil::createMissing $defList $repList columnRef createMissingTableCell
    syncutil::deleteSuperfluous $defList $repList columnRef
    syncutil::order $defList $repList columnRef
}


proc createMissingTableCell {formalParameter} {
    set file [c::getRoot $formalParameter]
    return [edit::createObject TableCell $file]
}

proc createMissingDataSourceActualParam {formalParameter} {
    set file [c::getRoot $formalParameter]
    return [edit::createObject DataSourceActualParam $file]
}

proc createMissingListOutputActualParam {formalParameter} {
    set file [c::getRoot $formalParameter]
    return [edit::createObject ListOutputActualParam $file]
}

proc createMissingDataSourceValue {formalParameter} {
    set file [c::getRoot $formalParameter]
    return [edit::createObject DataSourceValue $file]
}

proc edithook::create_Table {obj} {
    ### Initialize Table: Create a 1x2 structure

    set rowList    [c::get $obj.rows]
    set columnList [c::get $obj.columns]
    set file [c::getRoot $obj]

    edit::insert [c::icBelow $rowList] [edit::createObject TableRow $file]
    edit::insert [c::icBelow $columnList] [edit::createObject TableColumn $file]
    edit::insert [c::icBelow $columnList] [edit::createObject TableColumn $file]
}


proc edithook::create_FormularRadioGroupElement {obj} {
    set destination    [c::get $obj.elements]
    set file [c::getRoot $obj]

     edit::insert [c::icBelow $destination] [edit::createObject FormularRadioElement $file]
     edit::insert [c::icBelow $destination] [edit::createObject FormularRadioElement $file]
}

proc edithook::create_DataSourceTable {obj} {

    set columns [c::get $obj.columns]
    set file [c::getRoot $obj]

    set dbColumn1 [edit::createObject DBColumn $file]
    edit::insert [c::icBelow $columns] $dbColumn1
    set dbColumn2 [edit::createObject DBColumn $file]
    edit::insert [c::icBelow $columns] $dbColumn2
    
    edit::insert [c::icBelow [c::get $dbColumn1.type]] [edit::createObject DBInteger $file]
    
    change::setAttr $dbColumn1 name "[c::get $obj.name.VALUE]Id"
    change::setAttr $dbColumn1 primaryKey 1
}

proc edithook::create_DataSourceJoin {obj} {

    set database [c::get "$obj.INCLUDING DataBase"]
    set dataSources [c::get $database.dataSources]
    set file [c::getRoot $obj]

    set dsProjection [edit::createObject DataSourceProjection $file]
    edit::insert [c::icBelow $dataSources] $dsProjection
    change::setAttr $dsProjection sourceDataSource $obj
}

rename edit::setValue oldSetValue

proc edit::setValue {obj value} {

    oldSetValue $obj $value

    if [equal [c::getContext $obj] HtmlStyleSheetBlock.name] {
        set file [c::getRoot $obj]
        set pairs  [c::get $obj.PARENT.pairs]

        set pairsList [c::getList $obj.PARENT.pairs.CHILDREN]
        foreach {listEntry} $pairsList {
            edit::delete $listEntry
        }

		if {$value == "body"} {
            foreach {attr value} $::body {
            	### create pair and add to list
            	addPair $obj $attr $value
            }
		} elseif {$value == "p"} {
			foreach {attr value} $::p {
             	### create pair and add to list
            	addPair $obj $attr $value
 			}
		} elseif {$value == "h1" || $value == "h2" || $value == "h3" || $value == "h4" || $value == "h5" || $value == "h6"} {
			foreach {attr value} $::h {
            	### create pair and add to list
            	addPair $obj $attr $value
 			}
		} elseif {$value == "table"} {
			foreach {attr value} $::table {
             	### create pair and add to list
            	addPair $obj $attr $value
 			}
		} elseif {$value == "td"} {
			foreach {attr value} $::td {
            	### create pair and add to list
            	addPair $obj $attr $value
 			}
		} elseif {$value == "tr"} {
			foreach {attr value} $::tr {
             	### create pair and add to list
            	addPair $obj $attr $value
 			}
		} elseif {$value == "div"} {
			foreach {attr value} $::div {
            	### create pair and add to list
            	addPair $obj $attr $value
 			}
		}
    }
}

proc addPair {obj attr value} {
	# adds a new StyleSheetPair to the StyleSheetBlock

    set file [c::getRoot $obj]
    set pairs  [c::get $obj.PARENT.pairs]

    set pair [edit::createObject StyleSheetPair $file]
    edit::insert [c::icBelow $pairs] $pair

    ### set attribute
    set pair_attr [c::get $pair.attribute]
    oldSetValue $pair_attr $attr

    synchook::StyleSheetPair $pair
	setPairValue $pair $value
}


proc setPairValue {obj value} {
	# sets the value attr. of StyleSheetPair objects

    set valueObject [c::getList "$obj.value.CHILDREN"]

    if {[equal [c::getContext $valueObject] SizeValue]} {
    	# size
        edit::setValue [c::get $valueObject.value] $value
        # measure pt
	    edit::setValue [c::get $valueObject.measure] "pt"
	   	edit::setValue [c::get $obj.category] "none"
	} elseif  {[equal [c::getContext $valueObject] StyleValue]} {
		# style
        edit::setValue [c::get $valueObject.style] $value
	   	edit::setValue [c::get $obj.category] "Font"
	} elseif  {[equal [c::getContext $valueObject] FamilyValue]} {
		# family
        edit::setValue [c::get $valueObject.family] $value
	   	edit::setValue [c::get $obj.category] "Font"
	} elseif  {[equal [c::getContext $valueObject] ColorValue]} {
		# color
		edit::setValue [c::get $valueObject.color] $value
	   	edit::setValue [c::get $obj.category] "Color"
	} elseif  {[equal [c::getContext $valueObject] AlignmentValue]} {
		# alignment
		edit::setValue [c::get $valueObject.alignment] $value
	   	edit::setValue [c::get $obj.category] "none"
	} elseif  {[equal [c::getContext $valueObject] HeadingValue]} {
		# heading
		edit::setValue [c::get $valueObject.heading] $value
	   	edit::setValue [c::get $obj.category] "none"
	} elseif  {[equal [c::getContext $valueObject] ListValue]} {
		# listed
		edit::setValue [c::get $valueObject.listed] $value
	   	edit::setValue [c::get $obj.category] "none"
	}
}

proc synchook::Formular {obj} {
    if [nullptr [c::get {$obj.action.VALUE}]] {
        return
    }
    set defList [c::get {$obj.action.VALUE.formalParameter}]
    set repList [c::get {$obj.actualParams}]

    syncutil::createMissing $defList $repList formalParam createMissingWebnodeActualParam
    syncutil::deleteSuperfluous $defList $repList formalParam
    syncutil::order $defList $repList formalParam
}

proc createMissingWebnodeActualParam {definition} {
        return [edit::createObject WebNodeActualParam [edit::getFile $definition]]
}


proc synchook::ActionHandler {obj} {
    
    #if 1 return
    
    if { [nullptr [c::get {$obj.targetpage.VALUE}]] } return

    set defList [c::get {$obj.targetpage.VALUE.formalParameter}]
    set repList [c::get {$obj.params}]

    syncutil::createMissing $defList $repList formalParam createMissingWebnodeActualParam
    syncutil::deleteSuperfluous $defList $repList formalParam
    syncutil::order $defList $repList formalParam
}

proc createMissingWebnodeActualParam {definition} {
        return [edit::createObject WebNodeActualParam [edit::getFile $definition]]
}
