


proc getModifier {objId} {
  set modifierList [c::getList $objId.modifier.CHILDREN]
  foreach {modifier} $modifierList {
    set modifierClassName [c::getClassName $modifier]
    if [equal $modifierClassName "MD5Modifier"] {
      return "md5"
    } elseif [equal $modifierClassName "JoinModifier"] {
      return "join"
    }
  }
  return ""
}


proc _getWebNodeExpressionValue {objId value} {
  set modifier [getModifier $objId]
  if ![equal $modifier ""] {
    return [format "%s%s%s%s" $modifier "(" $value ")"]
  }
  return $value
}

proc getWebNodeVariableExpressionValue {objId} {
  set ref [c::get $objId.content.VALUE]
  if [nullptr $ref] {
    return "###"
  }
  return [_getWebNodeExpressionValue $objId [c::get $ref.name.VALUE]]
}

proc getWebNodeConstantExpressionValue {objId} {
  return [_getWebNodeExpressionValue $objId [getConstantExpressionValue $objId "\""]]
}

proc getWebNodeSessionExpressionValue {objId} {
  set ref [c::get $objId.content.VALUE]
  if [nullptr $ref] {
    return "null"
  }
  
  set name [c::get $ref.name.VALUE]
  set Ds [string totitle [c::get "$ref.INCLUDING DataSourceTable.name.VALUE"]]
  return [_getWebNodeExpressionValue $objId "SESSION\[$Ds.$name\]"]
}

#-----------------------------------Bilder---------------------

proc getImage {path {defaultImage "img::image"}} {
    variable imageCache
    if ![info exists imageCache($path)] {
        if { ![file exists $path] || ![file isfile $path] || ![equal [file extension $path] ".gif"] } {
            return $defaultImage
        }
        set imageCache($path) [image create photo -file $path]
    }
    return $imageCache($path)
}


proc getImageId {objId {defaultImage "img::image"}} {
    set path1 [file dirname [file::getFilename [file::getFileOfNode $objId]]]
    set path2 [c::get "$objId.INCLUDING WebApplication.imagePath.VALUE"]
    set name [c::get "$objId.imageName.VALUE"]
    set result [file join $path1 $path2 $name]
    return [getImage $result $defaultImage]
}


proc getImagePathEditor {objId} {
    if [file::getUnnamed [file::getFileOfNode $objId]] {
        tk_messageBox -message "Can't determine base location for image search path. You must save this specification first!"
        return "Entry"
    }
    set path [file dirname [file::getFilename [file::getFileOfNode $objId]]]
    return [list Path -filetypes dir -relativeTo $path]
}

proc getImageSelectorEditor {objId} {
    if [file::getUnnamed [file::getFileOfNode $objId]] {
        tk_messageBox -message "Can't compute image search path. You must save this specification first!"
        return "Entry"
    }
    set path1 [file dirname [file::getFilename [file::getFileOfNode $objId]]]
    set path2 [file join $path1 [c::get "$objId.INCLUDING WebApplication.imagePath.VALUE"]]
    if ![file exists $path2] {
        tk_messageBox -message "Invalid image Search path '$path2'. Adjust the 'imagePath' attribute in Root node!"
        return "Entry"
    }
    set list "(empty) {}"
    foreach file [glob -nocomplain "$path2/*.gif"] {
        set name [file tail $file]
        lappend list $name $name
    }
    return [list Listbox -values $list]
}

#-------------------------------------

proc getConverterList {} {
    set list "string \"\" integer IntegerConverter long LongConverter \"big integer\" BigIntegerConverter float FloatConverter double DoubleConverter boolean BooleanConverter timestamp DateTimeConverter \"long number\" IntegerNumberConverter \"double number\" DoubleNumberConverter \"custom pattern\" PatternConverter"
    return $list
}

proc getValidatorList {} {
    set list "\"----\" \"\" \"string length\" StringLengthValidator \"long range\" LongRangeValidator \"double range\" DoubleRangeValidator \"email address\" EmailValidator \"credit card\" CreditCardValidator \"regular expression\" RegExprValidator"
    return $list
}

#-------------------------------Dynamische Bestimmung der Attribute-----------------------------------------

proc getAttributeList {objId} {
    set category [c::get "$objId.INCLUDING StyleSheetPair.category.VALUE"]
    if {$category == "Color"} {
    set list "color color background-color background-color font-color font-color margin-color margin-color border-color border-color"
    } elseif {$category == "Font"} {
    set list "font-style font-style font-size font-size font-family font-family"
    } elseif {$category == "Padding"} {
    set list "width width height height left padding-left top padding-top right padding-right bottom padding-bottom padding padding alignment alignment float float"
    } elseif {$category == "Border"} {
    set list "left border-left-width top border-top-width right border-right-width bottom border-bottom-width width border-width"
    } elseif {$category == "List"} {
    set list "listStyle listStyle"
    } elseif {$category == "Link"} {
    set list "visited visited unvisited unvisited"
    } else {
    set list ""
    }
    return $list
}

proc getValueClass {attribute} {
    if {"$attribute" == "background-color" || "$attribute" == "font-color" || "$attribute" == "margin-color" || "$attribute" == "border-color" || "$attribute" == "visited" || "$attribute" ==                 "unvisited" || "$attribute" == "color" ||  "$attribute" == "border-color"} {
      return ColorValue
    } elseif {"$attribute" == "font-style"} {
      return StyleValue
    } elseif {"$attribute" == "font-size" || "$attribute" == "left" || "$attribute" == "top"  || "$attribute" == "right" ||                 "$attribute" == "bottom" || "$attribute" == "width" || "$attribute" == "margin-left" || "$attribute" == "margin-right" ||      "$attribute" == "margin-top" || "$attribute" == "margin-bottom" || "$attribute" == "border-width" ||  "$attribute" == "border-left-width" || "$attribute" == "border-right-width" || "$attribute" == "border-top-width" || "$attribute" == "border-bottom-width" ||  "$attribute" == "padding-left" || "$attribute" == "padding-right" || "$attribute" == "padding-top" || "$attribute" == "padding-bottom" || "$attribute" == "padding" || "$attribute" == "width" || "$attribute" == "height" || 
          "$attribute" == "border-spacing"} {
      return SizeValue
    } elseif {"$attribute" == "font-family"} {
      return FamilyValue
    } elseif {"$attribute" == "alignment" || "$attribute" == "float"} {
      return AlignmentValue
    } else {
     return ListValue
    }
}

proc getValueCategoryClass {valueCategory} {
    if {"$valueCategory" == "Color"} {
      return ColorValue
    } elseif {"$valueCategory" == "Style"} {
      return StyleValue
    } elseif {"$valueCategory" == "Size"} {
      return SizeValue
    } elseif {"$valueCategory" == "Family"} {
      return FamilyValue
    } elseif {"$valueCategory" == "Alignment"} {
      return AlignmentValue
    } elseif {"$valueCategory" == "Custom"} {
      return CustomValue
    } else {
     return ListValue
    }
}


#------------------------------- Editor PHPScript ----------------------------------------------


namespace eval dcPHPScript {}


proc dcPHPScript::create {name args} {
    dc::create $name
    oop::addClass $name dcPHPScript
    oop::data $name

    frame $name
    dc::bindDestroy $name

    label $name.head -text "<?php"
    pack $name.head -side top -anchor w

    frame $name.content
    text $name.content.text
    CDlg::mkScrollableXY $name.content text
    pack $name.content -side top  -expand 1 -fill both

    label $name.bottom -text "?>"
    pack $name.bottom -side top -anchor w

    bind $name.content.text <FocusOut> [list ::dc::changed $name]

    bindtags $name.content.text [linsert [bindtags $name.content.text] 2 BreakReturn]
    bind BreakReturn <Return> "break"
    bind $name.content.text <Tab> "dcPHPScript::focusNext $name; break"

    eval renew [list $name] $args
}


proc dcPHPScript::focusNext {name} {
    focus [tk_focusNext $name.content.text]
}


proc dcPHPScript::getValue {name} {
    return [string range [$name.content.text get 1.0 end] 0 end-1]
}


proc dcPHPScript::setValue {name value} {
    $name.content.text delete 1.0 end
    $name.content.text insert 1.0 $value
}


proc dcPHPScript::renew {name args} {
    oop::data $name
    set remainder [partialHandleArgs [list -width 30 -height 5 -wrap none -enabled 1]]
    dc::renew $name $remainder
    set data(shouldExpand) 1
    $name.content.text configure -width $arg(width) -height $arg(height) -wrap $arg(wrap) -state [lindex "disabled normal" $arg(enabled)]
}

#------------------------------- Editor JavaCodeEditor ----------------------------------------------


namespace eval dcJavaCodeEditor {}


proc dcJavaCodeEditor::create {name args} {
    dc::create $name
    oop::addClass $name dcJavaCodeEditor
    oop::data $name

    frame $name
    dc::bindDestroy $name

    label $name.head -text "public String submitAction() {"
    pack $name.head -side top -anchor w

    frame $name.content
    text $name.content.text
    CDlg::mkScrollableXY $name.content text
    pack $name.content -side top  -expand 1 -fill both

    label $name.bottom -text "}"
    pack $name.bottom -side top -anchor w

    bind $name.content.text <FocusOut> [list ::dc::changed $name]

    bindtags $name.content.text [linsert [bindtags $name.content.text] 2 BreakReturn]
    bind BreakReturn <Return> "break"
    bind $name.content.text <Tab> "dcJavaCodeEditor::focusNext $name; break"

    eval renew [list $name] $args
}


proc dcJavaCodeEditor::focusNext {name} {
    focus [tk_focusNext $name.content.text]
}


proc dcJavaCodeEditor::getValue {name} {
    return [string range [$name.content.text get 1.0 end] 0 end-1]
}


proc dcJavaCodeEditor::setValue {name value} {
    $name.content.text delete 1.0 end
    $name.content.text insert 1.0 $value
}


proc dcJavaCodeEditor::renew {name args} {
    oop::data $name
    set remainder [partialHandleArgs [list -width 30 -height 5 -wrap none -enabled 1]]
    dc::renew $name $remainder
    set data(shouldExpand) 1
    $name.content.text configure -width $arg(width) -height $arg(height) -wrap $arg(wrap) -state [lindex "disabled normal" $arg(enabled)]
}

#------------------------------- Editor VelocityCodeEditor ----------------------------------------------


namespace eval dcVelocityCodeEditor {}


proc dcVelocityCodeEditor::create {name args} {
    dc::create $name
    oop::addClass $name dcVelocityCodeEditor
    oop::data $name

    frame $name
    dc::bindDestroy $name

    label $name.head -text "## begin velocity template"
    pack $name.head -side top -anchor w

    frame $name.content
    text $name.content.text
    CDlg::mkScrollableXY $name.content text
    pack $name.content -side top  -expand 1 -fill both

    label $name.bottom -text "## end velocity template"
    pack $name.bottom -side top -anchor w

    bind $name.content.text <FocusOut> [list ::dc::changed $name]

    bindtags $name.content.text [linsert [bindtags $name.content.text] 2 BreakReturn]
    bind BreakReturn <Return> "break"
    bind $name.content.text <Tab> "dcVelocityCodeEditor::focusNext $name; break"

    eval renew [list $name] $args
}


proc dcVelocityCodeEditor::focusNext {name} {
    focus [tk_focusNext $name.content.text]
}


proc dcVelocityCodeEditor::getValue {name} {
    return [string range [$name.content.text get 1.0 end] 0 end-1]
}


proc dcVelocityCodeEditor::setValue {name value} {
    $name.content.text delete 1.0 end
    $name.content.text insert 1.0 $value
}


proc dcVelocityCodeEditor::renew {name args} {
    oop::data $name
    set remainder [partialHandleArgs [list -width 30 -height 5 -wrap none -enabled 1]]
    dc::renew $name $remainder
    set data(shouldExpand) 1
    $name.content.text configure -width $arg(width) -height $arg(height) -wrap $arg(wrap) -state [lindex "disabled normal" $arg(enabled)]
}

#---------------------------------------- drop effects --------------------------------------------------

dropEffect::add WebNode NavigationBarElement "dropEffect::createApplication NavigationBarElement target"
dropEffect::add WebNodeFormalParam WebNodeVariableExpression "dropEffect::createApplication WebNodeVariableExpression content"
dropEffect::add FormularElement WebNodeVariableExpression "dropEffect::createApplication WebNodeVariableExpression content"
dropEffect::add DBColumn ColumnValueExpression "dropEffect::createApplication ColumnValueExpression column"
dropEffect::add DataSourceFormalParam ParameterValueExpression "dropEffect::createApplication ParameterValueExpression parameter"
dropEffect::add DBColumn WebNodeSessionExpression "dropEffect::createApplication WebNodeSessionExpression content"
dropEffect::add WebRole WebRoleRef "dropEffect::createApplication WebRoleRef role"
dropEffect::add GlobalFrame FrameRef "dropEffect::createApplication FrameRef frame"



#---------------------------------------- templates -----------------------------------------------------

proc file::chooseTemplate {} {
    set templates [glob -nocomplain [file join $::main::appHome templates *.xml]]
    set values ""
    foreach t $templates {
        lappend values [file rootname [file tail $t]] $t
    }
    set data(template,value) ""
    set data(template,label) ""
    set data(template,editor) [list Listbox -values $values]
    set data(template,check) ""
    set ok [CDlg::stdInputDialog "Choose template" [list template] data]
    if { $ok } {
        return $data(template,value)
    } else {
        return ""
    }
}


proc file::new {} {
    set oid [oop::new]
    set template [chooseTemplate]
    setRoot $oid [xmlio::getNewFile $template]
    change::initUndoList [getRoot $oid]
    registerFile $oid
    return $oid
}


proc xmlio::getNewFile {templateFilePath} {
    if { ![empty $templateFilePath] } {
        set root [load $templateFilePath]
    } else {
        set root [edit::createRoot]
        variable data
        catch {unset data}
        set data(crossref) ""
        set data(file) $root
        loadLibraries
        connect
    }
    return $root    
}

#---------------------------------------Reihenfolge der Attribute-----------------------------------------
 set treedraw::attrOrder(TextFragment) "value link"
 set treedraw::attrOrder(ImageFragment) "imageName link width height text"

#---------------------------------------Code generators-------------------------------------

file::registerProcessor myCodeGen generateJ2EECode "Generate J2EE-webapp"

proc generateJ2EECode {root} {

  set mustHaveCommands "java javac ant"
  set downloadDest(javac) "http://java.sun.com/javase/downloads"
  set downloadDest(java) "http://java.sun.com/javase/downloads"
  set downloadDest(ant) "http://ant.apache.org"
  
  foreach {cmd} $mustHaveCommands {
  
    if {[catch {exec $cmd} results]} {  
      puts "executing $cmd results in $::errorCode"
      if {[lindex $::errorCode 0] eq "CHILDSTATUS"} {
        set status [lindex $::errorCode 2]
        
      } elseif {[lindex $::errorCode 0] eq "NONE"} {
        
      } {
        # Some kind of unexpected failure
        tk_messageBox -title "$cmd not available!" -message "$cmd command not available. Please make sure it is in your search path!\nIf you do not have $cmd installed, you should download and install the most recent version from $downloadDest($cmd)"
        return
      } 
    }
  }
   
  c::startProcessor codeGen $root
}

