
# Modul: abstractView
#
# Schnittstelle fr Sichten.
# Sicht-Implementierungen sollten von dieser Klasse erben.



xn::class AbstractView {

    xn::method init {root -size ?} {
        if { ![info exists size] } {
            set size $cfg::defaultClientWindowSize
        }
        set data(root) $root
        set data(file) [file::getFileOfNode $data(root)]
        set data(client) [::MDIFrame::New_Window "Dummy" $object -width [lindex $size 0] -height [lindex $size 1]]
        set data(toplevel) [::MDIFrame::GetUserpath $data(client)]
        set data(viewType) "Abstract View"
        file::addView $data(file) $object
    }


    xn::method wincmd {cmd args} {
        if { [llength [$object info methods $cmd]]==0 } {
            ::CDlg::Information [i18n::text "Command not supported by current window"]
        } else {
            eval [list $object $cmd] $args
        }
    }


    xn::method getRoot {} {
        return $data(root)
    }


    xn::method getFile {} {
        return $data(file)
    }


    xn::method getClient {} {
        return $data(client)
    }


    xn::method makeVisible {} {
        MDIFrame::MakeVisible $data(client)
    }


    xn::method close {} {
        MDIFrame::CloseClient $data(client)
    }


    xn::method destroy {} {
        file::removeView [$object getFile] $object
    }


    xn::method update {} {
        $object updateTitle
    }

    xn::method updateTitle {} {
        set filename [file::getFilename [$object getFile]]
        if [empty $filename] {
            set filename "Unnamed"
        } else {
            set filename [file tail $filename]
        }
        if [file::getModified [$object getFile]] {
            set modified "*"
        } else {
            set modified ""
        }
        MDIFrame::SetTitle [$object getClient] "$data(viewType) (${modified}$filename)"
    }


    xn::method wincmd_startProcessor {processor} {
        file::startProcessor [$object getFile] $processor
    }


    xn::method wincmd_undo {} {
        file::undo [$object getFile]
    }


    xn::method wincmd_save {} {
        file::interactiveSave [$object getFile]
    }


    xn::method wincmd_saveAs {} {
        file::interactiveSaveAs [$object getFile]
    }


    xn::method wincmd_exportRepresentation {} {
        file::exportRepresentationInteractive [$object getFile]
    }


    xn::method wincmd_fileInfo {} {
        file::showInfoDialog [$object getFile]
    }


    xn::method wincmd_showAbstree {} {
        file::openAbstractView [$object getFile]
    }


    xn::method wincmd_showMainView {} {
        file::openMainView [$object getFile]
    }


    xn::method wincmd_showErrWindow {} {
        file::openErrorView [$object getFile]
    }


    xn::method wincmd_findRegexp {} {
        file::findRegexp [$object getFile]
    }


    xn::method wincmd_closeFile {} {
        file::tryClose [$object getFile]
    }


    xn::method isReusable {node type} {
        return [expr [equal $node $data(root)] && [equal $type $data(viewType)]]
    }


    xn::method beforeDeletion {node} {
        if { [c::isAncestorOf $node $data(root)] } {
            MDIFrame::CloseClient $data(client)
        }
    }
}
