
# Tck/Tk kennt statdardmaessig die folgenden grafischen Primitive (siehe Tcl/Tk-Referenz):
# [arc], [bitmap], [image], [line], [oval], [polygon], [rectangle], [text] und [window]. 
# Die grafische Primitive lassen sich mittels der Funktion [CreateObject] verwenden, um
# visuelle Darstellungen auf den Canvas zu bringen. Die nachfolgend beschriebene
# Schnittstelle dient der Erweiterung dieser Menge von Grundtypen.
#
# Neue grafische Primitive koennen definiert werden, indem 
# eine Tcl-Funktion im Namespace ::canvasitems definiert wird, die zwei
# Listen als Parameter verarbeitet, nmlich [coords] und [params]. Die Funktion sollte
# eine Liste der folgenden Form zurueckliefern:
#
# [itemtype1 coords1 params1 itemtype2 coords2 params2 ....]
#
# Als Anwendungsbeispiel fuer die Schnittstelle enthaelt diese Datei drei
# Definitionen fuer neue grafische Primitive, naemlich:
# <ul>
#   <li>[line] (Redefinition), 
#   <li>[roundedrectangle] und 
#   <li>[solid3DRectangle].
# </ul>
#
# */

namespace eval canvasitems {}



#############################################
#   Erweiterung des "line"-Primitivs        #
#############################################

# Es koennen beliebige Pfeilformen definiert werden
# Koordinatensystem: Die Linie kommt von (0,-unendlich) und endet im Punkt (0,0)
# Dieses System wird defaultmaessig um den Faktor 10 skaliert.
# Datenstruktur: Genau wie der Rueckgabewert von der ::canvasitem - Funktionen
# Eine Pfeilform muss den Praefix "lineend_" besitzen!

set canvasitems::lineend_normal      {line    {-0.5 -1 0 0 0.5 -1} {}}
set canvasitems::lineend_inherits    {polygon {0 0 -0.5 -1 0.5 -1} { -fill white -outline black}}
set canvasitems::lineend_aggregation {polygon {0 0 -0.3 -0.5 0 -1 0.3 -0.5} { -fill white -outline black}}
set canvasitems::lineend_composition {polygon {0 0 -0.3 -0.5 0 -1 0.3 -0.5} { -fill black -outline black}}


proc canvasitems::line2 {coords params} {
  handleParams $params param "-arrowSize1 10 -arrowSize2 10 -arrowShape1 {} -arrowShape2 {} \
    -width x -fill x -stipple x -arrow x -arrowshape x -smooth x -dash x -tags x"

  #Linienende 1
  if { $param(-arrowShape1) == "" } {set arc1 ""} else {
    set last [lrange $coords 0 1]
    set prelast [lrange $coords 2 3]
    set rvec [vec::sub $last $prelast]
    set rvec [vec::normalize $rvec $param(-arrowSize1)]
    set qvec [vec::rotateRight90 $rvec]
    set arc1 [vec::transform [subst "\$canvasitems::lineend_$param(-arrowShape1)"] $rvec $qvec $last]
  }

  #Linienende 2
  if { $param(-arrowShape2) == "" } {set arc2 ""} else {
    set last [lrange $coords [expr [llength $coords]-2] end]
    set prelast [lrange $coords [expr [llength $coords]-4] [expr [llength $coords]-2]]
    set rvec [vec_sub $last $prelast]
    set rvec [vec_normalize $rvec $param(-arrowSize2)]
    set qvec [vec_rotateRight90 $rvec]
    set arc2 [vec_transform [subst "\$canvasitems::lineend_$param(-arrowShape2)"] $rvec $qvec $last]
  }

  unset param(-arrowSize1)
  unset param(-arrowSize2)
  unset param(-arrowShape1)
  unset param(-arrowShape2)
  set params [array get param]
  return [concat [list "line" $coords $params] $arc1 $arc2]
}



#############################################
#             Rounded Rectangle             #
#############################################


proc canvasitems::roundedRectangle {coords params} {
    handleParams $params param "-cornerradius 20 -width x -fill x -outline x -stipple x -tags x -dash {}"

    if { [lindex $coords 0] <= [lindex $coords 2] } {        
        set x1 [lindex $coords 0]
        set x2 [lindex $coords 2]
    } else {
        set x2 [lindex $coords 0]
        set x1 [lindex $coords 2]

    }
    if { [lindex $coords 1] <= [lindex $coords 3] } {
        set y1 [lindex $coords 1]
        set y2 [lindex $coords 3]
    } else {
        set y2 [lindex $coords 1]
        set y1 [lindex $coords 3]
    }
    set d $param(cornerradius)
    
    unset param(cornerradius)
    set params [getParamList param]
    return [list "polygon" [c::roundedRectanglePoints $x1 $y1 $x2 $y2 $d] [concat $params "-smooth yes"]]
}  



#############################################
#              solid3DRectangle             #
#############################################


proc canvasitems::embossedRectangle {coords params} {

    set x1 [lindex $coords 0]
    set y1 [lindex $coords 1]
    set x2 [lindex $coords 2]
    set y2 [lindex $coords 3]

    set res ""

    lappend res "rectangle"  "$x1 $y1  $x2 $y2"           "-fill gray"
    lappend res "line"       "$x1 $y2  $x1 $y1  $x2 $y1"  "-fill lightgray -width 2"
    lappend res "line"       "$x2 $y1  $x2 $y2  $x1 $y2"  "-fill darkgray -width 2"
    return $res
}

#############################################
#                 point                     #
#############################################

proc canvasitems::point {coords params} {

    set x [lindex $coords 0]
    set y [lindex $coords 1]

    set res ""

    lappend res "oval"  "[expr $x-2] [expr $y-2] [expr $x+2] [expr $y+2]" $params
    return $res
}

