
namespace eval clipboard {
    set data(obj) ""
}


proc clipboard::cut {elem} {
    variable data
    set data(obj) ""
    foreach e $elem {
        if ![edit::canExtract $e] {
            if {[llength $elem] == 1} {
                CDlg::Error [i18n::text "Can't cut this object!"] "Cut"
            } else {
                CDlg::Error [i18n::text "Can't cut at least one selected object!"] "Cut"
            }
            return
        }
    }
    foreach e $elem {
        change::extract $e
        lappend data(obj) $e
    }
    set data(copy) 0
}


proc clipboard::copy {elem} {
    variable data
    set data(obj) ""
    foreach e $elem {
        lappend data(obj) [change::copy $e [change::getFile $e]]
    }
    set data(copy) 1
}


proc clipboard::mark {elem} {
    variable data
    set data(obj) ""
    foreach e $elem {
        lappend data(obj) $e
    }
    set data(copy) 0
}


proc clipboard::linkExternalReferences {node} {
    # Die Referenzen des Teilbaums $node, die in die falsche Datei zeigen,
    # werden "intelligent" angepasst.
    
    if [c::isReference $node] {
        set file [d::getFile $node]
        set oldDest [c::getValue $node]
        if { ![nullptr $oldDest] && ![equal [d::getFile $oldDest] $file] } {
            if ![basiccomp::findByFullName $file [basiccomp::getFullName $oldDest] -var newDest] {
                set newDest 0x0
            }
            change::setValue $node $newDest
        }
    }
    foreach sub [c::getSubElements $node] {
        linkExternalReferences $sub
    }        
}


proc clipboard::paste {insContext} {
    variable data
    set insertedList ""
    if $data(copy) {
        foreach object $data(obj) {
            if {[edit::canInsert $insContext $object]} {
                set inserted [change::copy $object [change::getFile $insContext]]
                edit::insert $insContext $inserted
                linkExternalReferences $inserted
                lappend insertedList $inserted
                }
        }
    } else {
        foreach object $data(obj) {
            if {[edit::canInsert $insContext $object]} {
                if [equal [change::getFile $object] [change::getFile $insContext]] {
                    set inserted [change::copy $object [change::getFile $insContext]]

                } else {
                    set inserted $object
                }
                edit::insert $insContext $inserted
                linkExternalReferences $inserted
                lappend insertedList $inserted
            }
        }
        set data(obj) ""
    }
    return $insertedList
}


proc clipboard::canPaste {insContext} {
    variable data
    if [empty $data(obj)] {
        return 0
    }
    set result 0
    foreach object $data(obj) {
        set result 1
        set result [expr $result && [c::isSubtypeOf [c::getTypeId $object] [edit::getAcceptedType $insContext]]]
    }
    return $result
}
