
namespace eval console {}

proc console::open {} {
    set self ".console[newint]"
    upvar 0 ::console::$self data
    toplevel $self


    text $self.output
    pack $self.output -fill both -expand 1

    entry $self.entry
    pack $self.entry -fill x
    bind $self.entry <Return> [list console::onEnter $self]
    bind $self.entry <KeyPress-Down> [list console::onDown $self]
    bind $self.entry <KeyPress-Up> [list console::onUp $self]
    
    
    $self.output tag configure cmdTag -foreground red
    $self.output tag configure resultTag -foreground blue
    
    set data(addPointer) 0
    set data(showPointer) 0
    set data(maxEntries) $::cfg::consoleEntries
    set data(cmdQueue,0) ""
}


proc console::addText {self tag text} {
    $self.output insert end "$text\n" [list $tag]
}


proc console::evalCmd {cmd} {
    set code [catch $cmd result]
    if { [empty $result] } {
        set result "(empty result)"
    }
    return $result
}


proc console::onEnter {self} {
    set cmd [$self.entry get]
    $self.entry delete 0 end
    addText $self cmdTag $cmd

    set result [evalCmd $cmd]
    addText $self resultTag $result
    addCmdToCommandQueue $self $cmd
}

proc console::onDown {self} {
    set cmd [getNextCmdFromQueue $self]
    $self.entry delete 0 end
    $self.entry insert 0 $cmd
}

proc console::onUp {self} {
    set cmd [getLastCmdFromQueue $self]
    $self.entry delete 0 end
    $self.entry insert 0 $cmd
}


proc console::addCmdToCommandQueue {self cmd} {
    upvar 0 ::console::$self data
    set data(cmdQueue,$data(addPointer)) $cmd
    set data(addPointer) [expr [expr $data(addPointer) + 1] % $data(maxEntries)]
    set data(cmdQueue,$data(addPointer)) ""
    set data(showPointer) $data(addPointer)
}

proc console::getNextCmdFromQueue {self} {
    upvar 0 ::console::$self data
    set data(showPointer) [expr [expr $data(showPointer) - 1] % $data(maxEntries)]
    
    if { ![info exists data(cmdQueue,$data(showPointer))] } {
        set data(showPointer) [expr [expr $data(showPointer) + 1] % $data(maxEntries)]
       
    }
    return $data(cmdQueue,$data(showPointer))
}

proc console::getLastCmdFromQueue {self} {
    upvar 0 ::console::$self data
    set data(showPointer) [expr [expr $data(showPointer) + 1] % $data(maxEntries)]
    
    if { ![info exists data(cmdQueue,$data(showPointer))] } {
        set data(showPointer) [expr [expr $data(showPointer) - 1] % $data(maxEntries)]
       
    }
    return $data(cmdQueue,$data(showPointer))
}

