
# eventManager
#
# Verwaltet nderungsoperationen im Baum.
# Ruft die entsprechenden Listener in Fall von nderungen auf.
#

namespace eval eventManager {}

proc eventManager::new {} {
    set self [oop::new]
    oop::addClass $self "eventManager"
    oop::data $self
    set data(nextListenerId) 1
    set data(changeList) ""    
    set data(state) 0
    # 0: nix zu tun, 1: update after idle, 2: update luft
    return $self
}


proc eventManager::checkCmd_true {node} {
    return 1
}


proc eventManager::addListener {self changeDescription checkCmd priority cmd} {
    oop::data $self
    set listener "listener-$data(nextListenerId)"
    incr data(nextListenerId)
    addListenerObj $self $listener $changeDescription $checkCmd $priority $cmd
}


proc eventManager::addListenerObj {self listener changeDescription checkCmd priority cmd} {
    oop::data $self
    if {[empty $checkCmd]} {
        set checkCmd "eventManager::checkCmd_true"
    }
    set data(listener2priority,$listener) $priority
    set data(listener2cmd,$listener) $cmd
    set data(listener2autoRemove,$listener) [expr {![empty $changeDescription]}]
    set data(listener2checkCmd,$listener) $checkCmd
    set data(listener2nodes,$listener) $changeDescription

    foreach node $changeDescription {
        lappend data(node2listener,$node) $listener
    }
    if {[empty $changeDescription]} {
        lappend data(globalListeners) $listener
    }
    return $listener
}


proc eventManager::removeListener {self listener} {
    oop::data $self
    foreach node $data(listener2nodes,$listener) {
        lremove data(node2listener,$node) $listener
    }
    lremove data(globalListeners) $listener
}


proc eventManager::addChange {self node} {
    oop::data $self
    lappend data(changeList) $node
    scheduleUpdate $self
    log::log eventManager "Change occured: $node"
}


proc eventManager::callListener {self listener} {
    oop::data $self

    set start [clock clicks -milliseconds]
    eval $data(listener2cmd,$listener)
    set end [clock clicks -milliseconds]
    log::log eventManager [format "%3i ms for execution of '%s'" [expr {$end-$start}] $data(listener2cmd,$listener)]

    if { $data(listener2autoRemove,$listener) } {
        removeListener $listener
    }
}


proc eventManager::getInvolvedListeners {self change} {
    oop::data $self
    set result $data(globalListeners)
    if { [info exists data(node2listener,$change)] } {
        foreach listener $data(node2listener,$change) {
            if { [eval $data(listener2checkCmd,$listener) $change]==1 } {
                lappend result $listener
            }
        }
    }
    return $result
}


proc eventManager::actualizeListenerQueue {self queueName} {
    oop::data $self
    upvar $queueName queue
    foreach change $data(changeList) {
        foreach listener [getInvolvedListeners $self $change] {
            lappend queue [list $listener $data(listener2priority,$listener)]
        }
    }
    set data(changeList) ""
}


# privat
proc eventManager::processChanges {self} {
    oop::data $self
    if {$data(state)!=1} {
        error "eventManager::processChanges called in invalid state"
    }

    set data(state) 2
    set start [clock clicks -milliseconds]
    log::process "UPDATE" "starting update --------------------------------------"

    set callCount 0
    set listenerQueue ""
    actualizeListenerQueue $self listenerQueue
    while { ![empty $listenerQueue] } {
        set listenerQueue [lsort -unique $listenerQueue]
        set listenerQueue [lsort -index 1 $listenerQueue]
        callListener $self [lindex [lindex $listenerQueue 0] 0]
        set listenerQueue [lrange $listenerQueue 1 end]
        actualizeListenerQueue $self listenerQueue

        incr callCount
        if { $callCount > 100 } {
            log::error "Infinite synchronization?"
            break
        }
    }

    set data(state) 0
    set end [clock clicks -milliseconds]
    log::process "" "update ready; [expr $end - $start] ms ------------------------------------"
}

# privat
proc eventManager::scheduleUpdate {self} {
    oop::data $self
    if {$data(state)==0} {
        set data(state) 1
        after idle [list eventManager::processChanges $self]
    }
}
