
namespace eval exportHtml {}

proc exportHtml::quoteAnchor {s} {
    set result ""
    for {set i 0} {$i < [string length $s]} {incr i} {
        set c [string index $s $i]
        if [regexp {[a-zA-Z0-9-]} $c] {
            append result $c
        } else {
            append result "_"
        }
    }
    return $result
}

proc exportHtml::quoteFilename {s} {
    return [quoteAnchor $s]
}

proc exportHtml::saveEps {node viewName destPath} {

    set view [file::openView [file::getFileOfNode $node] $node $viewName -newView 1]

    update idletasks
    # xxx-id
    # warten, bis die Sicht tatschlich gezeichnet ist

    $view saveAsPostscript $destPath
    $view close
}

proc exportHtml::collectViews {node destDir listVar} {
    upvar $listVar list

    if [c::isLanguageObject $node] {
        set views [treedraw::getViewNames $node]
        if [llength $views]>0 {
            set type [lindex $views 0]
            #set name "$type [basiccomp::getFullName $node]"
            set name [basiccomp::getFullName $node]
            set epsPath [file join $destDir [quoteFilename $name].eps]
            saveEps $node $type $epsPath
            lappend list $name
        }
    }
    foreach subElem [c::getSubElements $node] {
        collectViews $subElem $destDir list
    }    
}


xn::proc exportHtml::export {file destDir type} {
    # destDir must exist
    >> xn::assert { [lsearch "html eps png" $type]!=-1 }

    ### Collect view informations
    set viewList ""
    collectViews [file::getRoot $file] $destDir viewList
    
    if {$type != "eps"} { 
        foreach viewName $viewList {
            set from [file join $destDir [quoteFilename $viewName].eps]
            set to   [file join $destDir [quoteFilename $viewName].png]

            convertEpsToPng $from $to
            file delete $from
        }
    }

    if {$type == "html"} {
        ### compute index.html
        set outFile [open [file join $destDir "index.html"] w]
        puts $outFile "<html>"
        puts $outFile "<head>"
        puts $outFile "<link rel='stylesheet' href='visExp.css' type='text/css'>"
        puts $outFile "<title>[file::getFilenameNoPath $file]</title>"
        puts $outFile "</head>"
        puts $outFile ""
        puts $outFile "<body>"
        puts $outFile "<h1>[file::getFilenameNoPath $file]</h1>"

        puts $outFile "<ul>"
        foreach viewName $viewList {
            puts $outFile "<li><a href='\#[quoteAnchor $viewName]'>$viewName</a>"
        }    
        puts $outFile "</ul>"

        foreach viewName $viewList {
            puts $outFile "<h2><a name='[quoteAnchor $viewName]'>$viewName</a></h2><p><img src='[quoteFilename $viewName].png' alt='Screenshot'>"
        }    
        puts $outFile "</body>"
        close $outFile

        ### copy css
        file copy -force [file join $::main::appHome "visExp.css"] $destDir

        ### convert eps -> png

    }
}

proc exportHtml::convertEpsToPng {from to} {
    upvar 0 ::tcl_platform tcl_platform
    ### Covert Parameters
    set res 92
    set antialias 4
    set scale 1
    
    ### END
    set fl [open $from]
    set data [read $fl]
    close $fl
    set lines [split $data \n]
    set found false;
    
    if {! [regexp {^%!PS-Adobe.*EPSF-} [lindex $lines 0]]} {
        append msg "Can't convert .eps to .png file!\n"
        append msg "EPS is not in a correct format"
        CDlg::Error $msg "HTML export"
        return false;
    }
    
    foreach line $lines {

        if {[regexp -nocase {^%%BoundingBox:\s*(\S+)\s+(\S+)\s+(\S+)\s+(\S+)} $line all 1 2 3 4]} {
            set found true
            break
        } else {
            if {[regexp -nocase {^%%EndComments} $line]} {
                append msg "Can't convert .eps to .png files!\n"
                append msg "EPS is not in a correct format: "
                append msg "no Bounding Box was found"
                CDlg::Error $msg "HTML export"
                return false;
                }
        }
        
    }
    
    if {$found} {
        
        set width [expr int (floor ([expr ($3 - $1) * $res/72 * $scale])) + 1]
        set height [expr int (floor ([expr ($4 - $2) * $res/72 * $scale])) + 1]
        
        if {$tcl_platform(platform) == "windows"} {
            set gsCommand "|gswin32c"
        } else {
            set gsCommand "|gs"
        }
        
        set fromList [split $from \\]
        set gsFrom [join $fromList /]
        set gsArgs "-q -dNOPAUSE -r$res -g${width}x$height -dTextAlphaBits=$antialias -dGraphicsAlphaBits=$antialias  -sDEVICE=png16m \"-sOutputFile=$to\""
        set gsInput "[expr 0 - $1] [expr 0 - $2] translate \n ($gsFrom) run\n showpage \n quit\n"
       
        set io [open "$gsCommand $gsArgs" w]
        puts "$gsCommand $gsArgs \n $gsInput"
        puts $io $gsInput
        flush $io
        close $io
        #exec $gsCommand $gsArgs <<$gsInput
        # Mit gs bzw. gswinc aus dem EPS ein PNG erzeugen
    
    }
    
    return $found
    

}
