
# Dieses Modul lst Fontangaben auf. Fonts knnen in DEViL
# durch symbolische Font-Familien spezifiziert werden.
# Dieses Modul ermittelt, welche Fonts bei solchen Angaben
# tatschlich verwendet werden.
#

namespace eval ::font {}



# Nachfolgend ist definiert, wie symbolische Familien umgesetzt werden.
# Symbolische Fontfamilien mssen mit einem Unterstrich beginnen.

# Typische Fonts:
# --------------
# Monospace: Courier, LucidaTypewriter, FotinosTypewriter, Terminal
# Proportional: Helvetica, Times


set font::class(_monospace)        "-family Courier  -size 10"
set font::class(_proportional)     "-family Arial    -size 10"
set font::class(_headline)         "-family Arial    -size 20"

set font::class(_standard)         $font::class(_monospace)
set font::class(_pp)               $font::class(_monospace)
set font::class(_treeView)         $font::class(_monospace)
set font::class(_messageView)      $font::class(_monospace)
set font::class(_textField)        $font::class(_monospace)


proc font::get {font} {
    # Lst eine symbolische Fontangabe auf. $font beginnt mit einem Font-Namen,
    # gefolgt von Attribut/Wert-Paaren

    variable class
    variable fontCache
    if [info exists fontCache($font)] {
        return $fontCache($font)
    }
    
    set family [lindex $font 0]
    if [equal [string index $family 0] "_"] {
        if [info exists class($family)] {
            array set arr $class($family)
        } else {
            log::error "Symbolic font '$font' is undefined!"
        }
    } else {
        set arr(-family) $family
    }
    array set arr [lrange $font 1 end]
    set result [array get arr]
    set result [eval font create $result]
    set fontCache($font) $result
    return $result
}



# PUBLIC
proc font::measure {text font} {
    # Achtung! Font ist die interne Font-Bezeichnung!!!

    variable measureCache
    if [info exists measureCache($font,$text)] {
        return $measureCache($font,$text)
    }
    
    set width 0
    set lines [split $text "\n"]
    foreach line $lines {
        set width [max $width [font measure $font $line]]
    }
    set height [expr [llength $lines]*[font metrics $font -linespace]]
    set result "$width $height"

    set measureCache($font,$text) $result
    return $result
}
