

xn::class SpecialDialogView {
    # does not inherit from AbstractView! Why?

    xn::classvar currObject "noCurrObject"

    xn::classmethod renewCreate {context kind optionList} {
        [$class private currObject] renewCreate $context $kind $optionList
    }


    xn::method init {root viewName} {
        set data(root) $root
        set data(file) [file::getFileOfNode $data(root)]
        set data(fileRoot) [c::getRoot $data(root)]
        set data(objects) ""
        set title "Edit [c::getClassName $root] properties"
        set data(toplevel) [CDlg::createDialog $title [list $object saveAndClose] [list $object cancel]]
        set data(workspace) [CDlg::Info $data(toplevel)]
        bind $data(workspace) <Destroy> [list $object destroy]
        set data(viewName) $viewName
        set data(viewType) $viewName
        set data(hidden) 1
        wm withdraw $data(toplevel)
        set data(initialEditState) [change::getEditState $data(fileRoot)]
        set data(normalTermination) 0
        
        file::addView $data(file) $object
        $object update
    }


    xn::method updateTitle {} {
    }


    xn::method update {} {
        $object renewStart
        c::startProcessor $data(viewName) $data(root)
        $object renewEnd
    }


    xn::method isReusable {node type} {
        return [expr [equal $node $data(root)] && [equal $type $data(viewType)]]
    }


    xn::method getFile {} {
        return $data(file)
    }


    xn::method destroy {} {
        # this function is automatically executed, after "destroy $data(toplevel)"
    
        file::removeView [$object getFile] $object
        if !$data(normalTermination) {
            # cancel button or window close button
            change::restoreState $data(fileRoot) $data(initialEditState)
        }
    }


    xn::method beforeDeletion {node} {
    }


    xn::method save {} {
        foreach obj $data(objects) {
            oop::call writeValue $data(workspace).[join $obj "-"]
        }
    }


    xn::method saveAndClose {} {
        change::clipCheckpoints $data(fileRoot) $data(initialEditState)
        $object save
        edit::forceUpdate $data(root)

        set data(normalTermination) 1
        destroy $data(toplevel)
    }


    xn::method cancel {} {
        destroy $data(toplevel)
    }


    xn::method read {} {
        foreach obj $data(objects) {
            oop::call readValue $data(workspace).[join $obj "-"]
        }    
    }


# --- renew handling ---------------------------------------------------------------


    xn::method renewStart {} {
        set data(renewObjects) ""
        $class private currObject $object
}


    xn::method renewEnd {} {
        $class private currObject "noCurrObject"
        set newObjects ""
        foreach {context kind optionList} $data(renewObjects) {
            set obj [list $context $kind]
            if ![elementOf $obj $data(objects)] {
                eval dcWithLabel::create $data(workspace).[join $obj "-"] $kind $optionList
                pack $data(workspace).[join $obj "-"] -fill both -expand [oop::call shouldExpand $data(workspace).[join $obj "-"]]
            } else {
                eval dcWithLabel::renew $data(workspace).[join $obj "-"] $optionList
            }
            lappend newObjects $obj
        }
        foreach old $data(objects) {
            if ![elementOf $old $newObjects] {
                destroy $data(workspace).[join $old "-"]
            }
        }
        set data(objects) $newObjects

        # reihenfolge anpassen
        set curr [lindex $newObjects 0]
        foreach obj [lrange $newObjects 1 end] {
            pack configure $data(workspace).[join $obj "-"] -after $data(workspace).[join $curr "-"]
            set curr $obj
        }


        $object read

        if $data(hidden) {
            set data(hidden) 0
            CDlg::Center $data(toplevel)
            CDlg::grabNoWait $data(toplevel)
        }
    }


    xn::method renewCreate {context kind optionList} {
        if { [equal $kind "None"] } return
        set kind "dc$kind"
        if [c::isValue $context] {
            lappend optionList -setCmd [list edit::setValue $context] -getCmd [list c::getValue $context]
        }
        if [c::isReference $context] {
            set values [treedraw::getVLPointerList $context]
            lappend optionList -setCmd [list edit::setValue $context] -getCmd [list c::getValue $context] -values $values
        }
        if [c::isAggregate $context] {
            #set values "(empty) {}"
            #foreach value [c::getAcceptedClassesForChild $context] {
            #    lappend values $value $value
            #}
            lappend optionList -setCmd [list edit::setClassOfChild $context] -getCmd [list c::getClassOfChild $context]
        }

        lappend optionList -onInput [list $object save]
        lappend data(renewObjects) $context $kind $optionList
    }
}
