
# treedraw.tcl
#
# Hilffunktionen zur Erzeugung von Kontextmenues und
# Editierdialogen, die in allen Sichten verwendet
# werden.


namespace eval ::treedraw {}



proc treedraw::showNodeInfo {objNode} {
    lappend message [i18n::text "Node ID: %s" $objNode]
    lappend message [i18n::text "Simple name: %s" [basiccomp::getSimpleName $objNode]]
    lappend message [i18n::text "Short name: %s" [basiccomp::getShortName $objNode]]
    lappend message [i18n::text "Full name: %s" [basiccomp::getFullName $objNode]]
    lappend message [i18n::text "Tree size: %s" [c::getObjCount $objNode]]

    CDlg::Information [join $message "\n\n"] [i18n::text "Node Information"]
}



proc treedraw::getViewNames {elem} {
    set className [c::getClassName $elem]
    if [info exists ::treedraw::viewnames($className)] {
        return $::treedraw::viewnames($className)
    } else {
        return ""
    }
}


proc treedraw::collectObjects {root listName} {
    upvar $listName list
    lappend list $root
    foreach aggObj [c::getSubElements $root] {
        foreach sub [c::getSubElements $aggObj] {
            collectObjects $sub list
        }
    }
}



proc treedraw::getVLPointerList {obj} {
    # Returns a list containing names and values of all valid references,
    # that can be stored in $obj.

    set currentValue [c::getValue $obj]
    set listContainsCurrentValue 0

    # collect objects
    set list ""
    collectObjects [c::getRoot $obj] list

    # check and add label
    set list2 ""
    foreach x $list {
        if [empty [checkutil::doValueCheck $obj $x]] {
            if [equal $x $currentValue] {
                set listContainsCurrentValue 1
            }
            set relName [basiccomp::getRelativeName $x $obj]
            switch -- [string first "::" $relName] {
                -1      {set sortPrefix a}
                0       {set sortPrefix c}
                default {set sortPrefix b}
            }
            lappend list2 [list $sortPrefix$relName $x]
        }
    }
    
    # sort
    set list2 [lsort -index 0 -dictionary $list2]

    # flatten
    set list ""
    lappend list [stringRep::nullValue2] "0x0"
    if {!$listContainsCurrentValue && ![nullptr $currentValue] } {
        lappend list "([basiccomp::getRelativeName $currentValue $obj])" $currentValue
    }
    foreach x $list2 {
        lappend list [string range [lindex $x 0] 1 end] [lindex $x 1]
    }
    return $list
}


proc treedraw::getRawEditor {obj showNoneAttributes} {

    # --- try to choose context based

    foreach context [c::getContexts $obj] {
       if [info exists treedraw::editor($context)] {
          if { [equal $treedraw::editor($context) "None"] && $showNoneAttributes} {
           return [getEditorByType $obj]
          }
           return $treedraw::editor($context)
        }
        if [info exists treedraw::editorComputedby($context)] {
            return [eval $treedraw::editorComputedby($context) $obj]
        }
    }

    # --- try to choose type based

    return [treedraw::getEditorByType $obj]
}

proc treedraw::getEditorByType {obj} {
    set type [c::getTypeName $obj]
    set ed(VLBoolean) Checkbox
    set ed(VLPointer) Listbox

    if { [info exists ed($type)] } {
        return $ed($type)
    }
    return "Entry"
}

proc treedraw::getEditor {obj showNoneAttributes} {
    set result [getRawEditor $obj $showNoneAttributes]
    if { [equal [c::getTypeName $obj] "VLPointer"] } {
        set values [getVLPointerList $obj]
        lappend result -values $values
    }
    return $result
}



proc treedraw::editAttributesEx {list title showNoneAttributes} {
    set dcList ""
    foreach obj $list {
        set role [c::getRole $obj]
        set className [c::getClassName [c::get $obj.PARENT]]

        set data($obj,doc) ""
        if [info exists ::helpToolTip::modelDoc($className.$role)] {
          set data($obj,doc) $::helpToolTip::modelDoc($className.$role)
        } {
          ### in Oberklassen suchen
          set contexts [c::getContexts [c::get $obj.PARENT]]
          foreach {element} $contexts {
           if [info exists ::helpToolTip::modelDoc($element.$role)] {
             set data($obj,doc) $::helpToolTip::modelDoc($element.$role)
             break
            }
          }
        }
        set data($obj,value) [c::getValue $obj]
        set data($obj,label) [c::getRole $obj]
        set data($obj,editor) [getEditor $obj $showNoneAttributes]
        set data($obj,check) "checkutil::doValueCheck $obj"
        if { ![equal [lindex $data($obj,editor) 0] "None"] } {
            lappend dcList $obj
        }
    }
    if ![empty $dcList] {
        set result [CDlg::stdInputDialog $title $dcList data]
        if $result {
            foreach obj $list {
                edit::setValue $obj $data($obj,value)
            }
        }
    }
}




proc treedraw::editAttribute {obj} {
    editAttributesEx [list $obj] [i18n::text "Edit Attribute"] 1
}

proc treedraw::orderAttributeList {list class} {
    if { ![info exists ::treedraw::attrOrder($class)] } {
        return $list
    }
    set order $::treedraw::attrOrder($class)
    set result ""
    foreach elem $order {
        set newList ""
        foreach id $list {
            if { [equal [c::getRole $id] $elem] } {
                lappend result $id
            } else {
                lappend newList $id
            }
        }
        set list $newList
    }
    return [concat $result $list]
}


proc treedraw::openStdAttributeDialog {obj} {
    # called by file::openAttributeDialog
    set list [lfilter [c::getSubElements $obj] {[c::isValue $x] || [c::isReference $x]}]
    set list [orderAttributeList $list [c::getClassName $obj]]
    editAttributesEx  $list [i18n::text "Edit %s properties" [c::getClassName $obj]] 0
}    


proc treedraw::editAttributes {obj} {
    file::openAttributeDialog [file::getFileOfNode $obj] $obj
}


proc treedraw::getSimpleName {elem} {
    if [nullptr $elem] {
        return [stringRep::nullValue]
    } else {
        return [basiccomp::getSimpleName $elem]
    }
}


proc treedraw::getShortName {elem} {
    if [nullptr $elem] {
        return [stringRep::nullValue]
    } else {
        return [basiccomp::getShortName $elem]
    }
}



proc treedraw::getFullName {elem} {
    if [nullptr $elem] {
        return [stringRep::nullValue]
    } else {
        return [basiccomp::getFullName $elem]
    }
}


proc treedraw::getValue {elem} {
    if [c::isReference $elem] {
        set result [getFullName [c::getValue $elem]]
    } else {
        set result [c::getValue $elem]
    }
    return [stringRep::quoted $result]
}



# --- context menu creation -------------------------------------------------------------------------------



proc treedraw::getSpecialContextmenu {elem menuListName} {
    upvar $menuListName menuList

    # Can be overwritten to support special operations in context menues.
    # elem can be a objNode or a valNode.
    #lappend menuList "Say hello" [list puts "Hello from node $elem!"]
}


proc treedraw::getGeneralContextmenu {elem menuListName} {
    upvar $menuListName menuList
    lappend menuList [i18n::text "Highlight in tree view"] [list file::highlightNodeInTreeView $elem]
    getSpecialContextmenu $elem menuList
}


proc treedraw::getEditContextmenu {obj prefix listName} {
    # obj ist ein EOLObject!
    upvar $listName commands

    set parent [c::getParent $obj]
    if { [c::maxMultiplicityIsOne $parent] && [c::getSubElementCount $parent]==1 } {
        return
    }

    set typelist [c::getSubtypes [edit::getAcceptedType $obj]]

    foreach type $typelist {
        if ![c::getIsAbstract $type] {
            lappend commands [i18n::text "%sInsert %s" $prefix [c::getNameOfType $type]] [list edit::insertNew $obj [c::getNameOfType $type]]
        }
    }
    if [clipboard::canPaste $obj] {
        lappend commands [i18n::text "%sPaste from Clipboard" $prefix] [list clipboard::paste $obj]
    }
}


proc treedraw::getObjNodeContextmenu {elem listName} {
    upvar $listName commands

    ### Part 1: general entries
    lappend commands [i18n::text "Edit attributes"] "treedraw::editAttributes $elem"
    lappend commands [i18n::text "Find usages"] "file::findUsages [file::getFileOfNode $elem] $elem"
    lappend commands [i18n::text "Show node infos"] "treedraw::showNodeInfo $elem"
    lappend commands [i18n::text "Export subtree"] "xmlio::exportSubtreeInteractive $elem"
    getGeneralContextmenu $elem commands
    lappend commands "---" ""

    ### Part 2: open other views
    foreach viewName [getViewNames $elem] {
        lappend commands [i18n::text "Open %s" $viewName] [list file::openView [file::getFileOfNode $elem] $elem $viewName]
    }
    lappend commands "---" ""

    ### Part 3: edit commands
    if [edit::canExtract $elem] {
        lappend commands [i18n::text "Delete"] [list edit::delete $elem]
        lappend commands [i18n::text "Cut"]    [list edit::cutIntoClipboard $elem]
        lappend commands [i18n::text "Mark"]   [list edit::mark $elem]
        lappend commands [i18n::text "Copy"]   [list edit::copyIntoClipboard $elem]
        getEditContextmenu [c::icBefore $elem] [i18n::text "Before -> "] commands
        getEditContextmenu [c::icAfter  $elem] [i18n::text "After -> "] commands
    }
    foreach sub [c::getAggregateSubElements $elem] {
        # creation commands for all SUB-Attributes
        getEditContextmenu [c::icBelow $sub] [i18n::text "Attribute %s -> " [c::getRole $sub]] commands

    }

     foreach sub [c::getSubElements $elem] {
        # creation of reveal commands for all REF-Attributes
        if [c::isReference $sub] {
          if ![nullptr [c::getValue $sub]] {
              lappend commands [i18n::text "Reveal -> %s" [c::getRole $sub]] [list file::highlightNode [list [c::getValue $sub]]]
          }
        }
    }


    lappend commands "---" ""
}


proc treedraw::getAttrNodeContextmenu {elem listName} {
    upvar $listName commands

    if [c::isLeaf $elem] {
        lappend commands [i18n::text "Edit attribute"] "treedraw::editAttribute $elem"        
    }
    if [c::isAggregate $elem] {
        getEditContextmenu [c::icBelow $elem] "" commands
    }
    if [c::isReference $elem] {
        lappend commands [i18n::text "Highlight referenced object"] [list file::highlightNodeInTreeView [c::getValue $elem]]
    }

    getGeneralContextmenu $elem commands
    lappend commands "---" ""
}


proc treedraw::contextMenu {elem x y} {
    # creates contextmenu for objNode or attrNode
    set menuList ""
    if [c::isLanguageObject $elem] {
        getObjNodeContextmenu $elem menuList
    }
    if [c::isRelationObject $elem] {
        getAttrNodeContextmenu $elem menuList
    }
    contextmenu::show $menuList $x $y
}

# ---------------------------------------------------------------------------------------------------



