
# Module: vlDraw
#
# Naming conventions:
#
# - objects: Objects that are visible; they consist of a set of primitives (special "groups")
#   Coding: obj-$treeContext
#
# - inspoints: Marks, that accept structure insertions
#   Coding: inspoint-$treeContext
#
# - dragHandles: Handles, that allow to define a reference attribute
#   Coding: dragHandle-$treeContext
#
# - primitives: tcl canvas items

namespace eval ::vlDraw {
}



###################################################################################### handler functions

# Levels:
#   0: normal
#   1: insertionLines
#   2: handles


proc vlDraw::acceptDropPredicate_clipboard {group} {
    if { ![equal [vlDraw::getType $group] "inspoint"] } {
        return 0
    }
    return [clipboard::canPaste [getTreeContext $group]]
}


proc vlDraw::acceptDropPredicate_type {className group} {
    if { ![equal [vlDraw::getType $group] "inspoint"] } {
        return 0
    }
    return [c::canInsertType [getTreeContext $group] [c::getTypeIdFromName $className]]
}


proc vlDraw::onDrop_insertNew {className userAction canvas fromObj toObj absPos delta} {
    set newNode [edit::createObject  $className [edit::getFile [getTreeContext $toObj]]]
    edit::insert [getTreeContext $toObj] $newNode
    setVisualPosition $canvas $newNode $toObj $absPos
    if { ![empty $userAction] } {
        eval $userAction $newNode
    }
    activeCanvas::selectAfterUpdate $canvas "obj-$newNode"
}


proc vlDraw::onDrop_insertClipboard {canvas fromObj toObj absPos delta} {
    set insertedL [clipboard::paste [getTreeContext $toObj]]
    foreach inserted $insertedL {
        setVisualPosition $canvas $inserted $toObj $absPos
        activeCanvas::selectAfterUpdate $canvas "obj-$inserted"
    }
}


proc vlDraw::onDrop_dragHandleDrop {canvas fromObj toObj absPos delta} {
    edit::setValue [getTreeContext $fromObj] [getTreeContext $toObj]
}


proc vlDraw::setPositionFkt_std {zeroPos attrName node newAbsPos} {
    set newPos [vec::sub $newAbsPos $zeroPos]
    set newPos [vec::round $newPos $cfg::positionQuantum]
    set positionObj [c::getSubElement $node $attrName]
    edit::setValue $positionObj $newPos
}


proc vlDraw::setVisualPosition {canvas fromNode toObj absPos} {
    if { [activeCanvas::getAttr $canvas $toObj setPositionFkt -dest fkt] } {
        eval $fkt [list $fromNode $absPos]
        edit::forceUpdate $fromNode
    }
}

proc vlDraw::old_setVisualPosition {canvas fromNode toObj absPos} {
    if { [activeCanvas::getAttr $canvas $toObj insertionRegionCoords -dest coords] } {
        if { ![activeCanvas::getAttr $canvas $toObj roundingValue -dest roundingValue] } {
            set roundingValue 1
        }        
        assign "x y a b" $coords
        set newPosition [vec::sub $absPos "$x $y"]
        set newPosition [vec::round $newPosition $roundingValue]
        set positionObj [c::getSubElement $fromNode position]
        edit::setValue $positionObj $newPosition
    }

}

proc vlDraw::onDrop_objectDrop {canvas fromObj toObj absPos delta} {
    if { [equal [getType $toObj] "reposition"] } {
        set fkt [activeCanvas::getAttr $canvas $fromObj "repositionFkt"]
        set data(absPos) $absPos
        set data(delta) $delta
        eval $fkt data
    } else {
        if { [dropEffect::isApplicable [getTreeContext $fromObj] [getTreeContext $toObj]] } {
            set resultObject [dropEffect::apply [getTreeContext $fromObj] [getTreeContext $toObj]]
        } else {
            # insert, aber vermeide Umsortierung bei VPSet!
            if { ![activeCanvas::getAttr $canvas $toObj setPositionFkt -dest dummy] || ![c::siblings [getTreeContext $toObj] [getTreeContext $fromObj]] } {
                edit::insert [getTreeContext $toObj] [getTreeContext $fromObj]
            }
            set resultObject [getTreeContext $fromObj]
        }
        if { ![empty $resultObject] } {
            setVisualPosition $canvas $resultObject $toObj $absPos
        }
    }
}


proc vlDraw::setSizeFkt_std {zeroPos obj absPos} {
    set newSize [vec::sub $absPos $zeroPos]
    set newSize [vec::round $newSize $cfg::positionQuantum]
    assign "x y" $newSize
    edit::setValue $obj [list [max $x 1] [max $y 1]]
}


proc vlDraw::setSizeEWFkt {zeroPos obj absPos} {
    set newSize [vec::sub $absPos $zeroPos]
    set newSize [vec::round $newSize $cfg::positionQuantum]
    assign "x y" $newSize
    edit::setValue $obj [max $x 1]
}

proc vlDraw::clickFkt_std {clickAttr} {

    edit::setValue $clickAttr [expr ! [c::getValue $clickAttr]]
}

proc vlDraw::onDrop_resizerDrop {canvas fromObj toObj absPos delta} {
    set cmd [activeCanvas::getAttr $canvas $fromObj setSizeFkt]
    eval $cmd [list $absPos]
    edit::forceUpdate [getTreeContext $fromObj]
}


proc vlDraw::old_onDrop_resizerDrop {canvas fromObj toObj absPos delta} {
    set newSize [vec::sub $absPos [activeCanvas::getAttr $canvas $fromObj zeroPosition]]
    edit::setValue [getTreeContext $fromObj] $newSize
}

proc vlDraw::onDoubleClick_removeIntermediate {obj index group} {
    set oldPos [c::getValue $obj]
    edit::setValue $obj [lreplace $oldPos [expr 2*$index] [expr 2*$index+1]]
}

proc vlDraw::attrRepositionFkt_adjustIntermediate {zeroPos obj index absPos} {
    set newPos [vec::sub $absPos $zeroPos]
    set newPos [vec::round $newPos $cfg::positionQuantum]
    assign "x y" $newPos
    set oldPos [c::getValue $obj]
    edit::setValue $obj [lreplace $oldPos [expr 2*$index] [expr 2*$index+1] $x $y]
}

proc vlDraw::attrRepositionFkt_addIntermediate {zeroPos obj index absPos} {
    set newPos [vec::sub $absPos $zeroPos]
    set newPos [vec::round $newPos $cfg::positionQuantum]
    assign "x y" $newPos
    set oldPos [c::getValue $obj]
    edit::setValue $obj [linsert $oldPos [expr 2*$index] $x $y]
}

proc vlDraw::attrRepositionFkt_createLinePartPoints {zeroPos obj index absPos } {
    
    set newPos [vec::sub $absPos $zeroPos]
    set newPos [vec::round $newPos $cfg::positionQuantum]
    assign "x y" $newPos
    assign "realX realY" $zeroPos
    set oldKoords [c::getValue $obj]
    
    set X [expr $x+$realX]
    set Y [expr $y+$realY]
    
    set len [llength $oldKoords]
    set newIndex [expr $index+2]
    
    set vFlag [lindex $oldKoords 0]
    if { $vFlag == "v" } {
     set vFlag 1
    } else { 
     set vFlag 0
    }
    
    set ident [expr ($vFlag+$index)%2]
    
    if {$ident==0} {    #horizontaler abschnitt
       set newKoord $Y
    } else {            #vertikaler abschnitt
       set newKoord $X
    }
    
    if {$index==0} {
       set newKoords [linsert $oldKoords 1 $newKoord ]
       if {$vFlag==1} {
         set newKoords [lreplace $newKoords 0 0 "h"]
       } else {
         set newKoords [lreplace $newKoords 0 0 "v"]
       }
    } elseif {$index==$len} {
       set newKoords [linsert $oldKoords $newIndex $newKoord ]
    } else {
       set newKoords [lreplace  $oldKoords $index $index $newKoord]
    }
    set len [llength $newKoords]
    set i 0
    set tmpKoords ""
    
    while {$i<$len} {
      set tmp1 [lindex $newKoords $i]
      lappend tmpKoords $tmp1
      
      set tmp2 [lindex $newKoords [expr $i+2]]
      if {$tmp1==$tmp2} {
        set i [expr $i+3]
      } else {
        incr i
      }
    }
    
    set newKoords $tmpKoords
    edit::setValue $obj $newKoords 
}


proc vlDraw::attrRepositionFkt_std {zeroPos obj absPos} {
    set newPos [vec::sub $absPos $zeroPos]
    set newPos [vec::round $newPos $cfg::positionQuantum]
    edit::setValue $obj $newPos
}

# XXX
proc vlDraw::onDrop_attrReposition {canvas fromObj toObj absPos delta} {
    set cmd [activeCanvas::getAttr $canvas $fromObj attrRepositionFkt]
    eval $cmd [list $absPos]
    edit::forceUpdate [getTreeContext $fromObj]
}


proc vlDraw::old_onDrop_attrReposition {canvas fromObj toObj absPos delta} {
    set newValue [vec::add [c::getValue [getTreeContext $fromObj]] $delta]
    edit::setValue [getTreeContext $fromObj] $newValue
}


proc vlDraw::acceptDropFkt_dragHandleDrop {canvas fromObj toObj} {
    if { ![equal [getType $toObj] "obj"] } {
        return 0
    }
    return [empty [checkutil::doValueCheck [getTreeContext $fromObj] [getTreeContext $toObj]]]
}


proc vlDraw::acceptDropFkt_objectDrop {canvas fromObj toObj} {
    if { [equal [getType $toObj] "reposition"] } {
        return [activeCanvas::getAttr $canvas $fromObj "repositionFkt" -dest dummy]
    }
    if { ![equal [getType $toObj] "inspoint"] } {
        return 0
    }
    if { [edit::canInsert [getTreeContext $toObj] [getTreeContext $fromObj]] } {
        return 1
    }
    if { [dropEffect::isApplicable [getTreeContext $fromObj] [getTreeContext $toObj]] } {
        return 1
    }
    return 0
}


proc vlDraw::acceptDropFkt_resizerDrop {canvas fromObj toObj} {
    set result [equal [getType $toObj] "resizerRegion"]
    return $result
}

proc vlDraw::acceptDropFkt_attrReposition {canvas fromObj toObj} {
    set result [equal [getType $toObj] "attrReposition"]
    return $result
}

proc ::vlDraw::onButtonClick_std {} {
    
}


################################################################################################ internal


proc vlDraw::start {activeCanvas root} {
    variable canvas
    variable objects
    variable data
    set canvas $activeCanvas
    activeCanvas::startRedraw $canvas
    catch {unset objects}
    catch {unset data}

    set data(root) $root
    set data(font) [font::get "_pp"]
    set data(deltay) [font metrics $data(font) -linespace]
    set data(space)  [font measure $data(font) " "]
    set data(indent) [expr 3*$data(space)]
    set data(kwColor) $cfg::vlTextKwColor
    set data(valColor) $cfg::vlTextValColor
    
}
   

proc vlDraw::end {} {
    variable canvas
    variable objects
    variable data
    foreach currn [array names objects] {
        activeCanvas::defGroup $canvas $currn \
            -onContextMenu vlDraw::showObjectContextMenu \
            -onDoubleClick [list vlDraw::onDoubleClick $data(root)] \
            -onDrop vlDraw::onDrop_objectDrop \
            -acceptDropFkt vlDraw::acceptDropFkt_objectDrop \
            -onVisualDrop activeCanvas::onVisualDrop_nothing
        if { ![info exists data($currn,noSelectionBorder)] } {
            activeCanvas::newElement $canvas $currn "rect" [activeCanvas::getBoundingBox $canvas $objects($currn)] \
                -normal "-outline {} -dropable 0" \
                -select [list -outline $cfg::selectionColor] \
                -highlight [list -outline $cfg::interactionColor -width 3]
        }
    }

    # resizer drop target
    activeCanvas::defGroup $canvas resizerRegion-0x0 -onVisualDrop "activeCanvas::onVisualDrop_movePreview"
    activeCanvas::newElement $canvas resizerRegion-0x0 "rect" "0 0 0 0" -normal "-outline {} -dropable 1"

    # attr reposition drop target
    activeCanvas::defGroup $canvas attrReposition-0x0 -onVisualDrop "activeCanvas::onVisualDrop_movePreview"
    #activeCanvas::defGroup $canvas attrReposition-0x0 -onVisualDrop "activeCanvas::onVisualDrop_nothing"
    activeCanvas::newElement $canvas attrReposition-0x0 "rect" "0 0 0 0" -normal "-outline {} -dropable 1"

    # object reposition drop target
    activeCanvas::newElement $canvas reposition-0x0 rectangle [activeCanvas::getBBox $canvas] \
            -normal "-outline {} -dropable 1" -highlight "-outline $cfg::interactionColor"
    activeCanvas::defGroup $canvas reposition-0x0 -onVisualDrop [list activeCanvas::onVisualDrop_insertionCursor "30 30"]

    # attr reposition drop target


    activeCanvas::endRedraw $canvas
    unset canvas
}


proc vlDraw::getTreeContext {string} {
    set result [lindex [split $string "-"] end]
    #old solution: regexp {.*-(0x[0-9a-h]+)$} $string dummy result
    return $result
}


proc vlDraw::getType {string} {
    set result [join [lrange [split $string "-"] 0 end-1] "-"]
    #old solution: regexp {(.*)-0x[0-9a-h]+$} $string dummy result
    return $result
}


proc vlDraw::showObjectContextMenu {object X Y} {
    treedraw::contextMenu [getTreeContext $object] $X $Y
}

proc vlDraw::onDoubleClick {viewRoot object} {

    set node [getTreeContext $object]
    set viewList [treedraw::getViewNames $node]
    if { ![empty $viewList] && ![equal $node $viewRoot] } {
        after 100 file::openView [file::getFileOfNode $node] $node [lindex $viewList 0]
        # XXX: ohne Verzoegerung wird das Fenster hinter dem angeklickten geoeffnet
    } else {
        if { [file::getReadonly [file::getFileOfNode $node]] } {
            vlDraw::doubleclickIfReadonly $node
        } else {
            treedraw::editAttributes $node
        }
    }
}



proc vlDraw::doubleclickIfReadonly {node} {
    CDlg::Error "File is readonly"
}



proc vlDraw::isHidden {coords} {
    set x [lindex $coords 0]
    return [expr "-20000 <= $x && $x < -10000"]
}


################################################################################################# object creation

# PUBLIC
proc vlDraw::disableSelectionBorder {treeContext} {
    variable data
    set data(obj-$treeContext,noSelectionBorder) 1
}


# PUBLIC
proc vlDraw::createObject {treeContext kind coords normalAttributes selectAttributes} {
    if { [isHidden $coords] } return

    variable canvas
    variable objects
    set clickable [expr ![nullptr $treeContext]]
    set newObj [activeCanvas::newElement $canvas obj-$treeContext $kind $coords \
       -normal [concat "-clickable $clickable -dropable 1" $normalAttributes]  \
       -select $selectAttributes  \
       -print 1 \
    ]
   
    lappend objects(obj-$treeContext) $newObj
    return $newObj
}

proc vlDraw::createClickableObject {treeContext kind coords normalAttributes selectAttributes clickAttr clickFkt} {
    
    variable canvas    
    set newObj [createObject $clickAttr $kind $coords $normalAttributes $selectAttributes]
    activeCanvas::defGroup $canvas clickAble-obj-$clickAttr -onClick $clickFkt
    activeCanvas::setAttr $canvas clickAble-obj-$clickAttr clickAttr $clickAttr
    
}

proc  vlDraw::addAnimation {cObj animPath} {
    variable canvas
    activeCanvas::addAnimation $canvas $cObj $animPath
}


# PUBLIC
proc vlDraw::createDragHandle {treeContext coords} {
    if { [isHidden $coords] } return

    variable canvas
    variable objects
    set obj [c::getParent $treeContext]
    activeCanvas::defGroup $canvas dragHandle-$treeContext \
            -onContextMenu vlDraw::showObjectContextMenu \
            -onDrop vlDraw::onDrop_dragHandleDrop \
            -acceptDropFkt vlDraw::acceptDropFkt_dragHandleDrop \
            -dominatorGroup obj-$obj

    activeCanvas::newElement $canvas dragHandle-$treeContext "oval" $coords \
      -normal [list -outline {} -fill {} -clickable 0] \
      -select [list -outline $cfg::interactionColor -fill $cfg::interactionColor -clickable 1] -level 2
}


# PUBLIC
proc vlDraw::createInsertionPoint {treeContext coords} {
    if { [isHidden $coords] } return

    if { [c::isRelationObject $treeContext] } {
        set treeContext [c::getEOL $treeContext]
    }
    variable canvas
    set id [activeCanvas::newElement $canvas inspoint-$treeContext rectangle $coords \
                -normal "-width 3 -fill $cfg::containerColor -outline {} -clickable 1 -dropable 1" \
      -highlight "-outline $cfg::interactionColor" -rotate 0]

}


# PUBLIC
proc vlDraw::createInsertionLine {treeContext coords} {
    if { [isHidden $coords] } return

    if { [c::isRelationObject $treeContext] } {
        set treeContext [c::getEOL $treeContext]
    }
    variable canvas
    set id [activeCanvas::newElement $canvas inspoint-$treeContext line $coords \
      -normal "-width 3 -fill {} -dropable 1" \
      -highlight "-fill black" -rotate 1 -level 1]

}

# PUBLIC
proc vlDraw::createInsertionRegion {treeContext coords setPositionFkt defaultCursorSize} {
    if { [isHidden $coords] } return

    set treeContext [c::getEOL $treeContext]
    variable canvas
    set id [activeCanvas::newElement $canvas inspoint-$treeContext rectangle $coords \
              -normal "-outline {} -dropable 1" -highlight "-outline $cfg::interactionColor"]
    activeCanvas::defGroup $canvas inspoint-$treeContext -onVisualDrop [list activeCanvas::onVisualDrop_insertionCursor $defaultCursorSize]
    #activeCanvas::setAttr $canvas inspoint-$treeContext insertionRegionCoords $coords
    activeCanvas::setAttr $canvas inspoint-$treeContext setPositionFkt $setPositionFkt
}

# PUBLIC
proc vlDraw::createObjectReposition {treeContext repositionFkt} {
    variable canvas
    activeCanvas::setAttr $canvas obj-$treeContext repositionFkt $repositionFkt
}


# PUBLIC
proc vlDraw::createResizer {sizeAttr coords setSizeFkt} {
    if { [isHidden $coords] } return

    variable canvas
    set obj [c::getParent $sizeAttr]
    assign "x1 y1 x2 y2" $coords
    set buttonCoords [vec::makeSquare "$x2 $y2" $cfg::handleSize]
    set id [activeCanvas::newElement $canvas obj-$obj rectangle $coords \
      -normal "-fill {} -outline {} -clickable 0" -select "-outline $cfg::interactionColor -clickable 1" -level 2]
    activeCanvas::newElement $canvas resizer-$sizeAttr rectangle $buttonCoords \
      -normal "-fill {} -outline {} -clickable 0" -select "-fill $cfg::interactionColor -clickable 1" -level 2
    activeCanvas::defGroup $canvas resizer-$sizeAttr -dominatorGroup obj-$obj \
      -acceptDropFkt vlDraw::acceptDropFkt_resizerDrop -onDrop vlDraw::onDrop_resizerDrop
    activeCanvas::setAttr $canvas resizer-$sizeAttr setSizeFkt $setSizeFkt
    activeCanvas::setAttr $canvas resizer-$sizeAttr visualDragParams "$id 1"
}


# XXX
# PUBLIC
proc vlDraw::createAttrReposition {posAttr coords attrRepositionFkt options} {
    if { [isHidden $coords] } return

    variable canvas
    handleParams $options option "-visualDragParams {} -onDoubleClick activeCanvas::onDoubleClick_nothing"

    set obj [c::getParent $posAttr]
    set attrReposition "attrReposition-[newint]"
    set id [activeCanvas::newElement $canvas $attrReposition-$posAttr rectangle $coords \
                -normal "-fill {} -outline {} -clickable 0" -select "-fill $cfg::interactionColor -clickable 1" -level 2]
    activeCanvas::defGroup $canvas $attrReposition-$posAttr -dominatorGroup obj-$obj \
      -acceptDropFkt vlDraw::acceptDropFkt_attrReposition -onDrop vlDraw::onDrop_attrReposition -onDoubleClick $option(onDoubleClick)

    activeCanvas::setAttr $canvas $attrReposition-$posAttr visualDragParams $option(visualDragParams)
    activeCanvas::setAttr $canvas $attrReposition-$posAttr attrRepositionFkt $attrRepositionFkt
}


# XXX
# PUBLIC

proc vlDraw::createButton {label onClickFunction pos} {
    variable canvas
    set params ""
    lappend params "-text"
    lappend params $label
    lappend params "-command"
    lappend params $onClickFunction
    activeCanvas::createWindowObject $canvas button $pos $params
}


################################################################### special support for pretty printing

# Format of the position record:
# x y indent state dryRun maxX


proc vlDraw::getTextRegion {coords1 coords2} {
    variable data

    assign "x1 y1 indent state dryRun maxX maxY lineMaxY" $coords1
    assign "x2 y2 indent state dryRun maxX maxY lineMaxY" $coords2
    incr y2 $data(deltay)
    return [list $x1 $y1 $x2 $y2]
}


proc vlDraw::state {name} {
    switch $name {
        nothing {return 0}
        blankOrSymbol {return 1}
        newLine {return 2}
        newParagraph {return 3}
        default {error "no such state"}
    }
}

proc vlDraw::setSpecialState {coords newState} {
    variable data

     assign "x y indent state dryRun maxX maxY lineMaxX" $coords

    if { $newState > $state } {
        set state $newState
    }
    set maxY 0
     set lineMaxY 0
    return "$x $y $indent $state $dryRun $maxX $maxY $lineMaxY"
}

proc vlDraw::incrIndent {coords factor} {
    variable data

     assign "x y indent state dryRun maxX maxY lineMaxX" $coords


    set maxY 0
    set lineMaxY 0
    incr indent [expr $factor*$data(indent)]
    return "$x $y $indent $state $dryRun $maxX $maxY $lineMaxY"
}    


proc vlDraw::isSymbol {text} {
    if { [string length $text]>2 } {
        return 0
    }
    if { [string is alnum [string index $text 0]] } {
        return 0
    }
    return 1
}

# PRIVATE
proc vlDraw::printNext {treeContext coords text color} {
    # Allgemein verwendete Ausgabefunktion

    variable data
    

     assign "x y indent state dryRun maxX maxY lineMaxY" $coords



    set maxY 0
    set lineMaxY 0

    set blank ""
    if { $state == [state blankOrSymbol] } {
        if { ![isSymbol $text] } {
            set blank " "
        }
    }
    if { $state >= [state newLine] } {
        set x $indent
        set y [expr $y+$data(deltay)]
        
    }
    if { $state >= [state newParagraph] } {
        incr y $data(deltay)
    }
    set width [font measure $data(font) "${blank}${text}"]
    if { !$dryRun } {
        #createInsertionLine $treeContext [list [expr $x-2] $y [expr $x-2] [expr $x+$data(deltay)]]
        createObject $treeContext "text" [list $x $y] [list -text "${blank}${text}" -fill $color -anchor nw -font $data(font)] ""
    }
    set endPos [expr $x+$width]
    if { $endPos > $maxX } {
        set maxX $endPos
    }
    set coords [list [expr $x+$width] $y $indent 0 $dryRun $maxX $maxY $lineMaxY]
    if { ![isSymbol $text] } {
        set coords [setSpecialState $coords [state blankOrSymbol]]
    }
    return $coords
}


# PRIVATE
proc vlDraw::createKeywordInternal {treeContext coords text} {
    # Intern verwendete Ausgabefunktion fr Schlsselwrter und Formatierungsanweisungen

    variable data
    if { [equal $text ""] } {
        # do nothing
    } elseif { [equal $text "!NL"] } {
        set coords [setSpecialState $coords [state newLine]]
    } elseif { [equal $text "!P"] } {
        set coords [setSpecialState $coords [state newParagraph]]
    } elseif { [equal $text "!INCR"] } {
        set coords [incrIndent $coords 1]
    } elseif { [equal $text "!DECR"] } {
        set coords [incrIndent $coords -1]
    } elseif { [equal $text "!ID"] } {
        set text [treedraw::getSimpleName $treeContext]
        set coords [printNext $treeContext $coords $text $cfg::vlTextDefColor]
    } else {
        set coords [printNext $treeContext $coords $text $data(kwColor)]
    }
    return $coords
}


# Einige, fr C-hnliche Sprachen ntzliche Formatierungsregeln.
# Sie knnen je nach Bedarf berschrieben werden.

set vlDraw::kw(\{)     [list "\{" !NL !INCR]
set vlDraw::kw(\})     [list !NL !DECR "\}" !NL]
set vlDraw::kw(\}\ )   [list !NL !DECR "\} "]
set vlDraw::kw(\;)     [list ";"   !NL]
set vlDraw::kw(!SPACE) [list " "]
set vlDraw::kw(!BEGIN) [list !INCR !NL]
set vlDraw::kw(!END)   [list !DECR !NL]
set vlDraw::kw(!HR)    [list !P "--------------------" !P]
set vlDraw::kw(!EOF)   [list !NL " "]


# PUBLIC
proc  vlDraw::createKeyword {treeContext coords text} {
    # Ausgabefunktion fr Schlsselwrter und Formatierungsanweisungen.
    # Formatierungsanweisungen knnen durch berschreiben von
    # vlDraw::kw den jeweiligen Bedrfnissen angepasst werden.

    variable kw
    if { [info exists kw($text)] } {
        foreach cmd $kw($text) {
            set coords [createKeywordInternal $treeContext $coords $cmd]
        }
    } else {
        set coords [createKeywordInternal $treeContext $coords $text]
    }
    return $coords
}



# PRIVATE
proc vlDraw::createRefText {treeContext coords text options} {
    # Erzeug einen Id-Text fr die gegebene Referenz.

    if { [empty $text] } {
        set text "\#\#\#"
    }
    set coords [printNext $treeContext $coords $text $cfg::vlTextRefColor]
    return $coords
}



# PUBLIC
proc vlDraw::createText {treeContext coords text options} {
    # Ausgabefunktion fr Text-Token.

    variable data
    if { ![empty $options] } {
        if { [empty $text] } {
            set output "???"
        } else {            
            if { [catch { set output [lindex $options $text] }] } {
                set output "???"
            } else {
                if { $text<0 || $text>=[llength $options] } {
                    set output "???"
                }
            }
            if { [equal output "???"] } {
                log::log error "While printing [c::getContext $treeContext]: Invalid attribute value: '$text' "
            }
        }
        set coords [createKeyword $treeContext $coords $output]
    } else {
        set coords [printNext $treeContext $coords $text $data(valColor)]
    }
    return $coords
}


# PUBLIC
proc vlDraw::createSeparators {treeContext coords separators} {
    foreach sep $separators {
        set coords [createKeyword $treeContext $coords $sep]
    }
    return $coords
}

proc vlDraw::getFlowStatePosition {flowState} {
    variable data
    assign "x y indent state dryRun maxX maxY lineMaxY" $flowState
    return "$x $y"

}

proc vlDraw::printNextObject {treeContext coords preferedSize internalState} {
    variable data
    assign "x y indent state dryRun maxX maxY lineMaxY" $coords
    assign "prefX prefY" $preferedSize
    if [expr [equal $internalState "\n"] || [isLastElement $treeContext]] {
      if [expr $prefY > $lineMaxY] {
       set maxY [expr $prefY + $maxY]
       set lineMaxY $prefY
      } {
       set maxY [expr $maxY + $lineMaxY]
      }
      set y [expr $y + $lineMaxY]
      set lineMaxY 0
      if [expr ($x + $prefX) > $maxX] {
       set maxX [expr $x + $prefX]
      }
      set x [expr $indent ]
    } {
      set x [expr $x + $prefX]
      if [expr $x > $maxX] {
       set maxX $x
      }
      if [expr $prefY > $lineMaxY] {
       set lineMaxY $prefY
      }
    }
    set coords "$x $y $indent $state $dryRun $maxX $maxY $lineMaxY"
 return $coords
}

proc vlDraw::getCurrentPosition {treeContext} {
 set thisList [c::getList $treeContext.PARENT.CHILDREN]
 return [lsearch -exact $thisList $treeContext]
}

proc vlDraw::isLastElement {treeContext} {
 set listLength [llength [c::getList $treeContext.PARENT.CHILDREN]]
 if [equal [getCurrentPosition $treeContext] [expr $listLength - 1]] {
  return 1
 }
 return 0
}

proc vlDraw::createCompactObject {treeContext coords preferedSize internalState} {
 set coords [printNextObject $treeContext $coords $preferedSize $internalState]



 return $coords
}



################################################################################## information retrieval

# PUBLIC
proc vlDraw::getSimpleName {treeContext} {
    return [basiccomp::getSimpleName $treeContext]
}

# PUBLIC
proc vlDraw::getShortName {treeContext} {
    return [basiccomp::getShortName $treeContext]
}

# PUBLIC
proc vlDraw::getFullName {treeContext} {
    return [basiccomp::getFullName $treeContext]
}


# PUBLIC
proc vlDraw::getRelativeName {treeContext base} {
    return [basiccomp::getRelativeName $treeContext $base]
}


# PUBLIC
proc vlDraw::getNamespace {treeContext} {
    return [basiccomp::getNamespace $treeContext]
}


# PUBLIC
proc vlDraw::measure {text font} {
    return [font::measure $text $font]
}



# PUBLIC
proc vlDraw::getCfg {key} {
    return [subst "\$::cfg::$key"]
}

# PUBLIC
proc vlDraw::measureImage {name} {
    return [list [image width $name] [image height $name]]
}

proc vlDraw::measureButton {label} {
    variable canvas
    set params ""
    lappend params "-text"
    lappend params $label
    return [activeCanvas::measureWindowObject $canvas button $params]
}

# PUBLIC
proc vlDraw::getValidImage {name} {
    if { [catch [list image type $name]] } {
        log::error "Image '$name' does not exist. Using 'img::defaultimage'."
        return "img::defaultimage";
    } else {
        return $name
    }
}


proc vlDraw::drawGrid {id grid origin size bgColor gridColor imageName} {

    if { [empty $imageName] } {
        set newImageSize [expr $grid*5]
        set imageName [makeGridImage $newImageSize $grid $bgColor $gridColor]
    }  
    set imageSize [vlDraw::measureImage $imageName]
    
    assign "xSize ySize" $size
    assign "xOrigin yOrigin" $origin
    assign "heigth width" $imageSize
    
    set i [expr $xSize/$heigth]
    set j [expr $ySize/$width]
    
    for {set k 0} {$k<$i} {incr k} {
        for {set n 0} {$n<$j} {incr n} {
            set xPos [expr { $k*$heigth + $xOrigin }]
            set yPos [expr { $n*$width + $yOrigin }]
            set options [list "-anchor" "nw" "-image" $imageName]
            ::vlDraw::createObject $id "image" "$xPos $yPos" $options ""
        }
    }
}


proc vlDraw::makeGridImage {size grid bgcolor gridcolor} {
 
 image create photo gridImage$grid
 set gridHalf [expr $grid/2]
 gridImage$grid put $bgcolor -to 0 0 $size $size
 
 for {set i $gridHalf} {$i<$size} {incr i $grid} {
  for {set j $gridHalf} {$j<$size} {incr j $grid} {
   gridImage$grid put $gridcolor -to  $i $j
   }
 }
 return "gridImage$grid"
}

proc test {text} {
 puts $text
}

